/*
 * linux/arch/arm/plat-s3c/hr-time.c
 *
 * S3C Timers
 *
 * Copyright (c) 2006 Samsung Electronics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/clocksource.h>
#include <linux/clockchips.h>
#include <linux/io.h>

#include <asm/system.h>
#include <mach/hardware.h>
#include <asm/leds.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/mach-types.h>
#include <mach/map.h>
#include <plat/regs-timer.h>
#include <mach/regs-irq.h>
#include <mach/tick.h>

#include <plat/clock.h>
#include <plat/cpu.h>

#ifndef TICK_MAX
#ifdef CONFIG_ARCH_S3C64XX
#define TICK_MAX (0xffffffff)
#else
#define TICK_MAX (0xffff)
#endif
#endif

/*
 * ---------------------------------------------------------------------------
 * PWM timer 4 ... count down to zero, interrupt, reload
 * ---------------------------------------------------------------------------
 */
static int s3c_timer4_start(unsigned long tcnt, int auto_reload)
{
	unsigned long tcon;
#ifdef CONFIG_ARCH_S3C64XX
	unsigned long tcstat;
#endif

	/* stop timer4 */
	tcon  = __raw_readl(S3C2410_TCON);
	tcon &= ~(S3C2410_TCON_T4START | S3C2410_TCON_T4MANUALUPD | S3C2410_TCON_T4RELOAD);
	__raw_writel(tcon, S3C2410_TCON);

#ifdef CONFIG_ARCH_S3C64XX
	tcstat = __raw_readl(S3C64XX_TINT_CSTAT);
	tcstat |=  0x10;
	__raw_writel(tcstat, S3C64XX_TINT_CSTAT);
#else
	if (tcnt > TICK_MAX) {
		panic("s3c_pwm_set_next_event : cycles is too large, cannot configure timer!");
		tcnt = TICK_MAX;
	}
#endif

	/* manual upate TCNT */
	__raw_writel(tcnt, S3C2410_TCNTB(4));
	__raw_writel(tcon | S3C2410_TCON_T4MANUALUPD, S3C2410_TCON);

	if (auto_reload) {
		tcnt--;
		__raw_writel(tcnt, S3C2410_TCNTB(4));
		tcon |= S3C2410_TCON_T4RELOAD;
	}

	tcon |= S3C2410_TCON_T4START;
	__raw_writel(tcon, S3C2410_TCON);

	return 0;
}

static int s3c_timer4_set_next_event(unsigned long cycles,
				struct clock_event_device *evt)
{
	return s3c_timer4_start(cycles, evt->mode == CLOCK_EVT_MODE_PERIODIC);
}

static void s3c_timer_resume(void);
static void s3c_timer4_set_mode(enum clock_event_mode mode,
			      struct clock_event_device *evt)
{
	switch (mode) {
	case CLOCK_EVT_MODE_PERIODIC:
		s3c_timer4_start(((uint64_t) NSEC_PER_SEC / HZ *
				    evt->mult) >> evt->shift, 1);
		break;
	case CLOCK_EVT_MODE_ONESHOT:
	case CLOCK_EVT_MODE_UNUSED:
	case CLOCK_EVT_MODE_SHUTDOWN:
	{
		uint32_t tcon = __raw_readl(S3C2410_TCON);
		tcon &= ~S3C2410_TCON_T4START;
		__raw_writel(tcon, S3C2410_TCON);
	}
		break;
	case CLOCK_EVT_MODE_RESUME:
		s3c_timer_resume();
		break;
	}
}

static struct clock_event_device clockevent_timer4 = {
	.name		= "pwm_timer4",
	.features       = CLOCK_EVT_FEAT_PERIODIC | CLOCK_EVT_FEAT_ONESHOT,
	.shift		= 32,
	.set_next_event	= s3c_timer4_set_next_event,
	.set_mode	= s3c_timer4_set_mode,
};

irqreturn_t s3c_timer4_interrupt(int irq, void *dev_id)
{
	struct clock_event_device *evt = &clockevent_timer4;

	evt->event_handler(evt);

	return IRQ_HANDLED;
}

static struct irqaction s3c_timer4_irq = {
	.name		= "pwm_timer4",
	.flags		= IRQF_DISABLED | IRQF_TIMER | IRQF_IRQPOLL,
	.handler	= s3c_timer4_interrupt,
};

static void s3c_init_timer4_clockevent(unsigned long rate)
{
	clockevent_timer4.mult = div_sc(rate, NSEC_PER_SEC,
					    clockevent_timer4.shift);
	clockevent_timer4.max_delta_ns =
		clockevent_delta2ns(TICK_MAX, &clockevent_timer4);
	clockevent_timer4.min_delta_ns =
		clockevent_delta2ns(1, &clockevent_timer4);

	clockevent_timer4.cpumask = cpumask_of(0);
	clockevents_register_device(&clockevent_timer4);
}


/*
 * ---------------------------------------------------------------------------
 * PWM timer 2 ... free running clock source and scheduler clock
 * ---------------------------------------------------------------------------
 */
static unsigned long s3c_timer2_overflows;

static void s3c_timer2_start(void)
{
	unsigned long tcon;
	unsigned long tcnt;
#ifdef CONFIG_ARCH_S3C64XX
	unsigned long tcstat;

	tcstat = __raw_readl(S3C64XX_TINT_CSTAT);
	tcstat |=  0x04;
	__raw_writel(tcstat, S3C64XX_TINT_CSTAT);
#endif

	tcon  = __raw_readl(S3C2410_TCON);

	tcnt = TICK_MAX;
	tcnt--;
	__raw_writel(tcnt, S3C2410_TCNTB(2));
	__raw_writel(tcnt, S3C2410_TCMPB(2));	

	tcon &= ~(0x0b<<12);

	tcon |= S3C2410_TCON_T2RELOAD;
	tcon |= S3C2410_TCON_T2MANUALUPD;

	__raw_writel(tcon, S3C2410_TCON);

	s3c_timer2_overflows = 0;

	/* start the timer running */
	tcon |= S3C2410_TCON_T2START;
	tcon &= ~S3C2410_TCON_T2MANUALUPD;
	__raw_writel(tcon, S3C2410_TCON);
}

irqreturn_t s3c_timer2_interrupt(int irq, void *dev_id)
{
	s3c_timer2_overflows++;
	return IRQ_HANDLED;
}

struct irqaction s3c_timer2_irq = {
	.name		= "pwm_timer2",
	.flags		= IRQF_DISABLED ,
	.handler	= s3c_timer2_interrupt,
};

static cycle_t s3c_timer2_read(void)
{
#if 0
	unsigned long flags;
	unsigned long value;

	raw_local_irq_save(flags);
	value = (s3c64xx_mpu_timer2_overflows & 0xFFFF) * (TICK_MAX+1) +
			((~__raw_readl(S3C_TIMERREG(0x2c))) & TICK_MAX);
	raw_local_irq_restore(flags);

	return (cycle_t) value;
#else
	return (TICK_MAX - (__raw_readl(S3C_TIMERREG(0x2c)) & TICK_MAX));
#endif
}

struct clocksource clocksource_timer2 = {
	.name		= "clock_source_timer2",
	.rating		= 300,
	.read		= s3c_timer2_read,
#ifdef CONFIG_ARCH_S3C64XX
	.mask		= CLOCKSOURCE_MASK(32),
#else
	.mask		= CLOCKSOURCE_MASK(16),
//	.mask		= CLOCKSOURCE_MASK(32),
#endif
	.shift		= 20,
	.flags		= CLOCK_SOURCE_IS_CONTINUOUS,
};

static void __init s3c_init_timer2_clocksource(unsigned long rate)
{
	static char err[] __initdata = KERN_ERR
			"%s: can't register clocksource!\n";

	clocksource_timer2.mult
		= clocksource_hz2mult(rate, clocksource_timer2.shift);

	s3c_timer2_start();

	if (clocksource_register(&clocksource_timer2))
		printk(err, clocksource_timer2.name);
}


/*
 * ---------------------------------------------------------------------------
 *  Timer initialization
 * ---------------------------------------------------------------------------
 */
static unsigned long s3c_set_timer_rate(void)
{
	static struct clk	*ck_ref = NULL;
	unsigned long	rate;
	unsigned long	tcfg1;
	unsigned long	tcfg0;

	if (ck_ref == NULL) {
		ck_ref = clk_get(NULL, "timers");
		clk_enable(ck_ref);
	}

	if (IS_ERR(ck_ref))
		panic("failed to get clock for system timer");

	rate = clk_get_rate(ck_ref);
	/* clk_put(ck_ref); */

	tcfg1 = __raw_readl(S3C2410_TCFG1);
	tcfg0 = __raw_readl(S3C2410_TCFG0);

	tcfg1 &= ~(S3C2410_TCFG1_MUX4_MASK | S3C2410_TCFG1_MUX2_MASK);
#ifdef CONFIG_ARCH_S3C64XX
	tcfg1 |= S3C2410_TCFG1_MUX4_DIV1;
	tcfg0 &= ~S3C2410_TCFG_PRESCALER1_MASK;
	tcfg0 |= (0) << S3C2410_TCFG_PRESCALER1_SHIFT;
#else
	tcfg1 |= (S3C2410_TCFG1_MUX4_DIV2 | S3C2410_TCFG1_MUX2_DIV2);
	tcfg0 &= ~S3C2410_TCFG_PRESCALER1_MASK;
#if 0
	tcfg0 |= (5) << S3C2410_TCFG_PRESCALER1_SHIFT;
	rate /= 12;
#else
	tcfg0 |= (32) << S3C2410_TCFG_PRESCALER1_SHIFT;
	rate /= 66;
#endif
#endif

	__raw_writel(tcfg1, S3C2410_TCFG1);
	__raw_writel(tcfg0, S3C2410_TCFG0);

	return rate;
}

static void s3c_timer_setup(void)
{
	unsigned long	rate = s3c_set_timer_rate();

	s3c_init_timer4_clockevent(rate);
	s3c_init_timer2_clocksource(rate);

	printk(KERN_INFO "s3c clock event/source resolution : %luhz\n", rate);
	printk(KERN_INFO "s3c clockevent : max_delta_ns(%lu), min_delta_ns(%lu)\n", clockevent_timer4.max_delta_ns, clockevent_timer4.min_delta_ns);
}

static void s3c_timer_resume(void)
{
	s3c_set_timer_rate();
	s3c_timer2_start();
}

static void __init s3c_timer_init(void)
{
	s3c_timer_setup();
	setup_irq(IRQ_TIMER2, &s3c_timer2_irq);	
	setup_irq(IRQ_TIMER4, &s3c_timer4_irq);
}

#ifdef CONFIG_ARCH_S3C64XX
struct sys_timer s3c64xx_timer = {
	.init		= s3c_timer_init,
};
#else
struct sys_timer s3c24xx_timer = {
	.init		= s3c_timer_init,
};
#endif
