/* linux/drivers/char/s3c_fimg2d2x.c
 *
 * Driver file for Samsung 2D Accelerator(FIMG-2D)
 *
 * Jonghun Han, Copyright (c) 2009 Samsung Electronics
 * [url]http://www.samsungsemi.com/[/url]
 *
 * add some notes by [email]tommy.hong@2009.05.09[/email]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */ 
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/clk.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/irq.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <mach/map.h>
#include <plat/regs-g2d.h>
#include "s3c_fimg2d2x.h"

#define S3C_G2D_MINOR	243

#define S3C_G2D_FIFO_NUM	32

#define S3C_G2D_TIMEOUT		HZ

/* rendering pipe line
 * 1. primitive drawing (Line/Point, Bitblt, Font Drawing)
 * 2. Rotation
 * 3. Clipping
 * 4. Stencil Test
 * 5. ROP
 * 6. Alpha Blending
 */

static struct s3c_g2d {
	struct clk		*clk;
	int			irq;
	struct resource 	*res;
	void __iomem		*regs;

	wait_queue_head_t	waitq;

	struct device           *dev;

	dma_addr_t              map_dma;        /* physical */
        u_char *                map_cpu;        /* virtual */
        u_int                   map_size;

        u_int                   smem_len;

	int			open_cnt;
} s3c_g2d;

static int s3c_g2d_wait_idle(struct s3c_g2d *g2d);
static int s3c_g2d_wait_empty_fifo(struct s3c_g2d *g2d, int nempty);

static inline int s3c_g2d_check_idle(struct s3c_g2d *g2d)
{
	u32	data = readl(g2d->regs+S3C_G2D_FIFO_STAT_REG);
	data &= S3C_G2D_FIFO_STAT_IDLE;
	return (data == S3C_G2D_FIFO_STAT_IDLE) ? 1 : 0;
}


static void s3c_g2d_dump_regs(struct s3c_g2d *g2d)
{
printk("===================================================================\n");
printk("INTEN_REG           : %08X\n", readl(g2d->regs + S3C_G2D_INTEN_REG));
printk("FIFO_INTC_REG       : %08X\n", readl(g2d->regs + S3C_G2D_FIFO_INTC_REG));
printk("INTC_PEND_REG       : %08X\n", readl(g2d->regs + S3C_G2D_INTC_PEND_REG));
printk("FIFO_STAT_REG       : %08X\n", readl(g2d->regs + S3C_G2D_FIFO_STAT_REG));
printk("SRC_ RES_REG        : %08X\n", readl(g2d->regs + S3C_G2D_SRC_RES_REG));
printk("SRC_HORI_RES_REG    : %08X\n", readl(g2d->regs + S3C_G2D_SRC_HORI_RES_REG));
printk("SRC_VERT_RES_REG    : %08X\n", readl(g2d->regs + S3C_G2D_SRC_VERT_RES_REG));
printk("SC_RES_REG          : %08X\n", readl(g2d->regs + S3C_G2D_SC_RES_REG));
printk("SC_HORI_RES_REG     : %08X\n", readl(g2d->regs + S3C_G2D_SC_HORI_RES_REG));
printk("SC_VERT_RES_REG     : %08X\n", readl(g2d->regs + S3C_G2D_SC_VERT_RES_REG));
printk("CW_LT_REG           : %08X\n", readl(g2d->regs + S3C_G2D_CW_LT_REG));
printk("CW_LT_X_REG         : %08X\n", readl(g2d->regs + S3C_G2D_CW_LT_X_REG));
printk("CW_LT_Y_REG         : %08X\n", readl(g2d->regs + S3C_G2D_CW_LT_Y_REG));
printk("CW_RB_REG           : %08X\n", readl(g2d->regs + S3C_G2D_CW_RB_REG));
printk("CW_RB_X_REG         : %08X\n", readl(g2d->regs + S3C_G2D_CW_RB_X_REG));
printk("CW_RB_Y_REG         : %08X\n", readl(g2d->regs + S3C_G2D_CW_RB_Y_REG));
printk("COORD0_REG          : %08X\n", readl(g2d->regs + S3C_G2D_COORD0_REG));
printk("COORD0_X_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD0_X_REG));
printk("COORD0_Y_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD0_Y_REG));
printk("COORD1_REG          : %08X\n", readl(g2d->regs + S3C_G2D_COORD1_REG));
printk("COORD1_X_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD1_X_REG));
printk("COORD1_Y_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD1_Y_REG));
printk("COORD2_REG          : %08X\n", readl(g2d->regs + S3C_G2D_COORD2_REG));
printk("COORD2_X_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD2_X_REG));
printk("COORD2_Y_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD2_Y_REG));
printk("COORD3_REG          : %08X\n", readl(g2d->regs + S3C_G2D_COORD3_REG));
printk("COORD3_X_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD3_X_REG));
printk("COORD3_Y_REG        : %08X\n", readl(g2d->regs + S3C_G2D_COORD3_Y_REG));
printk("ROT_OC_REG          : %08X\n", readl(g2d->regs + S3C_G2D_ROT_OC_REG));
printk("ROT_OC_X_REG        : %08X\n", readl(g2d->regs + S3C_G2D_ROT_OC_X_REG));
printk("ROT_OC_Y_REG        : %08X\n", readl(g2d->regs + S3C_G2D_ROT_OC_Y_REG));
printk("ROTATE_REG          : %08X\n", readl(g2d->regs + S3C_G2D_ROTATE_REG));
printk("X_INCR_REG          : %08X\n", readl(g2d->regs + S3C_G2D_X_INCR_REG));
printk("Y_INCR_REG          : %08X\n", readl(g2d->regs + S3C_G2D_Y_INCR_REG));
printk("ROP_REG             : %08X\n", readl(g2d->regs + S3C_G2D_ROP_REG));
printk("ALPHA_REG           : %08X\n", readl(g2d->regs + S3C_G2D_ALPHA_REG));
printk("FG_COLOR_REG        : %08X\n", readl(g2d->regs + S3C_G2D_FG_COLOR_REG));
printk("BG_COLOR_REG        : %08X\n", readl(g2d->regs + S3C_G2D_BG_COLOR_REG));
printk("BS_COLOR_REG        : %08X\n", readl(g2d->regs + S3C_G2D_BS_COLOR_REG));
printk("SRC_COLOR_MODE_REG  : %08X\n", readl(g2d->regs + S3C_G2D_SRC_COLOR_MODE_REG));
printk("DEST_COLOR_MODE_REG : %08X\n", readl(g2d->regs + S3C_G2D_DEST_COLOR_MODE_REG));
printk("STENCIL_CNTL_REG    : %08X\n", readl(g2d->regs + S3C_G2D_STENCIL_CNTL_REG));
printk("STENCIL_DR_MIN_REG  : %08X\n", readl(g2d->regs + S3C_G2D_STENCIL_DR_MIN_REG));
printk("STENCIL_DR_MAX_REG  : %08X\n", readl(g2d->regs + S3C_G2D_STENCIL_DR_MAX_REG));
printk("SRC_BASE_ADDR_REG   : %08X\n", readl(g2d->regs + S3C_G2D_SRC_BASE_ADDR_REG));
printk("DEST_BASE_ADDR_REG  : %08X\n", readl(g2d->regs + S3C_G2D_DEST_BASE_ADDR_REG));
printk("===================================================================\n");
}

static inline void s3c_g2d_set_fg_color(struct s3c_g2d *g2d, u32 fg_color)
{
	writel(fg_color, g2d->regs + S3C_G2D_FG_COLOR_REG);
}

static void s3c_g2d_set_alpha_blending(struct s3c_g2d *g2d, u32 alpha_mode, u32 alpha, u32 fading)
{
	void __iomem *regs = g2d->regs;
	u32 data;

	data = readl(regs + S3C_G2D_ROP_REG);
	data &= ~S3C_G2D_ROP_ALPHA_MODE_MASK;
	data |= (alpha_mode << S3C_G2D_ROP_ALPHA_MODE_SHIFT) & S3C_G2D_ROP_ALPHA_MODE_MASK;
	writel(data, regs + S3C_G2D_ROP_REG);

	data = ((alpha & S3C_G2D_ALPHA_ALPHA_MASK) | ((fading << S3C_G2D_ALPHA_FADING_SHIFT) & S3C_G2D_ALPHA_FADING_MASK));
	writel(data, regs + S3C_G2D_ALPHA_REG);
}

static void s3c_g2d_set_pattern(struct s3c_g2d *g2d, const u16 pattern[64], u32 poffsetX, u32 poffsetY)
{
	void __iomem *regs = g2d->regs;
	int i;
	u32 data;

	for (i = 0; i < 32; i++) {
		data = ((u32)pattern[2*i] << 16) | pattern[2*i+1];
		writel(data, regs + S3C_G2D_PATTERN_REG(i));
	}
	data = (poffsetY << S3C_G2D_PATOFF_Y_SHIFT) | poffsetX;
	writel(data, regs + S3C_G2D_PATOFF_REG);
	writel(poffsetX, regs + S3C_G2D_PATOFF_X_REG);
	writel(poffsetY, regs + S3C_G2D_PATOFF_Y_REG);
}

static void s3c_g2d_set_rop(struct s3c_g2d *g2d, u32 rop, u32 rop_os)
{
	void __iomem *regs = g2d->regs;
	u32 data;

	data = readl(regs + S3C_G2D_ROP_REG);
	data &= ~(S3C_G2D_ROP_ROP_MASK | S3C_G2D_ROP_OS_MASK);
	data |= rop & S3C_G2D_ROP_ROP_MASK;
	data |= (rop_os << S3C_G2D_ROP_OS_SHIFT) & S3C_G2D_ROP_OS_MASK;
	writel(data, regs + S3C_G2D_ROP_REG);
}

static void s3c_g2d_set_stencil(struct s3c_g2d *g2d, u32 ctrl, u32 key_min, u32 key_max)
{
	void __iomem *regs = g2d->regs;

	writel(ctrl, regs + S3C_G2D_STENCIL_CNTL_REG);
	writel(key_min, regs + S3C_G2D_STENCIL_DR_MIN_REG);
	writel(key_max, regs + S3C_G2D_STENCIL_DR_MAX_REG);
}

static void s3c_g2d_set_clipping(struct s3c_g2d *g2d, u32 x1, u32 y1, u32 x2, u32 y2)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	data = (x1 & 0x7F) | ((y1 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_CW_LT_REG);
	writel(x1, regs + S3C_G2D_CW_LT_X_REG);
	writel(y1, regs + S3C_G2D_CW_LT_Y_REG);
	data = (x2 & 0x7F) | ((y2 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_CW_RB_REG);
	writel(x2, regs + S3C_G2D_CW_RB_X_REG);
	writel(y2, regs + S3C_G2D_CW_RB_Y_REG);
}

/*   0 : (x         ,y)
 *  90 : (-y+(ox+oy),x-(ox-oy))
 * 180 : (-x+2ox    ,-y+2oy)
 * 270 : (y+(ox-oy) ,-x+(ox+oy))
 * xfl : (x         ,-y+2oy)
 * yfl : (-x+2ox    ,y)
 */
static void s3c_g2d_set_rotation(struct s3c_g2d *g2d, u32 mode, u32 ox, u32 oy)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	data = (ox & 0x7F) | ((oy & 0x7F) << 16);
	writel(data, regs + S3C_G2D_ROT_OC_REG);
	writel(ox, regs + S3C_G2D_ROT_OC_X_REG);
	writel(oy, regs + S3C_G2D_ROT_OC_Y_REG);
	writel(mode, regs + S3C_G2D_ROTATE_REG);
}

static void s3c_g2d_set_destination(struct s3c_g2d *g2d, u32 base_addr, u32 w, u32 h, u32 color_mode)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	writel(color_mode, regs + S3C_G2D_DEST_COLOR_MODE_REG);
	writel(base_addr, regs + S3C_G2D_DEST_BASE_ADDR_REG);
	data = (w & 0x7F) | ((h & 0x7F) << 16);
	writel(data, regs + S3C_G2D_SC_RES_REG);
	writel(w, regs + S3C_G2D_SC_HORI_RES_REG);
	writel(h, regs + S3C_G2D_SC_VERT_RES_REG);
#if 0
	/* reset clipping window */
	s3c_g2d_set_clipping(g2d, 0, 0, w-1, h-1);
#endif
}

static void s3c_g2d_set_source(struct s3c_g2d *g2d, u32 base_addr, u32 w, u32 h, u32 color_mode)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	writel(color_mode, regs + S3C_G2D_SRC_COLOR_MODE_REG);
	writel(base_addr, regs + S3C_G2D_SRC_BASE_ADDR_REG);
	data = (w & 0x7F) | ((h & 0x7F) << 16);
	writel(data, regs + S3C_G2D_SRC_RES_REG);
	writel(w, regs + S3C_G2D_SRC_HORI_RES_REG);
	writel(h, regs + S3C_G2D_SRC_VERT_RES_REG);
}

static void s3c_g2d_set_bitblt_src_range(struct s3c_g2d *g2d, u32 x1, u32 y1, u32 x2, u32 y2)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	data = (x1 & 0x7F) | ((y1 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_COORD0_REG);
	writel(x1, regs + S3C_G2D_COORD0_X_REG);
	writel(y1, regs + S3C_G2D_COORD0_Y_REG);
	data = (x2 & 0x7F) | ((y2 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_COORD1_REG);
	writel(x2, regs + S3C_G2D_COORD1_X_REG);
	writel(y2, regs + S3C_G2D_COORD1_Y_REG);
}

static void s3c_g2d_set_bitblt_dst_range(struct s3c_g2d *g2d, u32 x1, u32 y1, u32 x2, u32 y2)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	data = (x1 & 0x7F) | ((y1 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_COORD2_REG);
	writel(x1, regs + S3C_G2D_COORD2_X_REG);
	writel(y1, regs + S3C_G2D_COORD2_Y_REG);
	data = (x2 & 0x7F) | ((y2 & 0x7F) << 16);
	writel(data, regs + S3C_G2D_COORD3_REG);
	writel(x2, regs + S3C_G2D_COORD3_X_REG);
	writel(y2, regs + S3C_G2D_COORD3_Y_REG);
}

static void s3c_g2d_set_bitblt_transparent(struct s3c_g2d *g2d, u32 mode, u32 bg_color, u32 bs_color)
{
	void __iomem *regs = g2d->regs;
	u32 data;
	writel(bg_color, regs + S3C_G2D_BG_COLOR_REG);
	writel(bs_color, regs + S3C_G2D_BS_COLOR_REG);
	data = readl(regs + S3C_G2D_ROP_REG);
	data &= ~S3C_G2D_ROP_TRANSPARENT_MASK;
	data |= (mode << S3C_G2D_ROP_TRANSPARENT_SHIFT) & S3C_G2D_ROP_TRANSPARENT_MASK;
	writel(data, regs + S3C_G2D_ROP_REG);
}

static void s3c_g2d_start_bitblt_m2s(struct s3c_g2d *g2d, u32 mode, u32 incrx, u32 incry)
{
	void __iomem *regs = g2d->regs;

	writel(incrx, regs + S3C_G2D_X_INCR_REG);
	writel(incry, regs + S3C_G2D_Y_INCR_REG);

	/* s3c_g2d_dump_regs(g2d); */

	writel(mode, regs + S3C_G2D_CMD1_REG);
}

#if 0
static void s3c_g2d_calc_rotate_origin(u32 mode, u32 src_w, u32 src_h, u32 *ox, u32 *oy)
{
	switch (mode & 0xF) {
	case S3C_G2D_ROTATE_90:
		*ox = *oy = src_h/2;
		break;
	case S3C_G2D_ROTATE_180:
		*ox = src_w / 2;
		*oy = src_h / 2;
		break;
	case S3C_G2D_ROTATE_270:
		*ox = *oy = src_w/2;
		break;
	default:
		*ox = *oy = 0;
		if (mode & S3C_G2D_ROTATE_X_FLIP)
			*oy += src_h / 2;
		if (mode & S3C_G2D_ROTATE_Y_FLIP)
			*ox += src_w / 2;
		break;
	}
}
#endif

static void s3c_g2d_bitblt_set_param(struct s3c_g2d *g2d, struct s3c_g2d_bitblt_rotate *bitblt)
{
#if 0
	u32	dst_w, dst_h;
	u32	ox, oy;

	if ((bitblt->rotation_mode & S3C_G2D_ROTATE_90) ||
	    (bitblt->rotation_mode & S3C_G2D_ROTATE_270))
	{
		dst_w = bitblt->src_h;
		dst_h = bitblt->src_w;
	} else {
		dst_w = bitblt->src_w;
		dst_h = bitblt->src_h;
	}
#endif

	s3c_g2d_set_alpha_blending(g2d, bitblt->alpha.alpha_mode, bitblt->alpha.alpha, bitblt->alpha.fading);

	s3c_g2d_set_rop(g2d, S3C_G2D_ROPV_SOURCE, 0);

#if 0
	s3c_g2d_calc_rotate_origin(bitblt->rotation_mode, bitblt->src_w, bitblt->src_h, &ox, &oy);
#endif
	s3c_g2d_set_rotation(g2d, bitblt->rotation_mode, bitblt->ox,bitblt->oy);

	s3c_g2d_set_source(g2d, bitblt->src_fb.base_addr,
				bitblt->src_fb.w, bitblt->src_fb.h,
				bitblt->src_fb.color_mode);
	s3c_g2d_set_destination(g2d, bitblt->dst_fb.base_addr,
				     bitblt->dst_fb.w, bitblt->src_fb.h,
				     bitblt->dst_fb.color_mode);
	s3c_g2d_set_clipping(g2d, bitblt->clip_range.x1, bitblt->clip_range.y1,
			     bitblt->clip_range.x2, bitblt->clip_range.y2);

	s3c_g2d_set_bitblt_src_range(g2d, bitblt->src_range.x1,
					  bitblt->src_range.y1,
					  bitblt->src_range.x2,
					  bitblt->src_range.y2);

	s3c_g2d_set_bitblt_dst_range(g2d, bitblt->dst_range.x1,
					  bitblt->dst_range.y1,
					  bitblt->dst_range.x2,
					  bitblt->dst_range.y2);

	s3c_g2d_set_bitblt_transparent(g2d, bitblt->transp.mode,
					    bitblt->transp.bg_color,
					    bitblt->transp.bs_color);
}

static int s3c_g2d_bitblt_rotate(struct s3c_g2d *g2d, struct s3c_g2d_bitblt_rotate *bitblt)
{
	int ret;

	if ((ret = s3c_g2d_wait_idle(g2d)))
		return ret;

	s3c_g2d_bitblt_set_param(g2d, bitblt);
	s3c_g2d_start_bitblt_m2s(g2d, S3C_G2D_BITBLT_NORMAL, 0, 0);

	ret = s3c_g2d_wait_idle(g2d);
	return ret;
}

static int s3c_g2d_bitblt_host_rotate(struct s3c_g2d *g2d, struct s3c_g2d_bitblt_rotate *bitblt, u32 rot_mode)
{
	int ret;
	const u32	*sbuf = (u32 *) (g2d->map_cpu +
		(bitblt->src_fb.base_addr - g2d->map_dma) +
		(bitblt->src_range.y1 * bitblt->src_fb.w +
			bitblt->src_range.x1) * 4);
	const u32	*cbuf;

	u32 dw = bitblt->dst_range.x2 - bitblt->dst_range.x1 + 1;
	u32 dh = bitblt->dst_range.y2 - bitblt->dst_range.y1 + 1;
	u32 x, y;

	if ((ret = s3c_g2d_wait_idle(g2d))) {
		printk("wait error 1\n");
		return ret;
	}

	bitblt->rotation_mode = S3C_G2D_ROTATE_0;
	s3c_g2d_bitblt_set_param(g2d, bitblt);

	if (rot_mode == S3C_G2D_ROTATE_90)
		cbuf = sbuf + (dw-1) * bitblt->src_fb.w;
	else
		cbuf = sbuf + (dh-1);
	writel(*cbuf, g2d->regs + S3C_G2D_CMD2_REG);

	for (y = 0; y < dh; y++) {
		for (x = (y ? 0 : 1); x < dw; x++) {
			if ((ret = s3c_g2d_wait_empty_fifo(g2d, 1))) {
				printk("wait error 3\n");
				return ret;
			}
			writel(*cbuf, g2d->regs + S3C_G2D_CMD3_REG);

			if (rot_mode == S3C_G2D_ROTATE_90)
				cbuf -= bitblt->src_fb.w;
			else
				cbuf += bitblt->src_fb.w;
		}

		if (rot_mode == S3C_G2D_ROTATE_90)
			cbuf = sbuf + (y+1) + (dw-1) * bitblt->src_fb.w;
		else
			cbuf = sbuf + (dh-(y+2));
	}

#if 0
	if ((ret = s3c_g2d_wait_idle(g2d))) {
		printk("wait error 4\n");
		return ret;
	}
	return 0;
#else
	{
	u32 i;
	for (i = 0; i < 0xFFFFFFFF; i++)
		if (s3c_g2d_check_idle(g2d))
			return 0;
		else
			schedule();
	}
	return -EBUSY;
#endif
}

/*
 * Here is S3C FIMG2D command FIFO check or monitor ,must do it by user;
 * If FIFO Entries reach FIFO FULL ,and full interrupt occurs.
 * FIFO full number is 32.
 */
static inline int s3c_g2d_check_empty_fifo(struct s3c_g2d *g2d, int nempty)
{
	u32	data = readl(g2d->regs + S3C_G2D_FIFO_STAT_REG);
	return (S3C_G2D_FIFO_USED(data) > (S3C_G2D_FIFO_NUM - nempty)) ? 0 : 1;
}

static int s3c_g2d_wait_empty_fifo(struct s3c_g2d *g2d, int nempty)
{
#if 0
        while(! s3c_g2d_check_empty_fifo(g2d, nempty)) {
		if (interruptible_sleep_on_timeout(& g2d->waitq,
							S3C_G2D_TIMEOUT)  == 0)
		{
                        printk(KERN_ERR "\n%s: Waiting for interrupt is timeout\n", __FUNCTION__);
			if (s3c_g2d_check_empty_fifo(g2d, nempty))
				return 0;
			else
				return -EBUSY;
		}
	}
	return 0;
#else
	u32	i;
	for (i = 0; i < 0xFFFFFFFF; i++) {
		if (s3c_g2d_check_empty_fifo(g2d, nempty))
			return 0;
		else
			schedule();
	}
	return -EBUSY;
#endif
}

static int s3c_g2d_wait_idle(struct s3c_g2d *g2d)
{
	while (! s3c_g2d_check_idle(g2d)) {
		if (interruptible_sleep_on_timeout(& g2d->waitq,
							S3C_G2D_TIMEOUT)  == 0)
		{
                        printk(KERN_ERR "\n%s: Waiting for interrupt is timeout\n", __FUNCTION__);
			if (s3c_g2d_check_idle(g2d))
				return 0;
			else
				return -EBUSY;
		}
	}
	return 0;
}

static irqreturn_t s3c_g2d_irq(int irq, void *dev_id)
{
	struct s3c_g2d	*g2d = (struct s3c_g2d *) dev_id;
	u32	data = readl(g2d->regs + S3C_G2D_INTC_PEND_REG);

#if 0
        if(data & S3C_G2D_INTP_CMD_FIN) {
                writel(S3C_G2D_INTP_CMD_FIN, g2d->regs + S3C_G2D_INTC_PEND_REG);
                wake_up_interruptible(&g2d->waitq);
		printk("s3c_g2d_irq : S3C_G2D_INTC_PEND_REG -> %x\n", data);
                /* s3c_g2d_poll_flag = 1; */
        }
#else
        if(data) {
                writel(data, g2d->regs + S3C_G2D_INTC_PEND_REG);
                wake_up_interruptible(&g2d->waitq);
                /* s3c_g2d_poll_flag = 1; */
        }
#endif

        return IRQ_HANDLED;
}

static int s3c_g2d_open(struct inode *inode, struct file *file)
{
	if (s3c_g2d.open_cnt)
		return -EBUSY;
	file->private_data = &s3c_g2d;
	s3c_g2d.open_cnt++;
	return 0;
}

static int s3c_g2d_release(struct inode *inode, struct file *file)
{
	struct s3c_g2d	*g2d = (struct s3c_g2d *) file->private_data;
	if (g2d->open_cnt > 0)
		g2d->open_cnt--;
	return 0;
}

static int s3c_g2d_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct s3c_g2d	*g2d = (struct s3c_g2d *) file->private_data;
	void __user *argp = (void __user *)arg;

        if (cmd == S3C_G2D_BITBLT) {
                struct s3c_g2d_bitblt_rotate  bitblt;
                if (copy_from_user(&bitblt, argp, sizeof(bitblt)))
                        return -EFAULT;

		return s3c_g2d_bitblt_rotate(g2d, &bitblt);
	} else if (cmd == S3C_G2D_BITBLT_HOST) {
                struct s3c_g2d_bitblt_rotate  bitblt;
                if (copy_from_user(&bitblt, argp, sizeof(bitblt)))
                        return -EFAULT;

		if (bitblt.rotation_mode != S3C_G2D_ROTATE_90 &&
		    bitblt.rotation_mode != S3C_G2D_ROTATE_270)
		{
			return -EINVAL;
		}
		return s3c_g2d_bitblt_host_rotate(g2d, &bitblt, bitblt.rotation_mode);
	} else if (cmd == S3C_G2D_SET_MEMORY) {
                unsigned long   smem_len;
                unsigned long   map_size;
                u_char          *map_cpu;
                dma_addr_t      map_dma;

                if (copy_from_user(&smem_len, argp, sizeof(map_size)))
                        return -EFAULT;
                if (smem_len <= g2d->map_size) {
			g2d->smem_len = smem_len;
                        map_dma = g2d->map_dma;
                        if (copy_to_user(argp, &map_dma, sizeof(map_dma)))
                                return -EFAULT;
                        return 0;
                }

                map_size = PAGE_ALIGN(smem_len + PAGE_SIZE);
		if (map_size < 0x400000)
			map_size = 0x400000;

                if (g2d->map_cpu) {
                        dma_free_writecombine(g2d->dev,
                                              g2d->map_size,
                                              g2d->map_cpu,
                                              g2d->map_dma);
                        g2d->map_cpu = NULL;
                        g2d->map_size = 0;
                        g2d->map_dma = 0;
                        g2d->smem_len = 0;
                }

                map_cpu  = dma_alloc_writecombine(g2d->dev, map_size,
                                                  &map_dma, GFP_KERNEL);
                if (! map_cpu)
                        return -ENOMEM;

                g2d->map_size = map_size;
                g2d->map_cpu = map_cpu;
                g2d->map_dma = map_dma;
                g2d->smem_len = smem_len;
                if (copy_to_user(argp, &map_dma, sizeof(map_dma)))
                        return -EFAULT;
		return 0;
	}

	return -ENOIOCTLCMD;
}

static int
s3c_g2d_mmap(struct file *file, struct vm_area_struct * vma)
{
	struct s3c_g2d	*g2d = (struct s3c_g2d *) file->private_data;
        unsigned long off;
        unsigned long start;
        u32 len;

        if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
                return -EINVAL;
        off = vma->vm_pgoff << PAGE_SHIFT;

        if (! g2d->map_cpu)
                return -EINVAL;

        lock_kernel();

        /* frame buffer memory */
        start = g2d->map_dma;
        len = PAGE_ALIGN((start & ~PAGE_MASK) + g2d->smem_len);
        unlock_kernel();

        if (off >= len)
                return -EINVAL;

        start &= PAGE_MASK;
        if ((vma->vm_end - vma->vm_start + off) > len)
                return -EINVAL;
        off += start;
        vma->vm_pgoff = off >> PAGE_SHIFT;
        /* This is an IO map - tell maydump to skip this VMA */
        vma->vm_flags |= VM_IO | VM_RESERVED;
#if 0
        vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
#endif
        if (io_remap_pfn_range(vma, vma->vm_start, off >> PAGE_SHIFT,
                             vma->vm_end - vma->vm_start, vma->vm_page_prot))
                return -EAGAIN;
        return 0;
}

static const struct file_operations s3c_g2d_fops = {
        .owner          = THIS_MODULE,
        .open           = s3c_g2d_open,
        .release        = s3c_g2d_release,
        .ioctl          = s3c_g2d_ioctl,
        .mmap           = s3c_g2d_mmap,
};

static struct miscdevice s3c_g2d_dev = {
        .minor		= S3C_G2D_MINOR,
        .name		= "s3c_g2d",
        .fops		= &s3c_g2d_fops,
};

static int __init s3c_g2d_probe(struct platform_device *pdev)
{
        struct resource *res;
        int ret;

	s3c_g2d.dev = &pdev->dev;

        s3c_g2d.clk = clk_get(&pdev->dev, "g2d");
        if(s3c_g2d.clk == NULL || IS_ERR(s3c_g2d.clk)) {
                printk(KERN_ERR "%s : failed to find g2d clock source\n", __func__);
                return -ENOENT;
        }
        clk_enable(s3c_g2d.clk);

        res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
        if(res == NULL) {
                printk(KERN_ERR "%s : failed to get memory region resouce\n", __func__);
                ret = -ENOENT;
		goto err_clk;
        }

        s3c_g2d.irq = platform_get_irq(pdev, 0);
	if (s3c_g2d.irq == -1) {
                printk(KERN_ERR "%s : failed to get irq resouce\n", __func__);
                ret = -ENOENT;
		goto err_clk;
	}

        s3c_g2d.res = request_mem_region(res->start, res->end-res->start+1, pdev->name);
        if(s3c_g2d.res == NULL) {
                printk(KERN_ERR "%s : failed to reserve memory region\n", __func__);
                ret = -ENOENT;
		goto err_clk;
        }

        s3c_g2d.regs = ioremap(s3c_g2d.res->start, s3c_g2d.res->end-s3c_g2d.res->start+1);
        if(s3c_g2d.regs == NULL) {
                printk(KERN_ERR "%s : failed ioremap\n", __func__);
                ret = -ENXIO;
		goto err_req_region;
        }

        init_waitqueue_head(& s3c_g2d.waitq);

        ret = request_irq(s3c_g2d.irq, s3c_g2d_irq, 0, pdev->name, &s3c_g2d);
        if (ret) {
                printk(KERN_ERR "%s : request_irq failed.\n", __func__);
		goto err_ioremap;
        }

        writel(S3C_G2D_INTEN_ACF, s3c_g2d.regs + S3C_G2D_INTEN_REG);

        ret = misc_register(&s3c_g2d_dev);
        if (ret) {
                printk (KERN_ERR "%s : cannot register miscdev on minor=%d (%d)\n",
                        __func__, S3C_G2D_MINOR, ret);
		goto err_request_irq;
        }

        printk(KERN_INFO "S3C_G2D driver initialized\n");

        return 0;

err_request_irq:
        free_irq(s3c_g2d.irq, &s3c_g2d);
err_ioremap:
	iounmap(s3c_g2d.regs);
err_req_region:
	release_resource(s3c_g2d.res);
err_clk:
	clk_disable(s3c_g2d.clk);
        clk_put(s3c_g2d.clk);
	return ret;
}

static int s3c_g2d_remove(struct platform_device *dev)
{
        misc_deregister(&s3c_g2d_dev);
        free_irq(s3c_g2d.irq, &s3c_g2d);
	iounmap(s3c_g2d.regs);
	release_resource(s3c_g2d.res);
	clk_disable(s3c_g2d.clk);
        clk_put(s3c_g2d.clk);

	if (s3c_g2d.map_cpu) {
                dma_free_writecombine(s3c_g2d.dev,
                                      s3c_g2d.map_size,
                                      s3c_g2d.map_cpu,
                                      s3c_g2d.map_dma);
                s3c_g2d.map_cpu = NULL;
                s3c_g2d.map_size = 0;
                s3c_g2d.map_dma = 0;
                s3c_g2d.smem_len = 0;
        }

        return 0;
}

static int s3c_g2d_suspend(struct platform_device *pdev, pm_message_t state)
{
	clk_disable(s3c_g2d.clk);
	return 0;
}

static int s3c_g2d_resume(struct platform_device *pdev)
{
	clk_enable(s3c_g2d.clk);
        writel(S3C_G2D_INTEN_ACF, s3c_g2d.regs + S3C_G2D_INTEN_REG);
	return 0;
}

static struct platform_driver s3c_g2d_driver = {
	.probe		= s3c_g2d_probe,
	.remove		= s3c_g2d_remove,
	.suspend	= s3c_g2d_suspend,
	.resume		= s3c_g2d_resume,
	.driver		= {
		.name	= "s3c-g2d",
		.owner	= THIS_MODULE,
	},
};

int __devinit s3c_g2d_init(void)
{
	return platform_driver_register(&s3c_g2d_driver);
}

static void __exit s3c_g2d_cleanup(void)
{
	platform_driver_unregister(&s3c_g2d_driver);
}

module_init(s3c_g2d_init);
module_exit(s3c_g2d_cleanup);

MODULE_DESCRIPTION("G2D driver for S3C");
MODULE_LICENSE("GPL");
