#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/time.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/platform_device.h>

#include <mach/hardware.h>
#include <mach/regs-gpio.h>


#define ASI_REMOCON_MAJOR	62

#define ASI_REMOCON_IRQ		IRQ_EINT0	

/* #define DEBUG_REMOCON */	
#undef DEBUG_REMOCON

#ifdef DEBUG_REMOCON
#define TRACE(fmt, arg...) printk(KERN_WARNING "remocon: " fmt, ## arg)
#else
#define TRACE(fmt, arg...) do {} while (0)
#endif

#define LOW_USEC_HEADER1	36960
#define MIN_LOW_USEC_HEADER1	33960
#define MAX_LOW_USEC_HEADER1	40960

#define HIGH_USEC_HEADER2	 2200
#define MIN_HIGH_USEC_HEADER2	 1900
#define MAX_HIGH_USEC_HEADER2	 2600

#define LOW_USEC_HEADER3	41440	
#define MIN_LOW_USEC_HEADER3	18440
#define MAX_LOW_USEC_HEADER3	50440	

#define HIGH_USEC_HEADER4	31360
#define MIN_HIGH_USEC_HEADER4	18340
#define MAX_HIGH_USEC_HEADER4	50360

#define LOW_USEC_HEADER5	11200	
#define MIN_LOW_USEC_HEADER5	 9200	
#define MAX_LOW_USEC_HEADER5	14200

#define HIGH_USEC_HEADER6	80000
#define MIN_HIGH_USEC_HEADER6	78000
#define MAX_HIGH_USEC_HEADER6	84000

#define LOW_USEC_REPEAT1         3360 
#define MIN_LOW_USEC_REPEAT1   	 2960 
#define MAX_LOW_USEC_REPEAT1     3760

#define HIGH_USEC_REPEAT2       10080
#define MIN_HIGH_USEC_REPEAT2    8080
#define MAX_HIGH_USEC_REPEAT2   12080

#define LOW_USEC_REPEAT3        1120
#define MIN_LOW_USEC_REPEAT3    820 
#define MAX_LOW_USEC_REPEAT3    1420




#define LOW_USEC_START          9000
#define MIN_LOW_USEC_START      8000
#define MAX_LOW_USEC_START     	10000

#define LOW_USEC_CODE           1120 
#define MIN_LOW_USEC_CODE       800
#define MAX_LOW_USEC_CODE       1520

#define HIGH_USEC_LEAD          9000
#define MIN_HIGH_USEC_LEAD      8000
#define MAX_HIGH_USEC_LEAD      10000



#define HIGH_USEC_BIT0          1120 
#define MIN_HIGH_USEC_BIT0      620 
#define MAX_HIGH_USEC_BIT0      1620 
                                                                                
#define HIGH_USEC_BIT1		 3360 
#define MIN_HIGH_USEC_BIT1       2860
#define MAX_HIGH_USEC_BIT1       3860


#define KEY_CUSTOM_LO   	0x86
#define KEY_CUSTOM_HI   	0x79

#define KEY_CUSTOM_MASTER_LO   	0x80
#define KEY_CUSTOM_MASTER_HI   	0x7F
#define KEY_CODE_REPEAT         0xFF


static wait_queue_head_t	remocon_rwq;
static int			remocon_use_flag;

/* static int			last_triger_valid; */
static struct timeval		last_triger_time;
static int			last_triger_level;

static unsigned long		remocon_buffer;
static unsigned long		remocon_head;
static unsigned long		remocon_tail;
#if 0
static void process_header_start(unsigned long usec, unsigned long hl);
static void process_header_1(unsigned long usec, unsigned long hl);
static void process_header_2(unsigned long usec, unsigned long hl);
static void process_header_3(unsigned long usec, unsigned long hl);
static void process_header_4(unsigned long usec, unsigned long hl);
static void process_header_5(unsigned long usec, unsigned long hl);
static void process_header_6(unsigned long usec, unsigned long hl);

#endif
static void process_repeat_2(unsigned long usec, unsigned long hl);
static void process_repeat_3(unsigned long usec, unsigned long hl);

static void process_frame_start(unsigned long usec, unsigned long hl);
static void process_frame_type(unsigned long usec, unsigned long hl);
static void process_bit_high(unsigned long usec, unsigned long hl);
static void process_bit_low(unsigned long usec, unsigned long hl);

static int nInByte;
static int nInBit;
static unsigned char    remocom_temp_buf[4];

static void (*remocon_proc_fptr)(unsigned long usec, unsigned long hl) =
						process_frame_start;


static inline void remocon_incr_bp(unsigned long buffer, volatile unsigned long *index, int delta)
{
    	unsigned long	next = *index + delta;
	barrier();
	*index = (next >= (buffer + PAGE_SIZE)) ? buffer : next;
}

static inline void put_data(unsigned char data)
{
	*((unsigned char *) remocon_head) = data;
	remocon_incr_bp(remocon_buffer, & remocon_head, sizeof(unsigned char));
    	wake_up_interruptible(& remocon_rwq);
}



static void
process_repeat_2(unsigned long usec, unsigned long hl)
{
	if (!hl || MIN_HIGH_USEC_REPEAT2>usec || MAX_HIGH_USEC_REPEAT2<usec) {
		TRACE("process_repeat_2(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_frame_start;
	} else {
		remocon_proc_fptr = process_repeat_3;
	}
}

static void
process_repeat_3(unsigned long usec, unsigned long hl)
{
	if (hl || MIN_LOW_USEC_REPEAT3 > usec || MAX_LOW_USEC_REPEAT3 < usec) {
		TRACE("process_repeat_3(%lu, %lu) : error\n", usec, hl);
	} else {
		put_data(KEY_CODE_REPEAT);
	}
	remocon_proc_fptr = process_frame_start;
}

#if 0
static void
process_header_start(unsigned long usec, unsigned long hl)
{
	if (hl)
		remocon_proc_fptr = process_header_1;
	else {
		TRACE("process_header_start(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	}
}

static void
process_header_1(unsigned long usec, unsigned long hl)
{
	if(!hl && MIN_LOW_USEC_REPEAT1 < usec && MAX_LOW_USEC_REPEAT1 > usec) {
		remocon_proc_fptr = process_repeat_2;
	} else if (hl || MIN_LOW_USEC_HEADER1 > usec ||
		         MAX_LOW_USEC_HEADER1 < usec)
	{
		TRACE("process_header_1(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_header_2;
	}
}

static void
process_repeat_2(unsigned long usec, unsigned long hl)
{
	if (!hl || MIN_HIGH_USEC_REPEAT2>usec || MAX_HIGH_USEC_REPEAT2<usec) {
		TRACE("process_repeat_2(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_repeat_3;
	}
}

static void
process_repeat_3(unsigned long usec, unsigned long hl)
{
	if (hl || MIN_LOW_USEC_REPEAT3 > usec || MAX_LOW_USEC_REPEAT3 < usec) {
		TRACE("process_repeat_3(%lu, %lu) : error\n", usec, hl);
	} else {
		put_data(KEY_CODE_REPEAT);
	}
	remocon_proc_fptr = process_header_start;
}

static void
process_header_2(unsigned long usec, unsigned long hl)
{
	if (!hl || MIN_HIGH_USEC_HEADER2>usec || MAX_HIGH_USEC_HEADER2<usec) {
		TRACE("process_header_2(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_header_3;
	}
}

static void
process_header_3(unsigned long usec, unsigned long hl)
{
	if (hl || MIN_LOW_USEC_HEADER3 > usec || MAX_LOW_USEC_HEADER3 < usec) {
		TRACE("process_header_3(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_header_4;
	}
}

static void
process_header_4(unsigned long usec, unsigned long hl)
{
	if (!hl || MIN_HIGH_USEC_HEADER4>usec || MAX_HIGH_USEC_HEADER4<usec) {
		TRACE("process_header_4(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_header_5;
	}
}

static void
process_header_5(unsigned long usec, unsigned long hl)
{
	if (hl || MIN_LOW_USEC_HEADER5 > usec || MAX_LOW_USEC_HEADER5 < usec) {
		TRACE("process_header_5(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_header_6;
	}
}

static void
process_header_6(unsigned long usec, unsigned long hl)
{
	if (!hl || MIN_HIGH_USEC_HEADER6>usec || MAX_HIGH_USEC_HEADER6<usec) {
		TRACE("process_header_6(%lu, %lu) : error\n", usec, hl);
		remocon_proc_fptr = process_header_start;
	} else {
		remocon_proc_fptr = process_frame_start;
	}
}

#endif	



static void
process_frame_start(unsigned long usec, unsigned long hl)
{
	if(!hl && MIN_LOW_USEC_REPEAT1 < usec && MAX_LOW_USEC_REPEAT1 > usec) {
		remocon_proc_fptr = process_repeat_2;
	}else if (!hl && MIN_LOW_USEC_START < usec && usec < MAX_LOW_USEC_START) {
		remocon_proc_fptr = process_frame_type;
	} else {
		remocon_proc_fptr = process_frame_start;
	}
}

static void
process_frame_type(unsigned long usec, unsigned long hl)
{
	if (!hl) {
		remocon_proc_fptr = process_frame_start;
		TRACE("1frame type error(%lu, %lu)\n", usec, hl);
	} else if (MIN_HIGH_USEC_LEAD < usec && usec < MAX_HIGH_USEC_LEAD) {
		nInByte = 0;
		nInBit = 0;
		memset(remocom_temp_buf, 0, sizeof(remocom_temp_buf));
		remocon_proc_fptr = process_bit_low;
	}else {
		remocon_proc_fptr = process_frame_start;
		TRACE("2frame type error(%lu, %lu)\n", hl, usec);
	}
}

static void
process_bit_low(unsigned long usec, unsigned long hl)
{
	if (hl || usec < MIN_LOW_USEC_CODE || usec > MAX_LOW_USEC_CODE) {
		remocon_proc_fptr = process_frame_start;
		TRACE("code error(%lu, %lu)\n", usec, hl);
		return;
	}

	if (nInByte == 4) {
		remocon_proc_fptr = process_frame_start;
		if (((remocom_temp_buf[0] == KEY_CUSTOM_LO &&
		      remocom_temp_buf[1] == KEY_CUSTOM_HI) ||
		     (remocom_temp_buf[0] == KEY_CUSTOM_MASTER_LO &&
		      remocom_temp_buf[1] == KEY_CUSTOM_MASTER_HI)) &&
		    (remocom_temp_buf[2] | remocom_temp_buf[3]) == 0xFF)
		{
			put_data(remocom_temp_buf[2]);
		} else {
			TRACE("data error(%02X %02X %02X %02X)\n",
				remocom_temp_buf[0], remocom_temp_buf[1],
				remocom_temp_buf[2], remocom_temp_buf[3]);
				remocon_proc_fptr = process_frame_start;
		}
	} else {
		remocon_proc_fptr = process_bit_high;
	}
}

static void
process_bit_high(unsigned long usec, unsigned long hl)
{
	if (!hl) {
		remocon_proc_fptr = process_frame_start;
		TRACE("1bit error(%lu, %lu)\n", hl, usec);
		return;
	}

	if (MIN_HIGH_USEC_BIT0 < usec && usec < MAX_HIGH_USEC_BIT0) {
	} else if (MIN_HIGH_USEC_BIT1 < usec && usec < MAX_HIGH_USEC_BIT1) {
		remocom_temp_buf[nInByte] |= 0x01 << nInBit;
	} else {
		remocon_proc_fptr = process_frame_start;
		TRACE("2bit error(%lu, %lu)\n", hl, usec);
		return;
	}

	nInBit++;
	if (nInBit >= 8) {
		nInByte++;
		nInBit = 0;
	}

	remocon_proc_fptr = process_bit_low;
}

irqreturn_t remocon_interrupt(int irq, void *dev_id)
{
	struct timeval	tv;
	unsigned long	usec;
	unsigned long	hl;
	hl = s3c2410_gpio_getpin(S3C2410_GPF0);
	do_gettimeofday(&tv);
	if (tv.tv_sec - last_triger_time.tv_sec > 2100) {
	    	/* period > 35 min */
	    	usec = 0x7FFFFFFF; /* invalid */
	} else {
		usec = (tv.tv_sec - last_triger_time.tv_sec) * 1000000 +
		    	(tv.tv_usec - last_triger_time.tv_usec);
	}
	if (hl & last_triger_level) {
		TRACE("remocon_interrupt [%lu, %d] : error\n",hl,last_triger_level);
		remocon_proc_fptr = process_frame_start;//edit
	} else {
		(*remocon_proc_fptr)(usec, last_triger_level);
	}
	last_triger_time = tv;
	last_triger_level = hl;
	return IRQ_HANDLED;
}

unsigned int remocon_poll(struct file *filp, poll_table *wait)
{
    	unsigned int	mask = 0;

    	poll_wait(filp, & remocon_rwq, wait);
	if (remocon_head != remocon_tail) mask |= POLLIN | POLLRDNORM;
	return mask;
}

static ssize_t remocon_read(struct file *filp, char *buf, size_t count,
				loff_t *pos)
{
        int	count0;
    	while (remocon_head == remocon_tail) {
	    	if (filp->f_flags & O_NONBLOCK) return -EAGAIN;
    		interruptible_sleep_on(& remocon_rwq);
		if (signal_pending(current)) return -ERESTARTSYS;
	}
	count0 = remocon_head - remocon_tail;
	if (count0 < 0) count0 = remocon_buffer + PAGE_SIZE - remocon_tail;
	if (count0 < count) count = count0;

	if (copy_to_user(buf, (char *) remocon_tail, count)) return -EFAULT;
	remocon_incr_bp(remocon_buffer, & remocon_tail, count);

	return count;
}

static int remocon_open(struct inode *inode, struct file *filp)
{
        int	ret;

	if (remocon_use_flag) return -EBUSY;

	if (filp->f_mode & FMODE_WRITE) {
		return -EINVAL;
	}

	/* read mode file operation */
	init_waitqueue_head(&remocon_rwq);
	remocon_buffer = (unsigned long) kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (! remocon_buffer) {
	    	ret = -ENOMEM;
		goto nomem_error;
	}
	remocon_head = remocon_tail = remocon_buffer;

	/* last_triger_valid = 0; */
	remocon_proc_fptr = process_frame_start;//edit

	
	s3c2410_gpio_pullup(S3C2410_GPF0, S3C24XX_GPIO_PULL_UP);
	s3c2410_gpio_cfgpin(S3C2410_GPF0, S3C2410_GPF0_EINT0);
	set_irq_type(ASI_REMOCON_IRQ,IRQ_TYPE_EDGE_BOTH);

	last_triger_level = 1;
	ret = request_irq(ASI_REMOCON_IRQ, remocon_interrupt, IRQF_DISABLED,
			  "remocon", NULL);
	if (ret) {
	    	printk (KERN_ERR "remocon : IRQ %d already in use.\n",
			ASI_REMOCON_IRQ);
		goto irq_error;
	}
	remocon_use_flag = 1;

	return 0;

irq_error:
	kfree((void *) remocon_buffer);

nomem_error:
	return ret;
}

static int remocon_release(struct inode *inode, struct file *filp)
{
        if (remocon_use_flag && (filp->f_mode & FMODE_READ)) {
		kfree((void *) remocon_buffer);
	    	free_irq(ASI_REMOCON_IRQ, NULL);
	}
        remocon_use_flag = 0;
	return 0;
}


static struct file_operations remocon_fops = {
    	read:		remocon_read,
	poll:		remocon_poll,
	open:		remocon_open,
	release:	remocon_release,
};

static int __devinit yup110r_remocon_probe(struct platform_device *pdev)
{
    	int	result;
	result = register_chrdev(ASI_REMOCON_MAJOR, "remocon",&remocon_fops);
    	if (result < 0) {
	    	printk(KERN_WARNING "remocon : unagle to get major %d\n",
			ASI_REMOCON_MAJOR);
		return result;
	}
	printk(KERN_INFO "remocon : initialized (irq - %d)\n",
			ASI_REMOCON_IRQ);

	return 0;
}

static int __devexit yup110r_remocon_remove(struct platform_device *pdev)
{
    	unregister_chrdev(ASI_REMOCON_MAJOR, "remocon");
	return 0;
}

static struct platform_driver yup110r_remocon_driver = {
	.driver = {
		.name   = "yup110r-remocon",
		.owner  = THIS_MODULE,
	},
	.probe = yup110r_remocon_probe,
	.remove = __devexit_p(yup110r_remocon_remove),
};

static int __init remocon_init(void)
{
	return platform_driver_register(& yup110r_remocon_driver);
}

static void __exit remocon_exit(void)
{
	platform_driver_unregister(& yup110r_remocon_driver);
}

module_init(remocon_init);
module_exit(remocon_exit);

MODULE_AUTHOR("ASITEQ Corp.");
MODULE_LICENSE("GPL");
