#include <linux/module.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/i2c/ami602.h>

//#define DEBUG_TRIGGER_MODE
#undef DEBUG_TRIGGER_MODE
#ifdef DEBUG_TRIGGER_MODE
#include <mach/hardware.h>
#include <mach/regs-gpio.h>
#endif

/*
 * read command
 */
#define GET_COSR			0x11
#define GET_FINR			0x12
#define GET_MES_SUSPEND			0x28
#define GET_MES				0x14
#define GET_DAT				0x21
#define GET_FIRMWARE			0x17
#define GET_MES_PED_AUTO_SUSPEND	0x32
#define GET_MES_AVG			0x33
#define GET_GAIN			0x10
#define GET_PED_TH			0x36
#define GET_MES_T			0x15

/*
 * write command
 */
#define REQ_MES				0x55
#define SET_PWR_DOWN			0x57
#define SET_SUSPEND			0x75
#define SET_COSR			0x5b
#define SET_FINR			0x5c
#define SET_INDEX_DAT			0x63
#define SET_AEN				0x74
#define SET_MES_6CH_AUTO_START		0x64
#define SET_MES_6CH_AUTO_STOP		0x65
#define SET_MES_PED_AUTO_START		0x77
#define SET_MES_PED_AUTO_STOP		0x78
#define CLR_PED_SUSPEND			0x79
#define CHG_MES_AVG			0x80
#define SET_GAIN			0x61
#define SET_PED_TH			0x83

/*
 * sensor triggered measurement cycle
 */
#define SENSOR_CYCLE_20MS		0	/* 20ms */
#define SENSOR_CYCLE_40MS		1	/* 40ms */
#define SENSOR_CYCLE_60MS		2	/* 60ms */
#define SENSOR_CYCLE_80MS		3	/* 80ms */
#define SENSOR_CYCLE_100MS		4	/* 100ms */

#define MAX_12BIT                       ((1 << 12) - 1)

static u16 cmd_delay_tbl[][2] = {
	[SET_MES_PED_AUTO_START]	= {350, 450},
	[SET_MES_PED_AUTO_STOP]		= {350, 450},
	[GET_MES_PED_AUTO_SUSPEND]	= {200,  30},
	[CLR_PED_SUSPEND]		= {350,  30},
	[SET_MES_6CH_AUTO_START]	= {350, 450},
	[SET_MES_6CH_AUTO_STOP]		= {350, 450},
	[GET_MES]			= {200,  30},
	[GET_MES_SUSPEND]		= {200,  30},
	[GET_COSR]			= {200,  30},
	[GET_FINR]			= {200,  30},
	[GET_DAT]			= {200,  30},
	[GET_FIRMWARE]			= {200,  30},
	[GET_MES_AVG]			= {200,  30},
	[GET_GAIN]			= {200,  30},
	[GET_PED_TH]			= {550,  30},
	[GET_MES_T]			= {450,  30},
	[REQ_MES]			= {350,   0}, 
	[SET_PWR_DOWN]			= {350,  30},
	[SET_SUSPEND]			= {350,  30},
	[SET_COSR]			= {350,  30},
	[SET_FINR]			= {350,  30},
	[SET_INDEX_DAT]			= {350,  30},
	[SET_AEN]			= {350,  30},
	[CHG_MES_AVG]			= {350,  30},
	[SET_GAIN]			= {350,  30},
	[SET_PED_TH]			= {500,  30},
};

struct ami602 {
	struct input_dev	*input;

	struct i2c_client	*client;

	int			irq;
	struct workqueue_struct	*workqueue;
	struct work_struct	sensor_work;

	void (*set_trg)(int level);
	int (*get_busy)(void);
	void (*udelay)(unsigned int usec);

	/* status */
#define SENSOR_TRIGGER_PED	(1 << 1)
#define SENSOR_TRIGGER_6CH	(1 << 0)
	u8			trigger_mode;
	u8			trigger_cycle;
	u8			req_stop_trigger;
	struct completion	done;

	/* register shadow */
	u8			firmware[7];
	u16			origin[6];
	u16			sensitivity[6];
	u8			offset_coarse[6];
	u8			offset_fine[6];
	u8			gain[6];
	u8			ped_th[17];

	/* data buffer */
	u8			buf[32];

	/* for power management */
	u8			req_resume;
	u8			suspend_trigger_mode;
	u8			suspend_trigger_cycle;

	/* last fetched channel value */
	u16			chval[6];
	/* last fetched pedometer value */
	u32			step_num;
	u32			walk_time;
	u8			pedo_status;

	struct mutex		mutex;
};

static int ami602_conv_result(u8 result)
{
	switch (result) {
	case 0:	/* success */
		return 0;
	case 0xF7: /* pedometer not operation */
		printk("AMI602 : pedometer not operation\n");
		return -EPERM;
	case 0xFA: /* unusable command at Sensor Trigger Mode */
		printk("AMI602 : unusable command at Sensor Trigger Mode\n");
		return -EBUSY;
	case 0xFE: /* wrong packet */
		printk("AMI602 : wrong packet\n");
		return -EINVAL;
	default:
		printk("AMI602 : error code -> %02X\n", result);
		break;
	}
	return -EINVAL;
}

static int ami602_get_busy(struct ami602 *am)
{
	return (*(am->get_busy))();
}

static void ami602_udelay(struct ami602 *am, unsigned int usec)
{
	(*(am->udelay))(usec);
}

static void ami602_wakeup(struct ami602 *am)
{
	/* suspend -> active : 20us delay */
	(*(am->set_trg))(0);
	ami602_udelay(am, 10);
	(*(am->set_trg))(1);
	ami602_udelay(am, 15);
}

static int ami602_read(struct ami602 *am, u8 cmd, u8 *data, int ndata)
{
	int	rv;
	int	err;

	ami602_wakeup(am);

	/* cmd1 */
	rv = i2c_master_send(am->client, &cmd, 1);
	if (rv < 1) {
		printk(KERN_ERR "AMI602 : error on i2c_master_send(%d) at cmd(%02x)\n", rv, cmd);
		return -EIO;
	}

	ami602_udelay(am, cmd_delay_tbl[cmd][0] * 11 / 10);

	/* cmd2 */
	rv = i2c_master_recv(am->client, am->buf, ndata + 1);
	if (rv < 1) {
		printk(KERN_ERR "AMI602 : error on i2c_master_recv(%d) at cmd(%02x)\n", rv, cmd);
		return -EIO;
	}
	if ((err = ami602_conv_result(am->buf[0])) < 0)
		return err;
	if (rv < (ndata+1)) {
		printk(KERN_ERR "AMI602 : short receive at cmd(%02x)\n", cmd);
		return -EIO;
	}

	memcpy(data, am->buf+1, rv-1);

	return 0;
}

static int ami602_write(struct ami602 *am, u8 cmd, const u8 *data, int ndata)
{
	int	rv;

	ami602_wakeup(am);

	/* cmd1 */
	am->buf[0] = cmd;
	memcpy(am->buf+1, data, ndata);
	rv = i2c_master_send(am->client, am->buf, ndata+1); 
	if (rv < (ndata+1)) {
		printk(KERN_ERR "AMI602 : error on i2c_master_send(%d) at cmd(%02x)\n", rv, cmd);
		return -EIO;
	}

	ami602_udelay(am, cmd_delay_tbl[cmd][0] * 11 / 10);

	/* cmd2 */
	rv = i2c_master_recv(am->client, am->buf, 1);
	if (rv < 1) {
		printk(KERN_ERR "AMI602 : error on i2c_master_recv(%d) at cmd(%02x)\n", rv, cmd);
		return -EIO;
	}

	return ami602_conv_result(am->buf[0]);
}

static int ami602_read_delay(struct ami602 *am, u8 cmd, u8 *data, int ndata)
{
	int	rv;
	rv = ami602_read(am, cmd, data, ndata);
	ami602_udelay(am, cmd_delay_tbl[cmd][1] * 11 / 10);
	return rv;
}

static int ami602_write_delay(struct ami602 *am, u8 cmd, const u8 *data, int ndata)
{
	int	rv;
	rv = ami602_write(am, cmd, data, ndata);
	ami602_udelay(am, cmd_delay_tbl[cmd][1] * 11 / 10);
	return rv;
}

static int ami602_get_dat_at_index(struct ami602 *am, u8 index, u8 data[6])
{
	int	rv;
	rv = ami602_write_delay(am, SET_INDEX_DAT, &index, 1);
	if (rv)
		return rv;
	rv = ami602_read_delay(am, GET_DAT, data, 6);
	if (rv)
		return rv;
	return rv;
}

static int ami602_get_ch_origin_sensitivity(struct ami602 *am)
{
	int	rv;
	int	i;
	u8	data[5][6];

	for (i = 0; i < 5; i++) {
		rv = ami602_get_dat_at_index(am, i, data[i]);
		if (rv)
			return rv;
	}

	/* CH1,2,3 origin */
	for (i = 0; i < 3; i++)
		am->origin[i] = ((u16)data[0][2*i] << 8) | data[0][2*i+1];
	/* CH4,5,6 origin */
	for (i = 0; i < 3; i++)
		am->origin[i+3] = ((u16)data[1][2*i] << 8) | data[1][2*i+1];

	/* CH1,2 sensitivity */
	for (i = 0; i < 2; i++)
		am->sensitivity[i] = ((u16)data[2][2*i] << 8) | data[2][2*i+1];
	/* CH6 sensitivity */
	am->sensitivity[2] = ((u16)data[3][4] << 8) | data[3][5];
	/* CH3,4,5 sensitivity */
	for (i = 0; i < 3; i++)
		am->sensitivity[i+3] = ((u16)data[4][2*i] << 8) | data[4][2*i+1];

	return rv;
}

static int ami602_startup(struct ami602 *am)
{
	int rv;

	/* power down mode -> boot mode */
	ami602_wakeup(am);

	/* wait rising edge of busy (boot mode -> suspend mode) */
	do {
		msleep(1000/HZ);
	} while (! ami602_get_busy(am)) ;

	/* SET_COSR */
	rv = ami602_write_delay(am, SET_COSR, am->offset_coarse, 6);
	if (rv) {
		printk("SET_COSR -> error\n");
		return rv;
	}

	/* SET_FIN */
	rv = ami602_write_delay(am, SET_FINR, am->offset_fine, 6);
	if (rv)
		return rv;

	/* SET_PED_TH */
	rv = ami602_write_delay(am, SET_PED_TH, am->ped_th, 17);
	if (rv)
		return rv;

	/* SET_SUSPEND */
	rv = ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2);
	if (rv)
		return rv;

	return 0;
}

static void ami602_try_stop_sensor_trigger(struct ami602 *am)
{
	int	rv, i;

	for (i = 0; i < 10; i++) {
		rv = ami602_write_delay(am, SET_MES_6CH_AUTO_STOP, "ed", 2);
		if (rv == 0)
			break;
	}
	for (i = 0; i < 10; i++) {
		rv = ami602_write_delay(am, SET_MES_PED_AUTO_STOP, "ed", 2);
		if (rv == 0)
			break;
	}
	for (i = 0; i < 10; i++) {
		/* SET_SUSPEND */
		rv = ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2);
		if (rv == 0)
			break;
	}
}

static int ami602_powerdown(struct ami602 *am)
{
	int rv;

	/* SET_PWR_DOWN */
	rv = ami602_write_delay(am, SET_PWR_DOWN, (const u8 *) "pd", 2);
	if (rv)
		return rv;

	return 0;
}

static int ami602_get_initial_params(struct ami602 *am)
{
	int rv;

	rv = ami602_powerdown(am);
	if (rv)
		return rv;

	msleep(10);

	/* power down mode -> boot mode */
	ami602_wakeup(am);

	/* wait rising edge of busy (boot mode -> suspend mode) */
	do {
		msleep(1000/HZ);
	} while (! ami602_get_busy(am)) ;

	rv = ami602_read_delay(am, GET_FIRMWARE, am->firmware, 7);
	if (rv)
		return rv;
	rv = ami602_get_ch_origin_sensitivity(am);
	if (rv)
		return rv;
	rv = ami602_read_delay(am, GET_COSR, am->offset_coarse, 6);
	if (rv)
		return rv;
	rv = ami602_read_delay(am, GET_FINR, am->offset_fine, 6);
	if (rv)
		return rv;
	rv = ami602_read_delay(am, GET_GAIN, am->gain, 6);
	if (rv)
		return rv;
	rv = ami602_read_delay(am, GET_PED_TH, am->ped_th, 17);
	if (rv)
		return rv;

	/* SET_SUSPEND */
	rv = ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2);
	if (rv)
		return rv;

	return 0;
}

static void ami602_put_6ch_data(struct ami602 *am, const u8 data[9])
{
	int	i;
	for (i = 0; i < 3; i++, data += 3) {
		am->chval[i*2] = ((u16)data[0] << 4) | (data[1] >> 4);
		am->chval[i*2+1] = ((u16)(data[1] & 0xF) << 8) | data[2];
	}

	/* magnetometer */
	input_report_abs(am->input, ABS_Z, am->chval[0]);
	input_report_abs(am->input, ABS_X, am->chval[1]);
	input_report_abs(am->input, ABS_Y, am->chval[5]);

	/* accelerometer */
	input_report_abs(am->input, ABS_RZ, am->chval[2]);
	input_report_abs(am->input, ABS_RX, am->chval[3]);
	input_report_abs(am->input, ABS_RY, am->chval[4]);
}

static int ami602_get_mes_suspend(struct ami602 *am)
{
	int rv = 0;
	u8 data[9];

	rv = ami602_read(am, GET_MES_SUSPEND, data, 9);
	if (rv == 0)
		ami602_put_6ch_data(am, data);
	ami602_udelay(am, cmd_delay_tbl[GET_MES_SUSPEND][1] * 11 / 10);
	return rv;
}

static int ami602_set_aen(struct ami602 *am, u8 enable)
{
	return ami602_write_delay(am, SET_AEN, &enable, 1);
}

static int ami602_host_trigger(struct ami602 *am)
{
	int	rv;

	if ((rv = ami602_set_aen(am, 1)))
		return rv;

	if ((rv = ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2)))
		goto rtn;

	if ((rv = ami602_write_delay(am, REQ_MES, NULL, 0)))
		goto rtn;

	do {
		ami602_udelay(am, 10);
	} while (! ami602_get_busy(am)) ;

	rv = ami602_get_mes_suspend(am);

rtn:
	while (ami602_set_aen(am, 0))
		;
	while (ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2))
		;

	return rv;
}

static void ami602_put_ped_data(struct ami602 *am, const u8 data[9])
{
	am->step_num = ((u32)data[0] << 24);
	am->step_num |= ((u32)data[1] << 16);
	am->step_num |= ((u32)data[2] << 8);
	am->step_num |= data[3];
	am->walk_time = ((u32)data[4] << 24);
	am->walk_time |= ((u32)data[5] << 16);
	am->walk_time |= ((u32)data[6] << 8);
	am->walk_time |= data[7];
	am->pedo_status = data[8];

	input_report_abs(am->input, ABS_MISC+0, (int) am->step_num);
	input_report_abs(am->input, ABS_MISC+1, (int) am->walk_time);
	input_report_abs(am->input, ABS_MISC+2, am->pedo_status);
}

static int ami602_get_mes_ped_auto_suspend(struct ami602 *am)
{
	int rv = 0;
	u8 data[9];
	rv = ami602_read(am, GET_MES_PED_AUTO_SUSPEND, data, 9);
	if (rv == 0)
		ami602_put_ped_data(am, data);
	ami602_udelay(am, cmd_delay_tbl[GET_MES_PED_AUTO_SUSPEND][1] * 11 / 10);
	return rv;
}

static int ami602_set_sensor_trigger_mode_no_stop(struct ami602 *am, u8 trigger_mode, u8 cycle)
{
	int rv = 0;

	/*
         * start new sensor trigger mode
         */
	if (trigger_mode & SENSOR_TRIGGER_6CH) {
		u8 mes_6ch_auto_start_data[3];
		mes_6ch_auto_start_data[0] = cycle;
		mes_6ch_auto_start_data[1] = 's';
		mes_6ch_auto_start_data[2] = 't';
		rv = ami602_write_delay(am, SET_MES_6CH_AUTO_START, mes_6ch_auto_start_data, 3);
		if (rv)
			goto rtn;
		am->trigger_mode |= SENSOR_TRIGGER_6CH;
		am->trigger_cycle = cycle;
	}
	if (trigger_mode & SENSOR_TRIGGER_PED) {
		rv = ami602_write_delay(am, SET_MES_PED_AUTO_START, "st", 2);
		if (rv)
			goto rtn;
		am->trigger_mode |= SENSOR_TRIGGER_PED;
	}

rtn:
	return rv;
}

static int ami602_set_sensor_trigger_mode(struct ami602 *am, u8 trigger_mode, u8 cycle)
{
	/*
         * stop previous sensor trigger mode
         */
	while (am->trigger_mode) {
		am->req_stop_trigger = 1;
		wait_for_completion(& am->done);
	}
	am->req_stop_trigger = 0;

	return ami602_set_sensor_trigger_mode_no_stop(am, trigger_mode, cycle);
}

static void ami602_irq_worker(struct work_struct *work)
{
	struct ami602 *am = container_of(work, struct ami602, sensor_work);

	mutex_lock(&am->mutex);

	if (am->req_resume) {
		int rv;
retry:
		msleep(1000 / HZ);

		rv = ami602_startup(am);
		if (rv) {
			printk(KERN_ERR "AMI602 : error on ami602_startup\n");
			goto retry;
		}

		rv = ami602_set_sensor_trigger_mode_no_stop(am, am->suspend_trigger_mode, am->suspend_trigger_cycle);
		if (rv) {
			printk(KERN_ERR "AMI602 : error on ami602_set_sensor_trigger_mode\n");
			goto retry;
		}

		am->req_resume = 0;

		mutex_unlock(&am->mutex);
		return;
	}

	if (! am->trigger_mode) {
		if (am->req_stop_trigger) {
			am->req_stop_trigger = 0;
			complete(& am->done);
		}

		mutex_unlock(&am->mutex);
		return;
	}

	if (am->req_stop_trigger) {
		int	rv;
		if (am->trigger_mode & SENSOR_TRIGGER_6CH) {
			rv = ami602_write_delay(am, SET_MES_6CH_AUTO_STOP, "ed", 2);
			if (rv == 0)
				am->trigger_mode &= ~SENSOR_TRIGGER_6CH;
		}
		if (!(am->trigger_mode & SENSOR_TRIGGER_6CH) &&
		    am->trigger_mode & SENSOR_TRIGGER_PED)
		{
			rv = ami602_write_delay(am, SET_MES_PED_AUTO_STOP, "ed", 2);
			if (rv == 0)
				am->trigger_mode &= ~SENSOR_TRIGGER_PED;
		}

		if (! am->trigger_mode) {
			/* SET_SUSPEND */
			rv = ami602_write_delay(am, SET_SUSPEND, (const u8 *) "sp", 2);
			if (rv == 0) {
				am->req_stop_trigger = 0;
				complete(& am->done);
			}
		}
		mutex_unlock(&am->mutex);
		return;
	}

#ifdef DEBUG_TRIGGER_MODE
	s3c2410_gpio_setpin(S3C2443_GPH14, 1);
#endif
	if (am->trigger_mode & SENSOR_TRIGGER_6CH)
		ami602_get_mes_suspend(am);
	if (am->trigger_mode & SENSOR_TRIGGER_PED)
		ami602_get_mes_ped_auto_suspend(am);
	input_sync(am->input);
#ifdef DEBUG_TRIGGER_MODE
	s3c2410_gpio_setpin(S3C2443_GPH14, 0);
#endif

	mutex_unlock(&am->mutex);
}

static irqreturn_t ami602_irq(int irq, void *handle)
{
	struct ami602 *am = (struct ami602 *) handle;
	if (!work_pending(&am->sensor_work))
		queue_work(am->workqueue, &am->sensor_work);
	return IRQ_HANDLED;
}

static ssize_t show_firmware(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return snprintf(buf, PAGE_SIZE, "%d.%d(%02x%02x/%d/%d)\n",
			am->firmware[0], am->firmware[1],
			am->firmware[2], am->firmware[3],
			am->firmware[4], am->firmware[5]);
}

static ssize_t show_origin(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return snprintf(buf, PAGE_SIZE, "%d %d %d %d %d %d\n",
			am->origin[0], am->origin[1], am->origin[2],
			am->origin[3], am->origin[4], am->origin[5]);
}

static ssize_t show_sensitivity(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return snprintf(buf, PAGE_SIZE, "%d %d %d %d %d %d\n",
		am->sensitivity[0], am->sensitivity[1], am->sensitivity[2],
		am->sensitivity[3], am->sensitivity[4], am->sensitivity[5]);
}

static ssize_t show_ch_param(struct device *dev, char *buf, const u8 param[6])
{
	return snprintf(buf, PAGE_SIZE, "%d %d %d %d %d %d\n",
			param[0], param[1], param[2],
			param[3], param[4], param[5]);
}

static ssize_t store_ch_param(struct device *dev, const char *buf, size_t count,
			u8 cmd, u8 param[6])
{
	u8	trigger_mode;
	u8	trigger_cycle;
	u8	nparam[6];
	struct ami602 *am = dev_get_drvdata(dev);
	int	rv;
	if (sscanf(buf, "%hhu %hhu %hhu %hhu %hhu %hhu",
			nparam+0, nparam+1, nparam+2,
			nparam+3, nparam+4, nparam+5)
		!= 6)
	{
		return 0;
	}

	trigger_mode = am->trigger_mode;
	trigger_cycle = am->trigger_cycle;
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, 0, 0);

	mutex_lock(&am->mutex);
	rv = ami602_write_delay(am, cmd, nparam, 6);
	mutex_unlock(&am->mutex);
	if (rv) {
		count = 0;
		goto rtn;
	}
	memcpy(param, nparam, 6);
rtn:
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, trigger_mode, trigger_cycle);

	return count;
}

static ssize_t show_offset_coarse(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return show_ch_param(dev, buf, am->offset_coarse);
}

static ssize_t store_offset_coarse(struct device *dev,
			struct device_attribute *attr, const char *buf, size_t count)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return store_ch_param(dev, buf, count, SET_COSR, am->offset_coarse);
}

static ssize_t show_offset_fine(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return show_ch_param(dev, buf, am->offset_fine);
}

static ssize_t store_offset_fine(struct device *dev,
			struct device_attribute *attr, const char *buf, size_t count)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return store_ch_param(dev, buf, count, SET_FINR, am->offset_fine);
}

static ssize_t show_gain(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return show_ch_param(dev, buf, am->gain);
}

static ssize_t store_gain(struct device *dev,
			struct device_attribute *attr, const char *buf, size_t count)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return store_ch_param(dev, buf, count, SET_GAIN, am->gain);
}

static ssize_t show_ped_th(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	int rv;

	mutex_lock(&am->mutex);
	rv = snprintf(buf, PAGE_SIZE,
			"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
			am->ped_th[0], am->ped_th[1], am->ped_th[2],
			am->ped_th[3], am->ped_th[4], am->ped_th[5],
			am->ped_th[6], am->ped_th[7], am->ped_th[8],
			am->ped_th[9], am->ped_th[10], am->ped_th[11],
			am->ped_th[12], am->ped_th[13], am->ped_th[14],
			am->ped_th[15], am->ped_th[16]);
	mutex_unlock(&am->mutex);
	return rv;
}

static ssize_t store_ped_th(struct device *dev,
			struct device_attribute *attr, const char *buf, size_t count)
{
	u8	trigger_mode;
	u8	trigger_cycle;
	u8	nparam[17];
	int	rv;
	struct ami602 *am = dev_get_drvdata(dev);
	if (sscanf(buf, "%hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu",
			nparam+0, nparam+1, nparam+2, nparam+3, nparam+4,
			nparam+5, nparam+6, nparam+7, nparam+8, nparam+9,
			nparam+10, nparam+11, nparam+12, nparam+13, nparam+14,
			nparam+15, nparam+16)
		!= 17)
	{
		return 0;
	}
	trigger_mode = am->trigger_mode;
	trigger_cycle = am->trigger_cycle;
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, 0, 0);
	mutex_lock(&am->mutex);
	rv = ami602_write_delay(am, SET_PED_TH, nparam, 17);
	mutex_unlock(&am->mutex);
	if (rv) {
		count = 0;
		goto rtn;
	}
	memcpy(am->ped_th, nparam, 17);
rtn:
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, trigger_mode, trigger_cycle);
	return count;
}

static ssize_t show_trigger_mode(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	if ((am->trigger_mode & (SENSOR_TRIGGER_6CH|SENSOR_TRIGGER_PED)) ==
		(SENSOR_TRIGGER_6CH|SENSOR_TRIGGER_PED))
	{
		strcpy(buf, "both\n");
	} else if (am->trigger_mode & SENSOR_TRIGGER_6CH) {
		strcpy(buf, "6ch\n");
	} else if (am->trigger_mode & SENSOR_TRIGGER_PED) {
		strcpy(buf, "ped\n");
	} else {
		strcpy(buf, "stop\n");
	}
	return strlen(buf);
}

static ssize_t store_trigger_mode(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t count)
{
	struct ami602 *am = dev_get_drvdata(dev);
	u8 trigger_mode = 0;
	if (strncmp(buf, "both", 4) == 0)
		trigger_mode = SENSOR_TRIGGER_6CH|SENSOR_TRIGGER_PED;
	else if (strncmp(buf, "6ch", 3) == 0)
		trigger_mode = SENSOR_TRIGGER_6CH;
	else if (strncmp(buf, "ped", 3) == 0)
		trigger_mode = SENSOR_TRIGGER_PED;
	else if (strncmp(buf, "stop", 4) == 0)
		trigger_mode = 0;
	else
		return 0;

	if (ami602_set_sensor_trigger_mode(am, trigger_mode, am->trigger_cycle))
		return 0;
	return count;
}

static ssize_t show_trigger_cycle(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	return snprintf(buf, PAGE_SIZE, "%d\n", am->trigger_cycle);
}

static ssize_t store_trigger_cycle(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t count)
{
	struct ami602 *am = dev_get_drvdata(dev);
	u8 cycle;
	if (sscanf(buf, "%hhu", &cycle) != 1)
		return 0;
	if (cycle > 4)
		return 0;
	if (am->trigger_mode & SENSOR_TRIGGER_6CH) {
		if (ami602_set_sensor_trigger_mode(am, am->trigger_mode, cycle))
			return 0;
	} else
		am->trigger_cycle = cycle;
	return count;
}

static ssize_t show_6ch_value(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	int rv;
	mutex_lock(&am->mutex);
	if (! am->trigger_mode) {
		while (ami602_host_trigger(am))
			;
	}
	rv = snprintf(buf, PAGE_SIZE, "%d %d %d %d %d %d\n",
			am->chval[0], am->chval[1], am->chval[2],
			am->chval[3], am->chval[4], am->chval[5]);
	mutex_unlock(&am->mutex);
	return rv;
}

static ssize_t show_pedo_value(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	struct ami602 *am = dev_get_drvdata(dev);
	int rv;
	mutex_lock(&am->mutex);
	rv = snprintf(buf, PAGE_SIZE, "%u %u %d\n",
			am->step_num, am->walk_time, am->pedo_status);
	mutex_unlock(&am->mutex);
	return rv;
}

static ssize_t store_pedo_value(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t count)
{
	u8	trigger_mode;
	u8	trigger_cycle;
	int	rv;
	struct ami602 *am = dev_get_drvdata(dev);

	trigger_mode = am->trigger_mode;
	trigger_cycle = am->trigger_cycle;
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, 0, 0);
	mutex_lock(&am->mutex);
	rv = ami602_write_delay(am, CLR_PED_SUSPEND, NULL, 0);
	mutex_unlock(&am->mutex);
	if (rv)
		count = 0;
	if (trigger_mode)
		ami602_set_sensor_trigger_mode(am, trigger_mode, trigger_cycle);
	return count;
}

static DEVICE_ATTR(firmware,	S_IRUGO, show_firmware,		NULL);
static DEVICE_ATTR(origin,	S_IRUGO, show_origin,		NULL);
static DEVICE_ATTR(sensitivity,	S_IRUGO, show_sensitivity,	NULL);
static DEVICE_ATTR(offset_coarse, S_IRUGO|S_IWUSR|S_IWGRP, show_offset_coarse, store_offset_coarse);
static DEVICE_ATTR(offset_fine, S_IRUGO|S_IWUSR|S_IWGRP, show_offset_fine, store_offset_fine);
static DEVICE_ATTR(gain, S_IRUGO|S_IWUSR|S_IWGRP, show_gain, store_gain);
static DEVICE_ATTR(ped_th, S_IRUGO|S_IWUSR|S_IWGRP, show_ped_th, store_ped_th);
static DEVICE_ATTR(trigger_mode, S_IRUGO|S_IWUSR|S_IWGRP, show_trigger_mode, store_trigger_mode);
static DEVICE_ATTR(trigger_cycle, S_IRUGO|S_IWUSR|S_IWGRP, show_trigger_cycle, store_trigger_cycle);
static DEVICE_ATTR(6ch, S_IRUGO, show_6ch_value, NULL);
static DEVICE_ATTR(walk, S_IRUGO|S_IWUSR|S_IWGRP, show_pedo_value, store_pedo_value);

static struct attribute *ami602_attributes[] = {
	&dev_attr_firmware.attr,
	&dev_attr_origin.attr,
	&dev_attr_sensitivity.attr,
	&dev_attr_offset_coarse.attr,
	&dev_attr_offset_fine.attr,
	&dev_attr_gain.attr,
	&dev_attr_ped_th.attr,
	&dev_attr_trigger_mode.attr,
	&dev_attr_trigger_cycle.attr,
	&dev_attr_6ch.attr,
	&dev_attr_walk.attr,
	NULL,
};

static struct attribute_group ami602_attr_group = {
	.attrs = ami602_attributes,
};

static int ami602_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct ami602 *am;
	struct ami602_platform_data *pdata = client->dev.platform_data;
	struct input_dev *input_dev;
	int err;

	if (!pdata) {
		dev_err(&client->dev, "platform data is required!\n");
		return -EINVAL;
	}

	am = kzalloc(sizeof(struct ami602), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!am || !input_dev) {
		err = -ENOMEM;
		goto err_free_mem;
	}

	mutex_init(& am->mutex);
	init_completion(& am->done);

	am->client = client;
	i2c_set_clientdata(client, am);

	am->input = input_dev;

	am->set_trg = pdata->set_trg;
	am->get_busy = pdata->get_busy;
	am->udelay = pdata->udelay;
	pdata->init_platform_hw();

	input_dev->name = "AMI602 Motion Sensor";
	input_dev->id.bustype = BUS_I2C;
	input_dev->evbit[0] = BIT_MASK(EV_ABS);

	input_set_abs_params(input_dev, ABS_X, 0, MAX_12BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_Y, 0, MAX_12BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_Z, 0, MAX_12BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_RX, 0, MAX_12BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_RY, 0, MAX_12BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_RZ, 0, MAX_12BIT, 0, 0);

	input_set_abs_params(input_dev, ABS_MISC + 0, INT_MIN, INT_MAX, 0, 0);
	input_set_abs_params(input_dev, ABS_MISC + 1, INT_MIN, INT_MAX, 0, 0);
	input_set_abs_params(input_dev, ABS_MISC + 2, 0, 2, 0, 0);

#if 0
	am->workqueue = create_singlethread_workqueue(input_dev->name);
#else
	am->workqueue = create_rt_workqueue(input_dev->name);
#endif
	if (am->workqueue == NULL) {
		err = -EBUSY;
		goto err_create_workqueue;
	}

	INIT_WORK(&am->sensor_work, ami602_irq_worker);

	am->irq = client->irq;
	set_irq_type(am->irq, pdata->irqflags);
	err = request_irq(am->irq, ami602_irq, 0, client->dev.driver->name, am);
	if (err < 0) {
		dev_err(&client->dev, "irq %d busy?\n", am->irq);
		goto err_free_mem;
	}

	ami602_try_stop_sensor_trigger(am);

	err = ami602_get_initial_params(am);
	if (err)
		goto err_free_irq;

	err = input_register_device(input_dev);
	if (err)
		goto err_free_irq;

	sysfs_create_group(&client->dev.kobj, &ami602_attr_group);

	dev_info(&client->dev, "registered with irq (%d)\n", am->irq);

	return 0;

err_free_irq:
	free_irq(am->irq, am);
err_free_mem:
	destroy_workqueue(am->workqueue);
err_create_workqueue:
	input_free_device(input_dev);
	kfree(am);
	return err;
}

static int ami602_remove(struct i2c_client *client)
{
	struct ami602 *am = i2c_get_clientdata(client);
	struct ami602_platform_data *pdata = client->dev.platform_data;
	int rv;

	sysfs_remove_group(&client->dev.kobj, &ami602_attr_group);

	rv = ami602_set_sensor_trigger_mode(am, 0, 0);
	if (rv)
		return rv;

#if 0
	ami602_powerdown(am);
#endif

	pdata->exit_platform_hw();

	free_irq(am->irq, am);
	destroy_workqueue(am->workqueue);
	input_unregister_device(am->input);
	kfree(am);

	return 0;
}

#ifdef CONFIG_PM
static int ami602_suspend(struct i2c_client *client, pm_message_t state)
{
	struct ami602 *am = i2c_get_clientdata(client);
	struct ami602_platform_data *pdata = client->dev.platform_data;
	int rv;

	am->suspend_trigger_mode = am->trigger_mode;
	am->suspend_trigger_cycle = am->trigger_cycle;

	if (am->trigger_mode & SENSOR_TRIGGER_PED)
		return 0;

	rv = ami602_set_sensor_trigger_mode(am, 0, 0);
	if (rv)
		return rv;
	mutex_lock(&am->mutex);
	rv = ami602_powerdown(am);
	mutex_unlock(&am->mutex);
	if (rv)
		return rv;

	pdata->exit_platform_hw();

	return 0;
}

static int ami602_resume(struct i2c_client *client)
{
	struct ami602 *am = i2c_get_clientdata(client);
	struct ami602_platform_data *pdata = client->dev.platform_data;

	if (am->trigger_mode)
		return 0;

	pdata->init_platform_hw();

	am->req_resume = 1;
	queue_work(am->workqueue, &am->sensor_work);

	return 0;
}
#else
#define ami602_suspend	NULL
#define ami602_resume	NULL
#endif

static struct i2c_device_id	ami602_idtable[] = {
	{"ami602", 0},
	{}
};
MODULE_DEVICE_TABLE(i2c, ami602_idtable);

static struct i2c_driver	ami602_driver = {
	.driver = {
		.owner  = THIS_MODULE,
		.name	= "ami602",
	},
	.id_table =	ami602_idtable,
	.probe	  =	ami602_probe,
	.remove	  =	ami602_remove,
	.suspend  =	ami602_suspend,
	.resume  =	ami602_resume,
};

static int __init ami602_init(void)
{
	return i2c_add_driver(&ami602_driver);
}

static void __exit ami602_exit(void)
{
	i2c_del_driver(&ami602_driver);
}

module_init(ami602_init);
module_exit(ami602_exit);

MODULE_DESCRIPTION("AMI602 motion sensor driver");
MODULE_LICENSE("GPL");
