/* YUP150 Buttons Input Driver
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/irq.h>

#include <mach/hardware.h>
#include <mach/regs-gpio.h>

#define YUP150_BTN_MAX	2

struct yup150_input {
	int			btn_gpio[YUP150_BTN_MAX];
	int			btn_irq[YUP150_BTN_MAX];
	int			btn_map[YUP150_BTN_MAX];

	int			btn_pressed[YUP150_BTN_MAX];

	struct input_dev	*input_dev;
};

static int
yup150_btn_irq(int irq, void *data)
{
	int i;
	struct yup150_input *input = data;

	for (i = 0; i < YUP150_BTN_MAX; i++) {
		if (input->btn_irq[i] == irq)
			break;
	}
	if (i >= YUP150_BTN_MAX)
		return 0;

	if (s3c2410_gpio_getpin(input->btn_gpio[i])) {
		if (input->btn_pressed[i]) {
			input->btn_pressed[i] = 0;
			set_irq_type(irq, IRQ_TYPE_EDGE_FALLING);
			input_report_key(input->input_dev, input->btn_map[i], 0);
			input_sync(input->input_dev);
		}
	} else {
		if (! input->btn_pressed[i]) {
			input->btn_pressed[i] = 1;
			set_irq_type(irq, IRQ_TYPE_EDGE_RISING);
			input_report_key(input->input_dev, input->btn_map[i], 1);
			input_sync(input->input_dev);
		}
	}

	return IRQ_HANDLED;
}

static int __devinit yup150_input_probe(struct platform_device *pdev)
{
	struct yup150_input *input;
	struct input_dev *input_dev;
	int i;
	int ret;

	input = kzalloc(sizeof(*input), GFP_KERNEL);
	if (!input)
		return -ENOMEM;

	input_dev = input_allocate_device();
	if (!input_dev) {
		kfree(input);
		return -ENOMEM;
	}

	platform_set_drvdata(pdev, input);

	input->btn_gpio[0] = S3C2410_GPF0;
	input->btn_irq[0] = IRQ_EINT0;
	input->btn_map[0] = KEY_POWER;

	input->btn_gpio[1] = S3C2410_GPG1;
	input->btn_irq[1] = IRQ_EINT9;
	input->btn_map[1] = KEY_MENU;

	input->input_dev = input_dev;

	input_dev->name = "YUP150 button events";
	input_dev->id.bustype = BUS_HOST;
	input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_PWR);
	for (i = 0; i < YUP150_BTN_MAX; i++)
		set_bit(input->btn_map[i], input_dev->keybit);

	ret = input_register_device(input_dev);
	if (ret) {
		input_free_device(input_dev);
		kfree(input);
		return ret;
	}

	/* power button */
        s3c2410_gpio_pullup(S3C2410_GPF0, S3C24XX_GPIO_PULL_UP);
        s3c2410_gpio_cfgpin(S3C2410_GPF0, S3C2410_GPF0_EINT0);
        set_irq_type(input->btn_irq[0], IRQ_TYPE_EDGE_FALLING);
        ret = request_irq(input->btn_irq[0], yup150_btn_irq, 0,
			"yup150_pwr_btn", input);
	if (ret) {
		input_unregister_device(input_dev);
		input_free_device(input_dev);
		kfree(input);
		return ret;
	}

	/* menu button */
        s3c2410_gpio_pullup(S3C2410_GPG1, S3C24XX_GPIO_PULL_UP);
        s3c2410_gpio_cfgpin(S3C2410_GPG1, S3C2410_GPG1_EINT9);
        set_irq_type(input->btn_irq[1], IRQ_TYPE_EDGE_FALLING);
        ret = request_irq(input->btn_irq[1], yup150_btn_irq, 0,
			"yup150_menu_btn", input);
	if (ret) {
		free_irq(input->btn_irq[0], input);
		input_unregister_device(input_dev);
		input_free_device(input_dev);
		kfree(input);
		return ret;
	}

	return 0;
}

static int __devexit yup150_input_remove(struct platform_device *pdev)
{
	struct yup150_input *input  = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < YUP150_BTN_MAX; i++)
		free_irq(input->btn_irq[i], input);

	input_unregister_device(input->input_dev);
	input_free_device(input->input_dev);
	kfree(input);

	return 0;
}

#ifdef CONFIG_PM
static int yup150_input_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct yup150_input *input  = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < YUP150_BTN_MAX; i++) {
		if (input->btn_map[i] == KEY_POWER)
			enable_irq_wake(input->btn_irq[i]);
		set_irq_type(input->btn_irq[i], IRQ_TYPE_EDGE_FALLING);
		input->btn_pressed[i] = 0;
	}
	return 0;
}

static int yup150_input_resume(struct platform_device *pdev)
{
	struct yup150_input *input  = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < YUP150_BTN_MAX; i++) {
		if (input->btn_map[i] == KEY_POWER)
			disable_irq_wake(input->btn_irq[i]);
		set_irq_type(input->btn_irq[i], IRQ_TYPE_EDGE_FALLING);
		input->btn_pressed[i] = 0;
	}
	return 0;
}
#else
#define yup150_input_suspend	NULL
#define yup150_input_resume	NULL
#endif

static struct platform_driver yup150_input_driver = {
	.driver = {
		.name = "yup150-button",
		.owner  = THIS_MODULE,
	},
	.probe = yup150_input_probe,
	.remove = __devexit_p(yup150_input_remove),
	.suspend = yup150_input_suspend,
	.resume = yup150_input_resume,
};

static int __init yup150_input_init(void)
{
	return platform_driver_register(&yup150_input_driver);
}
module_init(yup150_input_init);

static void __exit yup150_input_exit(void)
{
	platform_driver_unregister(&yup150_input_driver);
}
module_exit(yup150_input_exit);

MODULE_DESCRIPTION("YUP150 button driver");
MODULE_LICENSE("GPL");
