/*
 *  linux/drivers/i2c/chips/ak4183.c
 *
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/workqueue.h>

//#include <asm/hardware.h>
//#include <asm/irq.h>

#include <mach/hardware.h>
#include <mach/regs-gpio.h>

//#define DEBUG
#undef DEBUG
#ifdef DEBUG
# define _DBG(fmt, args...) do{ printk(KERN_DEBUG"%s: " fmt "\n", __FUNCTION__, ##args); } while(0);
#else
# define _DBG(fmt, args...) do { } while(0);
#endif

/* command */
#define CMD_SLEEP_NO		0x80
#define CMD_ACCEL_OFF		0x40
#define CMD_CHAN_X		(0 << 4)
#define CMD_CHAN_Y		(1 << 4)
#define CMD_CHAN_Z1		(2 << 4)
#define CMD_CHAN_Z2		(3 << 4)
#define CMD_ADC_ON		0x04
#define CMD_MODE_8BIT		0x02

#if 0
enum touchscreen_state {
	STATE_WAIT_FOR_TOUCH,	/* Waiting for a PEN interrupt */
	STATE_SAMPLING		/* Actively sampling ADC */
};
#endif
struct ak4183_data {
	struct i2c_client	*client;

	/* lock */
	struct semaphore	lock;

	/* touch screen */
#ifdef CONFIG_INPUT
	struct timer_list	timer;
	struct input_dev	*input;
	struct work_struct	work;
//	enum touchscreen_state	state;
//	int			working;
	int			irq;
	int			nsample;
	int			oldx, oldy;
#endif
};
static unsigned long poll_sample_time = 20; /* Sample every 20 milliseconds */
#define X_AXIS_MAX		4095
#define X_AXIS_MIN		0
#define Y_AXIS_MAX		4095
#define Y_AXIS_MIN		0
#define PRESSURE_MAX    	65535
//#define PRESSURE_MAX    	4095
#define PRESSURE_MIN		0
#define PRESSURE_THRESHOLD	1

/*
 * AK4183 i2c IO
 */
static inline void
report_touch(struct ak4183_data *ak, int x, int y, int pressure, int pendown)
{
	if (pendown) {
		input_report_abs(ak->input, ABS_X, x);
		input_report_abs(ak->input, ABS_Y, y);
	}
	input_report_key(ak->input, BTN_TOUCH, pendown);
	input_report_abs(ak->input, ABS_PRESSURE, pressure);
	input_sync(ak->input);
}

static inline int
ak4183_read_axis(struct ak4183_data *ak, u8 chan)
{
	u8	command;
	u8	data[2];
	static const char	*chan_str[] = {
		"X", "Y", "Z1", "Z2",
	};

	command = CMD_SLEEP_NO | chan | CMD_ADC_ON;

	if (i2c_master_send(ak->client, &command, 1) != 1)
		printk("i2c_master_send in ak4183 : Error\n");
	if (i2c_master_recv(ak->client, data, 2) != 2)
		printk("i2c_master_recv in ak4183 : Error\n");

	return ((int)data[0] << 4) | (data[1] >> 4);
}

static inline void ak4183_powerdown(struct ak4183_data *ak)
{
	u8	command = CMD_SLEEP_NO | CMD_ACCEL_OFF;
	i2c_master_send(ak->client, &command, 1);
}

static inline void ak4183_adc_on(struct ak4183_data *ak)
{
	u8	command = CMD_SLEEP_NO | CMD_ADC_ON;
	i2c_master_send(ak->client, &command, 1);
}

/*
 * touch screen driver
 */
#ifdef CONFIG_INPUT
static void start_touch_read(struct ak4183_data *ak, int is_first)
{
	if (is_first) {
//		if (! ak->working) {
			schedule_work(&ak->work);
//		} else {
//			printk(KERN_ERR "unsyncronized ts_timer_callback\n");
//		}
	} else {
		unsigned long inc = (poll_sample_time * HZ) / 1000;
		if (inc < 1)
			inc = 1;
		mod_timer(&ak->timer, jiffies + inc);
	}
}

static void ts_timer_callback(unsigned long _ak)
{
	struct ak4183_data	*ak = (struct ak4183_data *) _ak;
//	if (! ak->working) {
		schedule_work(&ak->work);
//	} else {
//		printk(KERN_ERR "unsyncronized ts_timer_callback\n");
//	}
}

//#define N_TS_SAMPLE	4
#define N_TS_SAMPLE	1
#define swap_with_acsending(x1,x2) \
	do {if((x1)>(x2)){int t;t=(x1);(x1)=(x2);(x2)=t;}} while(0)
static void ts_work(struct work_struct *work)
{
	struct ak4183_data	*ak = container_of(work, struct ak4183_data, work);

	down(&ak->lock);

	if (s3c2410_gpio_getpin(S3C2410_GPF7)) {
		/* release */
		if (ak->nsample) {
			report_touch(ak, 0, 0, 0, 0);
		} else {
//			printk("ts_work : early touch up\n");
		}
		ak->nsample = 0;
		ak->oldx = ak->oldy = 0;
//		ak->state = STATE_WAIT_FOR_TOUCH;
	} else {
		/* press */
		int			x[N_TS_SAMPLE], y[N_TS_SAMPLE];
		int			z1, z2, rt;
		int			pressure[N_TS_SAMPLE];
		int			i, j;
		int			valid = 1;

//		ak->working = 1;

//		set_irq_type(ak->irq, IRQT_NOEDGE);
		ak4183_adc_on(ak);
		for (i = 0; i < N_TS_SAMPLE; i++) {
			x[i] = ak4183_read_axis(ak, CMD_CHAN_X);
			y[i] = ak4183_read_axis(ak, CMD_CHAN_Y);
			z1   = ak4183_read_axis(ak, CMD_CHAN_Z1);
			z2   = ak4183_read_axis(ak, CMD_CHAN_Z2);
			if (z1 == 0x00F || z2 == 0xFFF) {
				valid = 0;
				break;
			}
			if (z1 == 0) {
				pressure[i] = 0;
			} else {
				rt = x[i] * (z2 - z1) / z1;
				if (rt < 1) rt = 1;
				pressure[i] = PRESSURE_MAX / rt;
			}
#if 0
			printk("X,Y,Z1,Z2,P : %03X %03X %03X %03X %d\n",
				x[i], y[i], z1, z2, pressure[i]);
#endif
		}
		ak4183_powerdown(ak);
//		set_irq_type(ak->irq, IRQT_FALLING);

//		ak->working = 0;

#if 0
		if (!valid || pressure[0] < PRESSURE_THRESHOLD) {
			printk("ts_work : low pressure[z1=0x%x,z2=0x%x]\n",
				z1, z2);
		}
#endif
		if (valid) {
#if 1
			if (pressure[0] >= PRESSURE_THRESHOLD /* &&
			    (!ak->nsample || ak->oldx != x[0] || ak->oldy != y[0])*/)
			{
				report_touch(ak, x[0], y[0], pressure[0], 1);
				ak->oldx = x[0];
				ak->oldy = y[0];
				ak->nsample++;
			}
#else
			for (i = 0; i < N_TS_SAMPLE - 1; i++) {
				for (j = i+1; j < N_TS_SAMPLE; j++) {
					swap_with_acsending(x[i], x[j]);
					swap_with_acsending(y[i], y[j]);
					swap_with_acsending(pressure[i], pressure[j]);
				}
			}
			for (i = 2; i < N_TS_SAMPLE - 1; i++) {
				x[1] += x[i];
				y[1] += y[i];
				pressure[1] += pressure[i];
			}
			x[1] /= N_TS_SAMPLE - 2;
			y[1] /= N_TS_SAMPLE - 2;
			pressure[1] /= N_TS_SAMPLE - 2;

			if (pressure[1] >= PRESSURE_THRESHOLD &&
			    (!ak->nsample || ak->oldx != x[1] || ak->oldy != y[1]))
			{
				report_touch(ak, x[1], y[1], pressure[1], 1);
				ak->oldx = x[1];
				ak->oldy = y[1];
				ak->nsample++;
			}
#endif
		}
		start_touch_read(ak, 0);
	}

	up(&ak->lock);
}

static irqreturn_t pen_isr(int irq, void *_ak)
{
	struct ak4183_data	*ak = _ak;
//	if (ak->state == STATE_WAIT_FOR_TOUCH) {
//		ak->state = STATE_SAMPLING;
//		start_touch_read(ak, 1);
		start_touch_read(ak, 0);
//	}
	return IRQ_HANDLED;
}
#endif

/*
 * i2c client driver
 */
static int ak4183_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	int ret;
	struct ak4183_data *ak;

	ak = kmalloc(sizeof(struct ak4183_data), GFP_KERNEL);
	if (!ak) {
		ret = -ENOMEM;
		goto done;
	}
	memset(ak, 0, sizeof(struct ak4183_data));

	ak->client = client;
	i2c_set_clientdata(client, ak);

	/* initialize */
	ak4183_powerdown(ak);

	init_MUTEX(&ak->lock);

#ifdef CONFIG_INPUT
	ak->input = input_allocate_device();
	if (! ak->input) {
		ret = -ENOMEM;
		goto done;
	}
	ak->input->name = "ak4183_ts";
//	ak->input->private = ak;
	ak->input->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS) | BIT(EV_SYN);
	ak->input->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);
	input_set_abs_params(ak->input, ABS_X, X_AXIS_MIN, X_AXIS_MAX, 0, 0);
	input_set_abs_params(ak->input, ABS_Y, Y_AXIS_MIN, Y_AXIS_MAX, 0, 0);
	input_set_abs_params(ak->input, ABS_PRESSURE, PRESSURE_MIN, PRESSURE_MAX, 0, 0);
	input_register_device(ak->input);

//	ak->working = 0;
	INIT_WORK(&ak->work, ts_work);

//	ak->state = STATE_WAIT_FOR_TOUCH;
	ak->timer.function = ts_timer_callback;
	ak->timer.data = (unsigned long) ak;
	init_timer(&ak->timer);

//	ak->irq = IRQ_EINT7;
	ak->irq = client->irq;
	set_irq_type(ak->irq, IRQ_TYPE_EDGE_FALLING);
	ret = request_irq(ak->irq, pen_isr, 0, "ak4183_ts", ak);
	if (ret)
		goto errirq;
#endif
	ak4183_powerdown(ak);

#ifdef CONFIG_INPUT
errirq:
	if (ret) {
		del_timer_sync(&ak->timer);
		input_unregister_device(ak->input);
	}
#endif

done:
	if (ret) {
		if (ak) {
			if (ak->input)
				input_free_device(ak->input);
			kfree(ak);
		}
	}
	return ret;
}

static int ak4183_remove(struct i2c_client *client)
{
	struct ak4183_data	*ak = i2c_get_clientdata(client);

#ifdef CONFIG_INPUT
	del_timer_sync(&ak->timer);
	flush_scheduled_work();
	free_irq(ak->irq, ak);
	input_unregister_device(ak->input);
#endif

	input_free_device(ak->input);
	kfree(ak);
	return 0;
}

static struct i2c_device_id	ak4183_idtable[] = {
	{"ak4183", 0},
	{}
};
MODULE_DEVICE_TABLE(i2c, ak4183_idtable);

static struct i2c_driver ak4183_driver = {
	.driver = {
		.owner		= THIS_MODULE,
		.name		= "ak4183",
	},
	.id_table	= ak4183_idtable,
	.probe		= ak4183_probe,
	.remove		= ak4183_remove,
};

static __init int ak4183_init(void)
{
	return i2c_add_driver(&ak4183_driver);
}

static __exit void ak4183_exit(void)
{
	i2c_del_driver(&ak4183_driver);
}

MODULE_DESCRIPTION("AK4183 Driver");
MODULE_LICENSE("GPL");

module_init(ak4183_init);
module_exit(ak4183_exit);
