/*
 *  linux/drivers/mmc/s3c-hsmmc.c - Samsung S3C24XX HS-MMC driver
 *
 * $Id: s3c-hsmmc.c,v 1.73 2008/04/13 23:56:19 jsgood Exp $
 *
 *  Copyright (C) 2006 Samsung Electronics, All Rights Reserved.
 *  by Suh, Seung-Chull<sc.suh@samsung.com>
 *
 *  This driver is made for High Speed MMC interface. This interface
 *  is adopted and implemented since s3c2443 was made.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Modified by Ryu,Euiyoul <steven.ryu@samsung.com>
 *  Modified by Suh, Seung-chull to support s3c6400
 *
 * Note: s3c-hsmmc.c can be used for 1 channel only at same time.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/highmem.h>
#include <linux/dma-mapping.h>
#include <linux/mmc/host.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/proc_fs.h>
#include <linux/irq.h>

#include <asm/dma.h>
#include <asm/dma-mapping.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/sizes.h>

#include <mach/regs-gpio.h>
#include <plat/hsmmc.h>
#include <plat/clock.h>

#include "regs-hsmmc.h"

#ifdef CONFIG_S3CMMC_DEBUG
#define DBG(x...)       printk(PFX x)
#else
#define DBG(x...)       do { } while (0)
#endif

#include "s3c-hsmmc.h"

#define DRIVER_NAME "s3c-hsmmc"
#define PFX DRIVER_NAME ": "

#define RESSIZE(ressource) (((ressource)->end - (ressource)->start)+1)

//#define USE_FINISH_TASKLET
#undef USE_FINISH_TASKLET
static void sdhci_tasklet_finish(unsigned long param);
static void s3c_hsmmc_change_clock(struct s3c_hsmmc_host *host, unsigned int clock);
static void s3c_hsmmc_send_register(struct s3c_hsmmc_host *host);
static int s3c_set_bus_width (struct s3c_hsmmc_host *host, uint width);

static int card_detect = 0;

static struct s3c_hsmmc_cfg s3c_hsmmc_platform = {
	.hwport = 0,
	.enabled = 0,
//	.host_caps = (MMC_CAP_4_BIT_DATA | MMC_CAP_MULTIWRITE),
	.host_caps = MMC_CAP_4_BIT_DATA,
	.base = NULL,
	.highspeed = 0,

	/* ctrl for mmc */
	.fd_ctrl[0] = {
		.ctrl2    = 0xC0004100,	/* ctrl2 for mmc */
		.ctrl3[0] = 0x80808080,	/* ctrl3 for low speed */
		.ctrl3[1] = 0x00000080,	/* ctrl3 for high speed */
		.ctrl4 = 0,
	},
	
	/* ctrl for sd */
	.fd_ctrl[1] = {
		.ctrl2    = 0xC0000100,	/* ctrl2 for sd */
		.ctrl3[0] = 0,		/* ctrl3 for low speed */
		.ctrl3[1] = 0,		/* ctrl3 for high speed */
		.ctrl4 = 0,
	},
};

/* s3c_hsmmc_get_platdata
 *
 * get the platform data associated with the given device, or return
 * the default if there is none
 */

static struct s3c_hsmmc_cfg *s3c_hsmmc_get_platdata (struct device *dev)
{
	if (dev->platform_data != NULL)
		return (struct s3c_hsmmc_cfg *)dev->platform_data;

	return &s3c_hsmmc_platform;
}

static void sdhci_dumpregs(struct s3c_hsmmc_host *host)
{
	printk(KERN_DEBUG DRIVER_NAME ": ============== REGISTER DUMP ==============\n");

	printk(KERN_DEBUG DRIVER_NAME ": Sys addr: 0x%08x | Version:  0x%08x\n",
		readl(host->ioaddr + SDHCI_DMA_ADDRESS),
		readw(host->ioaddr + SDHCI_HOST_VERSION));
	printk(KERN_DEBUG DRIVER_NAME ": Blk size: 0x%08x | Blk cnt:  0x%08x\n",
		readw(host->ioaddr + SDHCI_BLOCK_SIZE),
		readw(host->ioaddr + SDHCI_BLOCK_COUNT));
	printk(KERN_DEBUG DRIVER_NAME ": Argument: 0x%08x | Trn mode: 0x%08x\n",
		readl(host->ioaddr + SDHCI_ARGUMENT),
		readw(host->ioaddr + SDHCI_TRANSFER_MODE));
	printk(KERN_DEBUG DRIVER_NAME ": Present:  0x%08x | Host ctl: 0x%08x\n",
		readl(host->ioaddr + SDHCI_PRESENT_STATE),
		readb(host->ioaddr + SDHCI_HOST_CONTROL));
	printk(KERN_DEBUG DRIVER_NAME ": Power:    0x%08x | Blk gap:  0x%08x\n",
		readb(host->ioaddr + SDHCI_POWER_CONTROL),
		readb(host->ioaddr + SDHCI_BLOCK_GAP_CONTROL));
	printk(KERN_DEBUG DRIVER_NAME ": Wake-up:  0x%08x | Clock:    0x%08x\n",
		readb(host->ioaddr + SDHCI_WAKE_UP_CONTROL),
		readw(host->ioaddr + SDHCI_CLOCK_CONTROL));
	printk(KERN_DEBUG DRIVER_NAME ": Timeout:  0x%08x | Int stat: 0x%08x\n",
		readb(host->ioaddr + SDHCI_TIMEOUT_CONTROL),
		readl(host->ioaddr + SDHCI_INT_STATUS));
	printk(KERN_DEBUG DRIVER_NAME ": Int enab: 0x%08x | Sig enab: 0x%08x\n",
		readl(host->ioaddr + SDHCI_INT_ENABLE),
		readl(host->ioaddr + SDHCI_SIGNAL_ENABLE));
	printk(KERN_DEBUG DRIVER_NAME ": AC12 err: 0x%08x | Slot int: 0x%08x\n",
		readw(host->ioaddr + SDHCI_ACMD12_ERR),
		readw(host->ioaddr + SDHCI_SLOT_INT_STATUS));
	printk(KERN_DEBUG DRIVER_NAME ": Caps:     0x%08x | Max curr: 0x%08x\n",
		readl(host->ioaddr + SDHCI_CAPABILITIES),
		readl(host->ioaddr + SDHCI_MAX_CURRENT));

	if (host->flags & SDHCI_USE_ADMA)
		printk(KERN_DEBUG DRIVER_NAME ": ADMA Err: 0x%08x | ADMA Ptr: 0x%08x\n",
		       readl(host->ioaddr + SDHCI_ADMA_ERROR),
		       readl(host->ioaddr + SDHCI_ADMA_ADDRESS));

	printk(KERN_DEBUG DRIVER_NAME ": ===========================================\n");
}

/*****************************************************************************\
 *                                                                           *
 * Low level functions                                                       *
 *                                                                           *
\*****************************************************************************/

static void sdhci_reset(struct s3c_hsmmc_host *host, u8 mask)
{
	unsigned long timeout;

	if (host->quirks & SDHCI_QUIRK_NO_CARD_NO_RESET) {
		if (!(readl(host->ioaddr + SDHCI_PRESENT_STATE) &
			SDHCI_CARD_PRESENT))
			return;
	}

	writeb(mask, host->ioaddr + SDHCI_SOFTWARE_RESET);

	if (mask & SDHCI_RESET_ALL)
		host->clock = 0;

	/* Wait max 100 ms */
	timeout = 100;

	/* hw clears the bit when it's done */
	while (readb(host->ioaddr + SDHCI_SOFTWARE_RESET) & mask) {
		if (timeout == 0) {
			printk(KERN_ERR "%s: Reset 0x%x never completed.\n",
				mmc_hostname(host->mmc), (int)mask);
			sdhci_dumpregs(host);
			return;
		}
		timeout--;
		mdelay(1);
	}
}

static void sdhci_init(struct s3c_hsmmc_host *host)
{
	u32 intmask;

	sdhci_reset(host, SDHCI_RESET_ALL);

	intmask = SDHCI_INT_BUS_POWER | SDHCI_INT_DATA_END_BIT |
		SDHCI_INT_DATA_CRC | SDHCI_INT_DATA_TIMEOUT | SDHCI_INT_INDEX |
		SDHCI_INT_END_BIT | SDHCI_INT_CRC | SDHCI_INT_TIMEOUT |
		SDHCI_INT_CARD_REMOVE | SDHCI_INT_CARD_INSERT |
		SDHCI_INT_DATA_AVAIL | SDHCI_INT_SPACE_AVAIL |
		SDHCI_INT_DMA_END | SDHCI_INT_DATA_END | SDHCI_INT_RESPONSE |
		SDHCI_INT_ADMA_ERROR;

	writel(intmask, host->ioaddr + SDHCI_INT_ENABLE);
	writel(intmask, host->ioaddr + SDHCI_SIGNAL_ENABLE);
}

static void sdhci_activate_led(struct s3c_hsmmc_host *host)
{
	u8 ctrl;

	ctrl = readb(host->ioaddr + SDHCI_HOST_CONTROL);
	ctrl |= SDHCI_CTRL_LED;
	writeb(ctrl, host->ioaddr + SDHCI_HOST_CONTROL);
}

static void sdhci_deactivate_led(struct s3c_hsmmc_host *host)
{
	u8 ctrl;

	ctrl = readb(host->ioaddr + SDHCI_HOST_CONTROL);
	ctrl &= ~SDHCI_CTRL_LED;
	writeb(ctrl, host->ioaddr + SDHCI_HOST_CONTROL);
}

#ifdef SDHCI_USE_LEDS_CLASS
static void sdhci_led_control(struct led_classdev *led,
        enum led_brightness brightness)
{
	struct s3c_hsmmc_host *host = container_of(led, struct s3c_hsmmc_host, led);
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);

	if (brightness == LED_OFF)
		sdhci_deactivate_led(host);
	else
		sdhci_activate_led(host);

	spin_unlock_irqrestore(&host->lock, flags);
}
#endif

/*****************************************************************************\
 *                                                                           *
 * Core functions                                                            *
 *                                                                           *
\*****************************************************************************/

#ifdef CONFIG_HSMMC_SCATTERGATHER
static inline uint s3c_hsmmc_build_dma_table (struct s3c_hsmmc_host *host,
						struct mmc_data *data)
{
	uint i;
	struct scatterlist * sg = data->sg;

	/* build dma table except the last one */
	for (i=0; i<(host->sg_len-1); i++) {
		host->sdma_descr_tbl[i].dma_address = sg[i].dma_address;
		host->sdma_descr_tbl[i].length_attr =
			(sg[i].length << 16) | (S3C_HSMMC_ADMA_ATTR_ACT_TRAN) |
			(S3C_HSMMC_ADMA_ATTR_VALID);

		DBG("  ADMA2 descr table[%d] - addr: %08x, size+attr: %08x\n",
			i, host->sdma_descr_tbl[i].dma_address,
			host->sdma_descr_tbl[i].length_attr);
	}

	/* the last one */
	host->sdma_descr_tbl[i].dma_address = sg[i].dma_address;
	host->sdma_descr_tbl[i].length_attr =
			(sg[i].length << 16) | (S3C_HSMMC_ADMA_ATTR_ACT_TRAN) |
			(S3C_HSMMC_ADMA_ATTR_END | S3C_HSMMC_ADMA_ATTR_VALID);

	DBG("  ADMA2 descr table[%d] - addr: %08x, size+attr: %08x\n",
		i, host->sdma_descr_tbl[i].dma_address,
		host->sdma_descr_tbl[i].length_attr);

 	return (i);
}
#endif /* CONFIG_HSMMC_SCATTERGATHER */

static void s3c_hsmmc_prepare_data(struct s3c_hsmmc_host *host, struct mmc_command *cmd)
{
	u8 reg8;
	u32 reg;
	struct mmc_data *data = cmd->data;

	if (data == NULL) {
		reg = s3c_hsmmc_readl(SDHCI_INT_ENABLE) | SDHCI_INT_RESPONSE;
		s3c_hsmmc_writel(reg, SDHCI_INT_ENABLE);
		return;
	}

	host->data_early = 0;

	reg = s3c_hsmmc_readl(SDHCI_INT_ENABLE) & ~SDHCI_INT_RESPONSE;
	s3c_hsmmc_writel(reg, SDHCI_INT_ENABLE);

	host->dma_dir = (data->flags & MMC_DATA_READ)
				? DMA_FROM_DEVICE : DMA_TO_DEVICE;

#ifdef CONFIG_HSMMC_SCATTERGATHER
	host->sg_len = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				host->dma_dir);

	reg = s3c_hsmmc_readl(SDHCI_INT_ENABLE);
	if (host->sg_len == 1) {
		reg &= ~SDHCI_INT_DMA_END;
	} else {
		reg |= SDHCI_INT_DMA_END;
	}
	s3c_hsmmc_writel(reg, SDHCI_INT_ENABLE);

	reg8 = s3c_hsmmc_readb(SDHCI_HOST_CONTROL);
	reg8 |= SDHCI_CTRL_ADMA32;
	s3c_hsmmc_writeb(reg8, SDHCI_HOST_CONTROL);
	DBG("HOSTCTL(0x28) = 0x%02x\n", s3c_hsmmc_readb(SDHCI_HOST_CONTROL));
	DBG("data->flags(direction) =  0x%x\n", data->flags);
	DBG("data->blksz: %d\n", data->blksz);
	DBG("data->blocks: %d\n", data->blocks);
	DBG("data->sg_len: %d\n", data->sg_len);

	DBG("data->sg->addr: 0x%x\n", data->sg->dma_address);
	DBG("data->sg->length: 0x%x\n", data->sg->length);

	host->dma_blk = s3c_hsmmc_build_dma_table(host, data);
	host->next_blk = 0;
#else /* CONFIG_HSMMC_SCATTERGATHER */
	DBG("data->flags(direction) =  0x%x\n", data->flags);
	DBG("data->blksz: %d\n", data->blksz);
	DBG("data->blocks: %d\n", data->blocks);
	DBG("data->sg_len: %d\n", data->sg_len);
	dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				host->dma_dir);
#endif /* CONFIG_HSMMC_SCATTERGATHER */
}

static inline void s3c_hsmmc_set_transfer_mode (struct s3c_hsmmc_host *host,
					struct mmc_data *data)
{
	u16 mode;

	mode = SDHCI_TRNS_DMA;

	if (data->stop)
		mode |= SDHCI_TRNS_ACMD12;

	if (data->blocks > 1)
		mode |= SDHCI_TRNS_MULTI | SDHCI_TRNS_BLK_CNT_EN;
	if (data->flags & MMC_DATA_READ)
		mode |= SDHCI_TRNS_READ;

	s3c_hsmmc_writew(mode, SDHCI_TRANSFER_MODE);
}

static void s3c_hsmmc_finish_data (struct s3c_hsmmc_host *host)
{
	struct mmc_data *data;
	u16 blocks;

	//BUG_ON(!host->data);
	if(!host->data) return;

	data = host->data;
	host->data = NULL;

	dma_unmap_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
			(data->flags & MMC_DATA_READ)
			? DMA_FROM_DEVICE : DMA_TO_DEVICE);

	/*
	 * Controller doesn't count down when in single block mode.
	 */
	if ((data->blocks == 1) && (data->error == 0))
		blocks = 0;
	else {
		blocks = s3c_hsmmc_readw(SDHCI_BLOCK_COUNT);
	}
	data->bytes_xfered = data->blksz * (data->blocks - blocks);

	if ((data->error == 0) && blocks) {
		printk(KERN_ERR "%s: Controller signalled completion even "
			"though there were blocks left. : %d\n",
			mmc_hostname(host->mmc), blocks);
		data->error = -EIO;
	}
	DBG("data->flags(direction) FINISHED =  %d\n", data->flags);
	DBG("data->blksz FINISHED =  %d\n", data->blksz);
	DBG("data->blocks FINISHED =  %d\n", data->blocks);
	DBG("Not FINISHED blocks =  %d\n", blocks);
	DBG("Ending data transfer (%d bytes)\n", data->bytes_xfered);
	DBG("\n");

#ifdef USE_FINISH_TASKLET
	tasklet_schedule(&host->finish_tasklet);
#else
	sdhci_tasklet_finish((unsigned long) host);
#endif
}

static inline void s3c_hsmmc_send_command (struct s3c_hsmmc_host *host,
						struct mmc_command *cmd)
{
	u32 mask=1;
	ulong timeout;

	while (s3c_hsmmc_readl(S3C_HSMMC_CONTROL4) & mask);

	DBG("Sending cmd=(%d), arg=0x%x\n", cmd->opcode, cmd->arg);

	/* Wait max 10 ms */
	timeout = 10;

	mask = SDHCI_CMD_INHIBIT;
	if ((cmd->data != NULL) || (cmd->flags & MMC_RSP_BUSY))
		mask |= SDHCI_DATA_INHIBIT;

	while (s3c_hsmmc_readl(SDHCI_PRESENT_STATE) & mask) {
		printk("########### waiting controller wakeup\n");
		if (timeout == 0) {
			printk(KERN_ERR "%s: Controller never released "
				"inhibit bit(s).\n", mmc_hostname(host->mmc));
			cmd->error = -EIO;

#ifdef USE_FINISH_TASKLET
			tasklet_schedule(&host->finish_tasklet);
#else
			sdhci_tasklet_finish((unsigned long) host);
#endif
			return;
		}
		timeout--;
		mdelay(1);
	}

	mod_timer(&host->timer, jiffies + 10 * HZ);

	host->cmd = cmd;

	s3c_hsmmc_prepare_data(host, cmd);
	s3c_hsmmc_send_register(host);
}

static void sdhci_finish_command(struct s3c_hsmmc_host *host)
{
	int i;

	BUG_ON(host->cmd == NULL);
//	if(!host->cmd) return;

	if (host->cmd->flags & MMC_RSP_PRESENT) {
		if (host->cmd->flags & MMC_RSP_136) {
			/* CRC is stripped so we need to do some shifting. */
			for (i=0; i<4; i++) {
				host->cmd->resp[i] = readl(host->ioaddr +
					SDHCI_RESPONSE + (3-i)*4) << 8;
				if (i != 3)
					host->cmd->resp[i] |=
						readb(host->ioaddr +
						SDHCI_RESPONSE + (3-i)*4-1);
			}
		} else {
			host->cmd->resp[0] = readl(host->ioaddr + SDHCI_RESPONSE);
		}
	}

	host->cmd->error = 0;

	if (host->cmd->data)
		host->data = host->cmd->data;

	if (!host->cmd->data)
#ifdef USE_FINISH_TASKLET
		tasklet_schedule(&host->finish_tasklet);
#else
		sdhci_tasklet_finish((unsigned long) host);
#endif

	host->cmd = NULL;
}

static void sdhci_set_clock(struct s3c_hsmmc_host *host, unsigned int clock)
{
        if (clock == host->clock)
                return;

	s3c_hsmmc_change_clock(host, clock);
}

static void s3c_hsmmc_change_clock(struct s3c_hsmmc_host *host, unsigned int clock)
{
	struct s3c_hsmmc_cfg *cfg = host->plat_data;
	int cardtype = 1;
	u32 val = 0, tmp_clk = 0, clk_src = 0, i, j;
	u16 div = (u16)-1;
	ulong timeout;	
	u8 ctrl;

	/* if we already set, just out. */
	if (clock == host->clock) {
		printk("%p:host->clock0 : %d Hz\n", host->ioaddr, host->clock);
		return;
	}

	/* before setting clock, clkcon must be disabled. */
	s3c_hsmmc_writew(0, SDHCI_CLOCK_CONTROL);

	s3c_hsmmc_writeb(S3C_HSMMC_TIMEOUT_MAX, SDHCI_TIMEOUT_CONTROL);

	/* change the edge type according to frequency */
	ctrl = s3c_hsmmc_readb(SDHCI_HOST_CONTROL);

	if (cfg->highspeed)
		ctrl |= SDHCI_CTRL_HISPD;
	else
		ctrl &= ~SDHCI_CTRL_HISPD;
	
	s3c_hsmmc_writeb(ctrl, SDHCI_HOST_CONTROL);

	if (clock == 0) {
		DBG(" In case of 0 Hz of clock, I'm afraid DO NOTHING ..\n");
		return;
	}

	/* calculate optimal clock. by scsuh */
	DBG("Requested clock is  %ld Hz\n", clock);

	for (i=0; i<NUM_OF_HSMMC_CLKSOURCES; i++) {
		if ((tmp_clk = clk_get_rate(host->clk[i])) <= clock) {
			if (tmp_clk >= val) {
				val = tmp_clk;
				div = 0;
				clk_src = i+1;
			}
		}

		for (j=0x1; j<=0x80; j <<= 1) {
			tmp_clk = clk_get_rate(host->clk[i]) / (j<<1);
			if ((val < tmp_clk) && (tmp_clk <= clock)) {
				val = tmp_clk;
				div = j;
				clk_src = i+1;
				break;
			}
		}
		DBG("   tmp_val[%d]: %d\n", i, val);
	}

	DBG("Optimal clock : %d Hz, div: 0x%x, SelBaseclk_src: %d\n", val, div, clk_src);

	/* CONTROL2 */
	s3c_hsmmc_writel(cfg->fd_ctrl[cardtype].ctrl2 | (clk_src << 4), S3C_HSMMC_CONTROL2);

	/* CONTROL3 */
	if (clock > 25000000)
		s3c_hsmmc_writel(cfg->fd_ctrl[cardtype].ctrl3[SPEED_HIGH], S3C_HSMMC_CONTROL3);
	else
		s3c_hsmmc_writel(cfg->fd_ctrl[cardtype].ctrl3[SPEED_NORMAL], S3C_HSMMC_CONTROL3);

	s3c_hsmmc_writel(cfg->fd_ctrl[cardtype].ctrl4 << 16, S3C_HSMMC_CONTROL4);

	s3c_hsmmc_writew(((div << 8) | SDHCI_CLOCK_INT_EN), SDHCI_CLOCK_CONTROL);

	timeout = 10;
	while (!((val = s3c_hsmmc_readw(SDHCI_CLOCK_CONTROL))
				& SDHCI_CLOCK_INT_STABLE)) {
		if (!timeout) {
			printk("Error in INTERNAL clock stabilization: %08x\n", val);
			return;
		}
		timeout--;
		mdelay(1);
	}

	s3c_hsmmc_writew(val | SDHCI_CLOCK_CARD_EN, SDHCI_CLOCK_CONTROL);

	timeout = 10;
	while (!((val = s3c_hsmmc_readw(SDHCI_CLOCK_CONTROL))
				& S3C_HSMMC_CLOCK_EXT_STABLE)) {
		if (!timeout) {
			printk("Error in EXTERNAL clock stabilization: %08x\n", val);
			return;
		}
		timeout--;
		mdelay(1);
	}
}

static void s3c_hsmmc_request (struct mmc_host *mmc, struct mmc_request *mrq)
{
	struct s3c_hsmmc_host *host = mmc_priv(mmc);
	unsigned long flags;

	DBG("hsmmc request: [CMD] opcode:%d arg:0x%08x flags:0x%02x retries:%u\n",
		mrq->cmd->opcode, mrq->cmd->arg, mrq->cmd->flags, mrq->cmd->retries);

	spin_lock_irqsave(&host->lock, flags);

	WARN_ON(host->mrq != NULL);

	sdhci_activate_led(host);

	host->mrq = mrq;

	if ((s3c_hsmmc_readl(SDHCI_PRESENT_STATE) & SDHCI_CARD_PRESENT) || card_detect) {
		s3c_hsmmc_send_command(host, mrq->cmd);
	} else {
		host->mrq->cmd->error = -ETIMEDOUT;

#ifdef USE_FINISH_TASKLET
		tasklet_schedule(&host->finish_tasklet);
#else
		sdhci_tasklet_finish((unsigned long) host);
#endif
	}


	mmiowb();
	spin_unlock_irqrestore(&host->lock, flags);
}

static void s3c_hsmmc_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct s3c_hsmmc_host *host = mmc_priv(mmc);

	unsigned long iflags;

	spin_lock_irqsave(&host->lock, iflags);

	/*
	 * Reset the chip on each power off.
	 * Should clear out any weird states.
	 */
	if (ios->power_mode == MMC_POWER_OFF) {
		s3c_hsmmc_writew(0, SDHCI_SIGNAL_ENABLE);
		sdhci_init(host);
	}

	if (host->plat_data->enabled)
		hsmmc_set_gpio(host->plat_data->hwport, host->plat_data->bus_width);

	DBG("\nios->clock: %d Hz\n", ios->clock);
	sdhci_set_clock(host, ios->clock);

	DBG("\nios->bus_width: %d\n", ios->bus_width);
	s3c_set_bus_width(host, ios->bus_width);

	DBG("S3C_HSMMC_CONTROL2(0x80) = 0x%08x\n", s3c_hsmmc_readl(S3C_HSMMC_CONTROL2));
	DBG("S3C_HSMMC_CONTROL3(0x84) = 0x%08x\n", s3c_hsmmc_readl(S3C_HSMMC_CONTROL3));
	DBG("SDHCI_CLOCK_CONTROL  (0x2c) = 0x%04x\n", s3c_hsmmc_readw(SDHCI_CLOCK_CONTROL));

	if (ios->power_mode == MMC_POWER_OFF)
		s3c_hsmmc_writeb(S3C_HSMMC_POWER_OFF, SDHCI_POWER_CONTROL);
	else
		s3c_hsmmc_writeb(S3C_HSMMC_POWER_ON_ALL, SDHCI_POWER_CONTROL);

	udelay(1000);
	spin_unlock_irqrestore(&host->lock, iflags);
}

static void sdhci_enable_sdio_irq(struct mmc_host *mmc, int enable)
{
        struct s3c_hsmmc_host *host;
        unsigned long flags;
        u32 ier;

        host = mmc_priv(mmc);

        spin_lock_irqsave(&host->lock, flags);

        if (host->flags & SDHCI_DEVICE_DEAD)
                goto out;

        ier = readl(host->ioaddr + SDHCI_INT_ENABLE);

        ier &= ~SDHCI_INT_CARD_INT;
        if (enable)
                ier |= SDHCI_INT_CARD_INT;

        writel(ier, host->ioaddr + SDHCI_INT_ENABLE);
        writel(ier, host->ioaddr + SDHCI_SIGNAL_ENABLE);

out:
        mmiowb();

        spin_unlock_irqrestore(&host->lock, flags);
}

static struct mmc_host_ops s3c_hsmmc_ops = {
	.request	= s3c_hsmmc_request,
	.set_ios	= s3c_hsmmc_set_ios,
	.enable_sdio_irq = sdhci_enable_sdio_irq,
};

/*****************************************************************************\
 *                                                                           *
 * Tasklets                                                                  *
 *                                                                           *
\*****************************************************************************/

static void sdhci_tasklet_card(unsigned long param)
{
	struct s3c_hsmmc_host *host;
	unsigned long flags;

	host = (struct s3c_hsmmc_host*)param;

	spin_lock_irqsave(&host->lock, flags);

	if (!(readl(host->ioaddr + SDHCI_PRESENT_STATE) & SDHCI_CARD_PRESENT)) {
		if (host->mrq) {
			printk(KERN_ERR "%s: Card removed during transfer!\n",
				mmc_hostname(host->mmc));
			printk(KERN_ERR "%s: Resetting controller.\n",
				mmc_hostname(host->mmc));

			sdhci_reset(host, SDHCI_RESET_CMD);
			sdhci_reset(host, SDHCI_RESET_DATA);

			host->mrq->cmd->error = -ENOMEDIUM;
#ifdef USE_FINISH_TASKLET
			tasklet_schedule(&host->finish_tasklet);
#else
			sdhci_tasklet_finish((unsigned long) host);
#endif
		}
	}
 
	spin_unlock_irqrestore(&host->lock, flags);

	mmc_detect_change(host->mmc, msecs_to_jiffies(200));
}

static void sdhci_tasklet_finish(unsigned long param)
{
	struct s3c_hsmmc_host *host;
	unsigned long flags;
	struct mmc_request *mrq;

	host = (struct s3c_hsmmc_host*)param;

	if(!host->mrq) return;

	spin_lock_irqsave(&host->lock, flags);

	del_timer(&host->timer);

	mrq = host->mrq;
	udelay(10);

	/*
	 * The controller needs a reset of internal state machines
	 * upon error conditions.
	 */
	if (!(host->flags & SDHCI_DEVICE_DEAD) &&
		(mrq->cmd->error ||
		 (mrq->data && (mrq->data->error ||
		  (mrq->data->stop && mrq->data->stop->error))) ||
		   (host->quirks & SDHCI_QUIRK_RESET_AFTER_REQUEST))) {

		/* Some controllers need this kick or reset won't work here */
		if (host->quirks & SDHCI_QUIRK_CLOCK_BEFORE_RESET) {
			unsigned int clock;

			/* This is to force an update */
			clock = host->clock;
			host->clock = 0;
			sdhci_set_clock(host, clock);
		}

		/* Spec says we should do both at the same time, but Ricoh
		   controllers do not like that. */
		sdhci_reset(host, SDHCI_RESET_CMD);
		sdhci_reset(host, SDHCI_RESET_DATA);
	}

	host->mrq = NULL;
	host->cmd = NULL;
	host->data = NULL;

#ifndef SDHCI_USE_LEDS_CLASS
	sdhci_deactivate_led(host);
#endif

	mmiowb();
	spin_unlock_irqrestore(&host->lock, flags);

	mmc_request_done(host->mmc, mrq);
}

/*****************************************************************************\
 *                                                                           *
 * Core functions                                                            *
 *                                                                           *
\*****************************************************************************/
static void s3c_hsmmc_send_register (struct s3c_hsmmc_host *host)
{
	struct mmc_command *cmd = host->cmd;
	struct mmc_data *data = cmd->data;

	u32 cmd_val;

	if (data) {
#ifdef CONFIG_HSMMC_SCATTERGATHER
		s3c_hsmmc_writew(SDHCI_MAKE_BLKSZ(0x7, data->blksz), SDHCI_BLOCK_SIZE);
		s3c_hsmmc_writel(virt_to_phys(host->sdma_descr_tbl), SDHCI_ADMA_ADDRESS);
		DBG("SDHCI_ADMA_ADDRESS(0x58) = 0x%08x\n", s3c_hsmmc_readl(SDHCI_ADMA_ADDRESS));
#else	/* CONFIG_HSMMC_SCATTERGATHER */
		s3c_hsmmc_writew(SDHCI_MAKE_BLKSZ(0x7, data->blksz), SDHCI_BLOCK_SIZE);
		s3c_hsmmc_writel(sg_dma_address(data->sg), SDHCI_DMA_ADDRESS);
#endif /* CONFIG_HSMMC_SCATTERGATHER */
		s3c_hsmmc_writew(data->blocks, SDHCI_BLOCK_COUNT);
		s3c_hsmmc_set_transfer_mode(host, data);
	}

	s3c_hsmmc_writel(cmd->arg, SDHCI_ARGUMENT);

	cmd_val = (cmd->opcode << 8);
	if (cmd_val == (12<<8))
		cmd_val |= (3 << 6);

	if (cmd->flags & MMC_RSP_136)		/* Long RSP */
		cmd_val |= SDHCI_CMD_RESP_LONG;
	else if (cmd->flags & MMC_RSP_BUSY)	/* R1B */
		cmd_val |= SDHCI_CMD_RESP_SHORT_BUSY;
	else if (cmd->flags & MMC_RSP_PRESENT)	/* Normal RSP */
		cmd_val |= SDHCI_CMD_RESP_SHORT;

	if (cmd->flags & MMC_RSP_OPCODE)
		cmd_val |= SDHCI_CMD_INDEX;

	if (cmd->flags & MMC_RSP_CRC)
		cmd_val |= SDHCI_CMD_CRC;

	if (data)
		cmd_val |= SDHCI_CMD_DATA;

	s3c_hsmmc_writew(cmd_val, SDHCI_COMMAND);
}

/*****************************************************************************\
 *                                                                           *
 * Interrupt handling                                                        *
 *                                                                           *
\*****************************************************************************/

static void sdhci_cmd_irq(struct s3c_hsmmc_host *host, u32 intmask)
{
	BUG_ON(intmask == 0);

	if (!host->cmd) {
		printk(KERN_ERR "%s: Got command interrupt 0x%08x even "
			"though no command operation was in progress.\n",
			mmc_hostname(host->mmc), (unsigned)intmask);
		sdhci_dumpregs(host);
		return;
	}

	if (intmask & SDHCI_INT_TIMEOUT)
		host->cmd->error = -ETIMEDOUT;
	else if (intmask & (SDHCI_INT_CRC | SDHCI_INT_END_BIT |
			SDHCI_INT_INDEX))
		host->cmd->error = -EILSEQ;

	if (host->cmd->error) {
#ifdef USE_FINISH_TASKLET
		tasklet_schedule(&host->finish_tasklet);
#else
		sdhci_tasklet_finish((unsigned long) host);
#endif
		return;
	}

	/*
	 * The host can send and interrupt when the busy state has
	 * ended, allowing us to wait without wasting CPU cycles.
	 * Unfortunately this is overloaded on the "data complete"
	 * interrupt, so we need to take some care when handling
	 * it.
	 *
	 * Note: The 1.0 specification is a bit ambiguous about this
	 *       feature so there might be some problems with older
	 *       controllers.
	 */
	if (host->cmd->flags & MMC_RSP_BUSY) {
		if (host->cmd->data)
			DBG("Cannot wait for busy signal when also "
				"doing a data transfer");
		else if (!(host->quirks & SDHCI_QUIRK_NO_BUSY_IRQ))
			return;

		/* The controller does not support the end-of-busy IRQ,
		 * fall through and take the SDHCI_INT_RESPONSE */
	}

	if (intmask & SDHCI_INT_RESPONSE)
		sdhci_finish_command(host);
}

static void sdhci_data_irq (struct s3c_hsmmc_host *host, u32 intmask)
{
	BUG_ON(intmask == 0);

	if (!host->data) {
		/*
		 * The "data complete" interrupt is also used to
		 * indicate that a busy state has ended. See comment
		 * above in sdhci_cmd_irq().
		 */
		if (host->cmd && (host->cmd->flags & MMC_RSP_BUSY)) {
			if (intmask & SDHCI_INT_DATA_END) {
				sdhci_finish_command(host);
			return;
			}
		}

		printk(KERN_ERR "%s: Got data interrupt 0x%08x even "
			"though no data operation was in progress.\n",
			mmc_hostname(host->mmc), (unsigned)intmask);
		sdhci_dumpregs(host);

		return;
	}

	if (intmask & SDHCI_INT_DATA_TIMEOUT)
		host->data->error = -ETIMEDOUT;
	else if (intmask & (SDHCI_INT_DATA_CRC | SDHCI_INT_DATA_END_BIT))
		host->data->error = -EILSEQ;
	else if (intmask & SDHCI_INT_ADMA_ERROR)
		host->data->error = -EIO;

	if (host->data->error)
		s3c_hsmmc_finish_data(host);
	else {
		/*
		 * We currently don't do anything fancy with DMA
		 * boundaries, but as we can't disable the feature
		 * we need to at least restart the transfer.
		 */
		if (intmask & SDHCI_INT_DMA_END)
			writel(readl(host->ioaddr + SDHCI_DMA_ADDRESS),
				host->ioaddr + SDHCI_DMA_ADDRESS);

		if (intmask & SDHCI_INT_DATA_END) {
			if (host->cmd) {
				/*
				 * Data managed to finish before the
				 * command completed. Make sure we do
				 * things in the proper order.
				 */
				host->data_early = 1;
			} else {
				s3c_hsmmc_finish_data(host);
			}
		}
	}
}

static irqreturn_t s3c_hsmmc_irq (int irq, void *dev_id)
{
	irqreturn_t result = 0;
	struct s3c_hsmmc_host *host = dev_id;
	struct mmc_request *mrq;
	u32 intsts;

#ifdef CONFIG_HSMMC_S3C_IRQ_WORKAROUND
	uint i, org_irq_sts;
#endif
	spin_lock(&host->lock);

	mrq = host->mrq;

	intsts = s3c_hsmmc_readw(SDHCI_INT_STATUS);

	/* Sometimes, hsmmc does not update its status bit immediately
	 * when it generates irqs. by scsuh.
	 */
#ifdef CONFIG_HSMMC_S3C_IRQ_WORKAROUND
	for (i=0; i<0x1000; i++) {
		if ((intsts = s3c_hsmmc_readw(SDHCI_INT_STATUS)))
			break;
	}
#endif

	if (unlikely(!intsts)) {
		result = IRQ_NONE;
		goto out;
	}
	intsts = s3c_hsmmc_readl(SDHCI_INT_STATUS);
#ifdef CONFIG_HSMMC_S3C_IRQ_WORKAROUND
	org_irq_sts = intsts;
#endif

	DBG(PFX "Got interrupt = 0x%08x\n", intsts);
 
	if (unlikely(intsts & S3C_HSMMC_INT_CARD_CHANGE)) {
		u32 reg16;

		if (intsts & SDHCI_INT_CARD_INSERT)
			printk(PFX "card inserted.\n");
		else if (intsts & SDHCI_INT_CARD_REMOVE)
			printk(PFX "card removed.\n");

		reg16 = s3c_hsmmc_readw(SDHCI_INT_ENABLE);
		s3c_hsmmc_writew(reg16 & ~S3C_HSMMC_INT_CARD_CHANGE,
							SDHCI_INT_ENABLE);
		s3c_hsmmc_writew(S3C_HSMMC_INT_CARD_CHANGE, SDHCI_INT_STATUS);
		s3c_hsmmc_writew(reg16, SDHCI_INT_ENABLE);

		intsts &= ~S3C_HSMMC_INT_CARD_CHANGE;

		tasklet_schedule(&host->card_tasklet);
		goto insert;
	}
 
	if (likely(!(intsts & SDHCI_INT_ERROR))) {
		s3c_hsmmc_writel(intsts, SDHCI_INT_STATUS);

		if (intsts & SDHCI_INT_RESPONSE) {
			DBG("command done\n");
			sdhci_finish_command(host);
		}

		if (intsts & SDHCI_INT_DATA_END) {
			DBG("transfer done\n\n");
			sdhci_finish_command(host);
			s3c_hsmmc_finish_data(host);
			intsts &= ~SDHCI_INT_DMA_END;
		}
	} else {
		DBG("command FAIL : found bad irq [0x%8x]\n", intsts);
		DBG("\n");
		if (intsts & S3C_HSMMC_INT_CMD_MASK) {
			s3c_hsmmc_writel(intsts & S3C_HSMMC_INT_CMD_MASK, SDHCI_INT_STATUS);
			sdhci_cmd_irq(host, intsts & S3C_HSMMC_INT_CMD_MASK);
		}

		if (intsts & SDHCI_INT_DATA_MASK) {
			s3c_hsmmc_writel(intsts & SDHCI_INT_DATA_MASK, SDHCI_INT_STATUS);
			sdhci_finish_command(host);
			sdhci_data_irq(host, intsts & SDHCI_INT_DATA_MASK);
		}

		intsts &= ~(S3C_HSMMC_INT_CMD_MASK | SDHCI_INT_DATA_MASK);
	}

#ifdef CONFIG_HSMMC_S3C_IRQ_WORKAROUND
	for (i=0; i<0x1000; i++) {
		if (org_irq_sts != s3c_hsmmc_readl(SDHCI_INT_STATUS))
			break;
	}
#endif

insert:
	result = IRQ_HANDLED;
	mmiowb();

out:
	spin_unlock(&host->lock);

	return result;
}

static irqreturn_t s3c_hsmmc_irq_cd (int irq, void *dev_id)
{
	struct s3c_hsmmc_host *host = dev_id;
	int ext_CD_int = 0;

	if (IRQ_EINT0 <= irq && irq <= IRQ_EINT3) {
		ext_CD_int = readl(S3C2410_GPFDAT);
		ext_CD_int &= (1 << (irq - IRQ_EINT0));
	} else if (IRQ_EINT4 <= irq && irq <= IRQ_EINT7) {
		ext_CD_int = readl(S3C2410_GPFDAT);
		ext_CD_int &= (1 << (irq - IRQ_EINT4 + 4));
	} else if (IRQ_EINT8 <= irq && irq <= IRQ_EINT15) {
		ext_CD_int = readl(S3C2410_GPGDAT);
		ext_CD_int &= (1 << (irq - IRQ_EINT8));
	}

	if(ext_CD_int && card_detect) {
		printk("s3c-hsmmc channel-0(EXT): card removed.\n");
		set_irq_type(host->irq_cd, IRQF_TRIGGER_FALLING);
		card_detect = 0;
	}
	else if(!ext_CD_int && !card_detect) {
		printk("s3c-hsmmc channel-0(EXT): card inserted.\n");
		set_irq_type(host->irq_cd, IRQF_TRIGGER_RISING);
		card_detect = 1;
	}
	else
		return IRQ_HANDLED;

	tasklet_schedule(&host->card_tasklet);
	mmiowb();

	spin_unlock(&host->lock);

	return IRQ_HANDLED;
}

static void s3c_hsmmc_check_status(unsigned long data)
{
        struct s3c_hsmmc_host *host = (struct s3c_hsmmc_host *)data;

	s3c_hsmmc_irq(0, host);
}

/* return 0: OK
 * return -1: error
 */
static int s3c_set_bus_width (struct s3c_hsmmc_host *host, uint width)
{
	u8 reg;

	reg = s3c_hsmmc_readb(SDHCI_HOST_CONTROL);

	switch (width) {
	case MMC_BUS_WIDTH_1:
		reg &= ~(SDHCI_CTRL_4BITBUS | S3C_HSMMC_CTRL_8BIT);
		DBG("bus width: 1 bit\n");
		break;
		
	case MMC_BUS_WIDTH_4:
		DBG("bus width: 4 bit\n");
		reg &= ~(S3C_HSMMC_CTRL_8BIT);
		reg |= SDHCI_CTRL_4BITBUS;
		break;
		
	case MMC_BUS_WIDTH_8:
		reg |= S3C_HSMMC_CTRL_8BIT;
		DBG("bus width: 8 bit\n");
		break;
		
	default:
		DBG("bus width: Error\n");
		return -1;
	}

	s3c_hsmmc_writeb(reg, SDHCI_HOST_CONTROL);

	DBG("HOSTCTL(0x28) = 0x%02x\n", s3c_hsmmc_readb(SDHCI_HOST_CONTROL));

	return 0;
}

static int s3c_hsmmc_probe(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct s3c_hsmmc_host *host;
	struct s3c_hsmmc_cfg *plat_data;
	uint i;
	int ret;
	unsigned int caps;

	mmc = mmc_alloc_host(sizeof(struct s3c_hsmmc_host), &pdev->dev);
	if (!mmc)
		ret = -ENOMEM;

	plat_data = s3c_hsmmc_get_platdata(&pdev->dev);

	host = mmc_priv(mmc);

	host->mmc = mmc;

	host->plat_data = plat_data;

	host->mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!host->mem) {
		printk("Failed to get io memory region resouce.\n");
		ret = -ENOENT;
		goto probe_free_host;
	}

	host->mem = request_mem_region(host->mem->start,
		RESSIZE(host->mem), pdev->name);
	if (!host->mem) {
		printk("Failed to request io memory region.\n");
		ret = -ENOENT;
		goto probe_free_host;
	}

	host->ioaddr = ioremap(host->mem->start, (host->mem->end - host->mem->start)+1);
	if (host->ioaddr == NULL) {
		printk(KERN_ERR "Failed to remap register block\n");
		return -ENOMEM;
	}

	host->hw_name = "samsung-hsmmc";
	host->quirks = 0;

	host->irq = platform_get_irq(pdev, 0);
	if (host->irq == 0) {
		printk("Failed to get interrupt resouce.\n");
		ret = -EINVAL;
		goto untasklet;
	}

	platform_set_drvdata(pdev, host);

	if ((plat_data->enabled == 1)) {
		host->irq_cd = plat_data->irq_cd;
		if (host->irq_cd == 0) {
			printk("Failed to get interrupt resouce.\n");
			ret = -EINVAL;
			goto untasklet;
		}
		set_irq_type(host->irq_cd, IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING);
	}

	host->flags |= SDHCI_USE_DMA;

	sdhci_reset(host, SDHCI_RESET_ALL);

	/* register some clock sources if exist */
	for (i=0; i<NUM_OF_HSMMC_CLKSOURCES; i++) {
		host->clk[i] = clk_get(&pdev->dev, plat_data->clk_name[i]);
		if (IS_ERR(host->clk[i])) {
			ret = PTR_ERR(host->clk[i]);
			host->clk[i] = ERR_PTR(-ENOENT);
		}

		if (clk_enable(host->clk[i])) {
			host->clk[i] = ERR_PTR(-ENOENT);
		}
		if (!IS_ERR(host->clk[i])) {
			DBG("MMC clock source[%d], %s is %ld Hz\n",i, plat_data->clk_name[i], clk_get_rate(host->clk[i]));
		}
	}

        host->version = readw(host->ioaddr + SDHCI_HOST_VERSION);
        host->version = (host->version & SDHCI_SPEC_VER_MASK)
                                >> SDHCI_SPEC_VER_SHIFT;
        if (host->version > SDHCI_SPEC_200) {
                printk(KERN_ERR "%s: Unknown controller version (%d). "
                        "You may experience problems.\n", mmc_hostname(mmc),
                        host->version);
        }

	caps = readl(host->ioaddr + SDHCI_CAPABILITIES);

	if (host->quirks & SDHCI_QUIRK_FORCE_DMA)
                host->flags |= SDHCI_USE_DMA;
        else if (!(caps & SDHCI_CAN_DO_DMA))
                DBG("Controller doesn't have DMA capability\n");
        else
                host->flags |= SDHCI_USE_DMA;

	if ((host->quirks & SDHCI_QUIRK_BROKEN_DMA) &&
                (host->flags & SDHCI_USE_DMA)) {
                DBG("Disabling DMA as it is marked broken\n");
                host->flags &= ~SDHCI_USE_DMA;
        }

	if (host->flags & SDHCI_USE_DMA) {
                if ((host->version >= SDHCI_SPEC_200) &&
                                (caps & SDHCI_CAN_DO_ADMA2))
                        host->flags |= SDHCI_USE_ADMA;
        }

	if ((host->quirks & SDHCI_QUIRK_BROKEN_ADMA) &&
                (host->flags & SDHCI_USE_ADMA)) {
                DBG("Disabling ADMA as it is marked broken\n");
                host->flags &= ~SDHCI_USE_ADMA;
        }

        if (host->flags & SDHCI_USE_ADMA) {
                /*
                 * We need to allocate descriptors for all sg entries
                 * (128) and potentially one alignment transfer for
                 * each of those entries.
                 */
                host->adma_desc = kmalloc((128 * 2 + 1) * 4, GFP_KERNEL);
                host->align_buffer = kmalloc(128 * 4, GFP_KERNEL);
                if (!host->adma_desc || !host->align_buffer) {
                        kfree(host->adma_desc);
                        kfree(host->align_buffer);
                        printk(KERN_WARNING "%s: Unable to allocate ADMA "
                                "buffers. Falling back to standard DMA.\n",
                                mmc_hostname(mmc));
                        host->flags &= ~SDHCI_USE_ADMA;
                }
        }

        /*
         * If we use DMA, then it's up to the caller to set the DMA
         * mask, but PIO does not need the hw shim so we set a new
         * mask here in that case.
         */
        if (!(host->flags & SDHCI_USE_DMA)) {
                host->dma_mask = DMA_BIT_MASK(64);
                mmc_dev(host->mmc)->dma_mask = &host->dma_mask;
        }

	/*
	 * Set host parameters.
	 */
	mmc->ops = &s3c_hsmmc_ops;
#if 1
	mmc->f_min = 400 * 1000; /* at least 400kHz */
	mmc->f_max = 100 * MHZ;
#else
	mmc->f_min = host->max_clk / 256;
        mmc->f_max = host->max_clk;
#endif

	mmc->caps = MMC_CAP_4_BIT_DATA | MMC_CAP_SDIO_IRQ;

	if (caps & SDHCI_CAN_DO_HISPD)
                mmc->caps |= MMC_CAP_SD_HIGHSPEED;

        mmc->ocr_avail = 0;
        if (caps & SDHCI_CAN_VDD_330)
                mmc->ocr_avail |= MMC_VDD_32_33|MMC_VDD_33_34;
        if (caps & SDHCI_CAN_VDD_300)
                mmc->ocr_avail |= MMC_VDD_29_30|MMC_VDD_30_31;
        if (caps & SDHCI_CAN_VDD_180)
                mmc->ocr_avail |= MMC_VDD_165_195;

        if (mmc->ocr_avail == 0) {
                printk(KERN_ERR "%s: Hardware doesn't report any "
                        "support voltages.\n", mmc_hostname(mmc));
                return -ENODEV;
        }

	spin_lock_init(&host->lock);

        /*
         * Maximum number of segments. Depends on if the hardware
         * can do scatter/gather or not.
         */
        if (host->flags & SDHCI_USE_ADMA)
                mmc->max_hw_segs = 128;
        else if (host->flags & SDHCI_USE_DMA)
                mmc->max_hw_segs = 1;
        else /* PIO */
                mmc->max_hw_segs = 128;
        mmc->max_phys_segs = 128;

        /*
         * Maximum number of sectors in one transfer. Limited by DMA boundary
         * size (512KiB).
         */
        mmc->max_req_size = 524288;

        /*
         * Maximum segment size. Could be one segment with the maximum number
         * of bytes. When doing hardware scatter/gather, each entry cannot
         * be larger than 64 KiB though.
         */
        if (host->flags & SDHCI_USE_ADMA)
                mmc->max_seg_size = 65536;
        else
                mmc->max_seg_size = mmc->max_req_size;

        /*
         * Maximum block size. This varies from controller to controller and
         * is specified in the capabilities register.
         */
        mmc->max_blk_size = (caps & SDHCI_MAX_BLOCK_MASK) >> SDHCI_MAX_BLOCK_SHIFT;
        if (mmc->max_blk_size >= 3) {
                printk(KERN_WARNING "%s: Invalid maximum block size, "
                        "assuming 512 bytes\n", mmc_hostname(mmc));
                mmc->max_blk_size = 512;
        } else
                mmc->max_blk_size = 512 << mmc->max_blk_size;

        /*
         * Maximum block count.
         */
        mmc->max_blk_count = 65535;

	init_timer(&host->timer);
        host->timer.data = (unsigned long)host;
        host->timer.function = s3c_hsmmc_check_status;
        host->timer.expires = jiffies + HZ;

	/*
	 * Init tasklets.
	 */
	tasklet_init(&host->card_tasklet,
		sdhci_tasklet_card, (unsigned long)host);

#ifdef USE_FINISH_TASKLET
	tasklet_init(&host->finish_tasklet,
		sdhci_tasklet_finish, (unsigned long)host);
#endif

	ret = request_irq(host->irq, s3c_hsmmc_irq, 0, DRIVER_NAME, host);
	if (ret)
		goto untasklet;

	if (plat_data->enabled == 1 && host->irq_cd) {
		ret = request_irq(host->irq_cd, s3c_hsmmc_irq_cd, 0, DRIVER_NAME, host);
		if (ret)
			goto untasklet;
	}

	sdhci_init(host);

	mmc_add_host(mmc);

	printk(KERN_INFO "[s3c_hsmmc_probe]: %s.%d: at 0x%p with irq %d. clk src:",
			pdev->name, pdev->id, host->ioaddr, host->irq);

	for (i=0; i<NUM_OF_HSMMC_CLKSOURCES; i++) {
		if (!IS_ERR(host->clk[i]))
			printk(" %s", host->clk[i]->name);
	}
	printk("\n");

	return 0;

untasklet:
	tasklet_kill(&host->card_tasklet);
#ifdef USE_FINISH_TASKLET
	tasklet_kill(&host->finish_tasklet);
#endif

	for (i=0; i<NUM_OF_HSMMC_CLKSOURCES; i++) {
		if (host->clk[i] != ERR_PTR(-ENOENT)) {
			clk_disable(host->clk[i]);
			clk_put(host->clk[i]);
		}
	}

probe_free_host:
	mmc_free_host(mmc);

	return ret;
}

static int s3c_hsmmc_remove(struct platform_device *dev)
{
	struct s3c_hsmmc_host *host = platform_get_drvdata(dev);
	struct mmc_host *mmc;
	int i;

	mmc = host->mmc;

	mmc_remove_host(mmc);

	sdhci_reset(host, SDHCI_RESET_ALL);

	for (i=0; i<NUM_OF_HSMMC_CLKSOURCES; i++) {
		clk_disable(host->clk[i]);
		clk_put(host->clk[i]);
	}

	free_irq(host->irq, host);

	del_timer_sync(&host->timer);

	tasklet_kill(&host->card_tasklet);
#ifdef USE_FINISH_TASKLET
	tasklet_kill(&host->finish_tasklet);
#endif
	mmc_free_host(mmc);

	return 0;
}

#ifdef CONFIG_PM
static int s3c_hsmmc_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct s3c_hsmmc_host *host = platform_get_drvdata(pdev);

#if 0
	disable_irq(host->irq_cd);
#else
	enable_irq_wake(host->irq_cd);
#endif
	mmc_suspend_host(host->mmc, state);

	return 0;
}

static int s3c_hsmmc_resume(struct platform_device *pdev)
{
	struct s3c_hsmmc_host *host = platform_get_drvdata(pdev);

	sdhci_init(host);
	mmc_resume_host(host->mmc);
#if 0
	enable_irq(host->irq_cd);
#else
	disable_irq_wake(host->irq_cd);
#endif

	return 0;
}
#else
#define s3c_hsmmc_suspend NULL
#define s3c_hsmmc_resume NULL
#endif

static struct platform_driver s3c_hsmmc_driver =
{
        .probe          = s3c_hsmmc_probe,
        .remove         = s3c_hsmmc_remove,
        .suspend 	= s3c_hsmmc_suspend,
	.resume		= s3c_hsmmc_resume,
	.driver		= {
		.name	= "s3c-hsmmc",
		.owner	= THIS_MODULE,
	},
};

static int __init s3c_hsmmc_drv_init(void)
{
	return platform_driver_register(&s3c_hsmmc_driver);
}

static void __exit s3c_hsmmc_drv_exit(void)
{
	platform_driver_unregister(&s3c_hsmmc_driver);
}

module_init(s3c_hsmmc_drv_init);
module_exit(s3c_hsmmc_drv_exit);

MODULE_DESCRIPTION("S3C SD HOST I/F 1.0 and 2.0 Driver");
MODULE_LICENSE("GPL");
