/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/interrupt.h>
#include <linux/irq.h>

#include <plat/adc.h>

#include <mach/hardware.h>
#include <mach/regs-gpio.h>

static struct s3c_adc_client	*adc_client;

static void yup150_power_init_pin(void)
{
#if 0
	s3c2410_gpio_setpin(S3C2410_GPF3, S3C2410_GPF3_INP);
	s3c2410_gpio_setpin(S3C2410_GPF5, S3C2410_GPF5_INP);
	s3c2410_gpio_setpin(S3C2410_GPF6, S3C2410_GPF6_INP);
#else
	s3c2410_gpio_setpin(S3C2410_GPF3, S3C2410_GPF3_EINT3);
	s3c2410_gpio_setpin(S3C2410_GPF5, S3C2410_GPF5_EINT5);
	s3c2410_gpio_setpin(S3C2410_GPF6, S3C2410_GPF6_EINT6);
#endif
	s3c2410_gpio_setpin(S3C2410_GPG0, S3C2410_GPG0_INP);

	s3c2410_gpio_pullup(S3C2410_GPF3, S3C24XX_GPIO_PULL_NONE);
	s3c2410_gpio_pullup(S3C2410_GPF5, S3C24XX_GPIO_PULL_NONE);
	s3c2410_gpio_pullup(S3C2410_GPF6, S3C24XX_GPIO_PULL_NONE);
	s3c2410_gpio_pullup(S3C2410_GPG0, S3C24XX_GPIO_PULL_NONE);
}

static inline int yup150_is_ext_power_online(void)
{
	return s3c2410_gpio_getpin(S3C2410_GPF6);
}
static inline int yup150_is_usb_vbus_online(void)
{
	return ! s3c2410_gpio_getpin(S3C2410_GPF3);
}
static inline int yup150_is_usb_ac_adapter(void)
{
	return s3c2410_gpio_getpin(S3C2410_GPG0);
}

static inline int yup150_is_battery_charging(void)
{
	if (yup150_is_ext_power_online())
		return ! s3c2410_gpio_getpin(S3C2410_GPF5);
	else
		return 0;
}

static int yup150_is_cradle_online(void)
{
	if (yup150_is_ext_power_online()) {
		return ! yup150_is_usb_vbus_online();
	} else
		return 0;
}
static int yup150_is_ac_online(void)
{
	if (yup150_is_ext_power_online() &&
	    yup150_is_usb_vbus_online())
	{
		return yup150_is_usb_ac_adapter();
	} else
		return 0;
}
static int yup150_is_usb_online(void)
{
	if (yup150_is_ext_power_online() &&
	    yup150_is_usb_vbus_online())
	{
		return ! yup150_is_usb_ac_adapter();
	} else
		return 0;
}

static int yup150_get_bat_voltage(void)
{
	/*
         * adc_val * 2 * 3300 * 1000 / 4096
         */
	return s3c_adc_read(adc_client, 0) * 1611;
}

static char *yup150_power_supplied_to[] = {
	"main-battery",
};

static enum power_supply_property yup150_power_props[] = {
	POWER_SUPPLY_PROP_ONLINE,
};

static enum power_supply_property yup150_bat_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_TECHNOLOGY,
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
	POWER_SUPPLY_PROP_VOLTAGE_MIN_DESIGN,
	POWER_SUPPLY_PROP_VOLTAGE_MAX_DESIGN,
};

static int yup150_power_get_property(struct power_supply *psy,
				  enum power_supply_property psp,
				  union power_supply_propval *val)
{
	switch (psp) {
	case POWER_SUPPLY_PROP_ONLINE:
		if (strcmp(psy->name, "cradle") == 0) {
			val->intval = yup150_is_cradle_online();
		} else if (strcmp(psy->name, "ac") == 0) {
			val->intval = yup150_is_ac_online();
		} else if (strcmp(psy->name, "usb") == 0) {
			val->intval = yup150_is_usb_online();
		} else {
			return -EINVAL;
		}
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int yup150_bat_get_property(struct power_supply *psy,
				  enum power_supply_property psp,
				  union power_supply_propval *val)
{
	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
		if (yup150_is_battery_charging())
			val->intval = POWER_SUPPLY_STATUS_CHARGING;
		else
			val->intval = POWER_SUPPLY_STATUS_NOT_CHARGING;
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:
		val->intval = POWER_SUPPLY_TECHNOLOGY_LION;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_MIN_DESIGN:
		val->intval = 3550 * 1000;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_MAX_DESIGN:
		val->intval = 4200 * 1000;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
		if (! adc_client)
			return -EINVAL;
		val->intval = yup150_get_bat_voltage();
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static struct power_supply yup150_psy_cradle = {
	.name = "cradle",
	.type = POWER_SUPPLY_TYPE_MAINS,
	.supplied_to = yup150_power_supplied_to,
	.num_supplicants = ARRAY_SIZE(yup150_power_supplied_to),
	.properties = yup150_power_props,
	.num_properties = ARRAY_SIZE(yup150_power_props),
	.get_property = yup150_power_get_property,
};

static struct power_supply yup150_psy_ac = {
	.name = "ac",
	.type = POWER_SUPPLY_TYPE_MAINS,
	.supplied_to = yup150_power_supplied_to,
	.num_supplicants = ARRAY_SIZE(yup150_power_supplied_to),
	.properties = yup150_power_props,
	.num_properties = ARRAY_SIZE(yup150_power_props),
	.get_property = yup150_power_get_property,
};

static struct power_supply yup150_psy_usb = {
	.name = "usb",
	.type = POWER_SUPPLY_TYPE_USB,
	.supplied_to = yup150_power_supplied_to,
	.num_supplicants = ARRAY_SIZE(yup150_power_supplied_to),
	.properties = yup150_power_props,
	.num_properties = ARRAY_SIZE(yup150_power_props),
	.get_property = yup150_power_get_property,
};

static struct power_supply yup150_psy_bat = {
	.name = "main-battery",
	.type = POWER_SUPPLY_TYPE_BATTERY,
	.properties = yup150_bat_props,
	.num_properties = ARRAY_SIZE(yup150_bat_props),
	.get_property = yup150_bat_get_property,
};

static struct device *dev;
static struct timer_list polling_timer;

static int new_cradle_status = -1;
static int new_ac_status = -1;
static int new_usb_status = -1;
static int new_charge_status = -1;
static int cradle_status = -1;
static int ac_status = -1;
static int usb_status = -1;
static int charge_status = -1;

static void update_status(void)
{
	new_cradle_status = !! yup150_is_cradle_online();
	new_ac_status = !! yup150_is_ac_online();
	new_usb_status = !! yup150_is_usb_online();
	new_charge_status = !! yup150_is_battery_charging();
}

static void polling_timer_func(unsigned long unused)
{
	dev_dbg(dev, "polling...\n");

	update_status();

	if (new_cradle_status != cradle_status) {
		power_supply_changed(&yup150_psy_cradle);
		cradle_status = new_cradle_status;
	}

	if (new_ac_status != ac_status) {
		power_supply_changed(&yup150_psy_ac);
		ac_status = new_ac_status;
	}

	if (new_usb_status != usb_status) {
		power_supply_changed(&yup150_psy_usb);
		usb_status = new_usb_status;
	}

	if (new_charge_status != charge_status) {
		power_supply_changed(&yup150_psy_bat);
		charge_status = new_charge_status;
	}
#if 0
	mod_timer(&polling_timer,
		  jiffies + msecs_to_jiffies(1000));
#endif
}

static int yup150_power_irq(int irq, void *data)
{
	mod_timer(&polling_timer,
		  jiffies + msecs_to_jiffies(1000));
	return IRQ_HANDLED;
}

static int yup150_power_probe(struct platform_device *pdev)
{
	int ret = 0;

	dev = &pdev->dev;

	if (pdev->id != -1) {
		dev_err(dev, "it's meaningless to register several "
			"yup150_powers; use id = -1\n");
		ret = -EINVAL;
		goto wrongid;
	}

	yup150_power_init_pin();

	ret = power_supply_register(&pdev->dev, &yup150_psy_cradle);
	if (ret) {
		dev_err(dev, "failed to register %s power supply\n",
			yup150_psy_cradle.name);
		goto cradle_supply_failed;
	}

	ret = power_supply_register(&pdev->dev, &yup150_psy_ac);
	if (ret) {
		dev_err(dev, "failed to register %s power supply\n",
			yup150_psy_ac.name);
		goto ac_supply_failed;
	}

	ret = power_supply_register(&pdev->dev, &yup150_psy_usb);
	if (ret) {
		dev_err(dev, "failed to register %s power supply\n",
			yup150_psy_usb.name);
		goto usb_supply_failed;
	}

	adc_client = s3c_adc_register(pdev, NULL, NULL, 0);
	if (adc_client) {
		ret = power_supply_register(&pdev->dev, &yup150_psy_bat);
		if (ret) {
			dev_err(dev, "failed to register %s power supply\n",
				yup150_psy_bat.name);
			goto bat_supply_failed;
		}
	} else {
		dev_err(&pdev->dev, "cannot register adc\n");
	}

	setup_timer(&polling_timer, polling_timer_func, 0);
#if 0
	mod_timer(&polling_timer, jiffies + msecs_to_jiffies(10));
#endif

	set_irq_type(IRQ_EINT6, IRQ_TYPE_EDGE_BOTH);
	ret = request_irq(IRQ_EINT6, yup150_power_irq, 0,
			"yup150-power-eint6", NULL);
	if (ret) {
		dev_err(dev, "request eint6 irq failed\n");
		goto eint6_irq_failed;
	}

	set_irq_type(IRQ_EINT3, IRQ_TYPE_EDGE_BOTH);
	ret = request_irq(IRQ_EINT3, yup150_power_irq, 0,
			"yup150-power-eint3", NULL);
	if (ret) {
		dev_err(dev, "request eint3 irq failed\n");
		goto eint3_irq_failed;
	}

	set_irq_type(IRQ_EINT5, IRQ_TYPE_EDGE_BOTH);
	ret = request_irq(IRQ_EINT5, yup150_power_irq, 0,
			"yup150-power-eint5", NULL);
	if (ret) {
		dev_err(dev, "request eint5 irq failed\n");
		goto eint5_irq_failed;
	}

	return 0;

eint5_irq_failed:
	free_irq(IRQ_EINT3, NULL);
eint3_irq_failed:
	free_irq(IRQ_EINT6, NULL);
eint6_irq_failed:
	power_supply_unregister(&yup150_psy_bat);
bat_supply_failed:
	power_supply_unregister(&yup150_psy_usb);
usb_supply_failed:
	power_supply_unregister(&yup150_psy_ac);
ac_supply_failed:
	power_supply_unregister(&yup150_psy_cradle);
cradle_supply_failed:
wrongid:
	return ret;
}

static int yup150_power_remove(struct platform_device *pdev)
{
	free_irq(IRQ_EINT5, NULL);
	free_irq(IRQ_EINT3, NULL);
	free_irq(IRQ_EINT6, NULL);

	del_timer_sync(&polling_timer);

	if (adc_client) {
		power_supply_unregister(&yup150_psy_bat);
		s3c_adc_release(adc_client);
	}
	power_supply_unregister(&yup150_psy_usb);
	power_supply_unregister(&yup150_psy_ac);
	power_supply_unregister(&yup150_psy_cradle);

	return 0;
}

#ifdef CONFIG_PM
static int yup150_power_suspend(struct platform_device *pdev, pm_message_t state)
{
	set_irq_type(IRQ_EINT6, IRQ_TYPE_EDGE_RISING);
	enable_irq_wake(IRQ_EINT6);
	return 0;
}

static int yup150_power_resume(struct platform_device *pdev)
{
	disable_irq_wake(IRQ_EINT6);
	set_irq_type(IRQ_EINT6, IRQ_TYPE_EDGE_BOTH);
	return 0;
}
#else
#define yup150_power_suspend	NULL
#define yup150_power_resume	NULL
#endif

static struct platform_driver yup150_power_pdrv = {
	.driver = {
		.name = "yup150-power",
	},
	.probe = yup150_power_probe,
	.remove = yup150_power_remove,
	.suspend = yup150_power_suspend,
	.resume = yup150_power_resume,
};

static int __init yup150_power_init(void)
{
	return platform_driver_register(&yup150_power_pdrv);
}

static void __exit yup150_power_exit(void)
{
	platform_driver_unregister(&yup150_power_pdrv);
}

module_init(yup150_power_init);
module_exit(yup150_power_exit);

MODULE_LICENSE("GPL");
