/* hsspi-s3c2443.c
 *
 * Copyright (C) 2006 Samsung Electronics Co. Ltd.
 *
 * S3C2443 SPI Controller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/device.h>
#include <linux/platform_device.h>

#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <mach/dma.h>
#include <mach/hardware.h>
#include <mach/regs-gpio.h>
#include <mach/regs-gpiol.h>
#include <mach/regs-s3c2443-clock.h>
#include <plat/regs-hsspi.h>

#include <mach/spi.h>

#define S3C_SPI_FIFO_DEPTH	(16 * 4)
//#define S3C_SPI_FIFO_DEPTH	16

#define CONFIG_SPICLK_PCLK	1

struct s3c_spi {
	/* bitbang has to be first */
        struct spi_bitbang       bitbang;
        struct completion        done;

        void __iomem            *regs;
        int                      irq;

        void                    (*set_cs)(struct s3c2410_spi_info *spi,
                                          int cs, int pol);

        /* data buffers */
        int                      len;
        const unsigned char     *tx;
	int			ntx;
        unsigned char           *rx;
	int			nrx;

        struct clk              *clk;
        struct resource         *ioarea;
        struct spi_master       *master;
        struct device           *dev;
        struct s3c2410_spi_info *pdata;
};

static inline struct s3c_spi *to_hw(struct spi_device *sdev)
{
        return spi_master_get_devdata(sdev->master);
}

static irqreturn_t s3c_spi_irq(int irq, void *dev)
{
        struct s3c_spi *hw = dev;
	unsigned long	sts;

	sts = readl(hw->regs + S3C_SPI_STATUS);

//        dev_dbg(hw->dev, "s3c_spi_irq -> status : %X\n", sts);

#if 0
	if (sts & SPI_STUS_TRAILCNT_ZERO) {
                printk("hspi : trailing count is zero [nrx:%d ntx:%d len:%d]\n",
			hw->nrx, hw->ntx, hw->len);
	}
#endif

	if (sts & SPI_STUS_RX_OVERRUN_ERR) {
                printk("hspi : Rx overrun error detected\n");
        }

        if (sts & SPI_STUS_RX_UNDERRUN_ERR) {
                printk("hspi : Rx underrun error detected\n");
        }

        if (sts & SPI_STUS_TX_OVERRUN_ERR) {
                printk("hspi : Tx overrun error detected\n");
        }

        if (sts & SPI_STUS_TX_UNDERRUN_ERR) {
                printk("hspi : Tx underrun error detected\n");
        }

	while (readl(hw->regs + S3C_SPI_STATUS) & SPI_STUS_RX_FIFORDY) {
		u8	data = readb(hw->regs + S3C_SPI_RX_DATA);

		if ((hw->nrx >= hw->ntx) || (hw->nrx >= hw->len))
			printk("SPI : unmatched RX [nrx:%d ntx:%d len:%d]\n",
				hw->nrx, hw->ntx, hw->len);

		if (hw->nrx < hw->len) {
			if (hw->rx)
				hw->rx[hw->nrx] = data;
			hw->nrx++;
		}
		if (hw->nrx >= hw->len) {
			unsigned long	inten;
			hw->rx = NULL;
			inten = readl(hw->regs + S3C_SPI_INT_EN);
			inten &= ~(SPI_INT_RX_FIFORDY_EN|SPI_INT_RX_UNDERRUN_EN|SPI_INT_RX_OVERRUN_EN|SPI_INT_TRAILING_EN);
			writel(inten, hw->regs + S3C_SPI_INT_EN);

			complete(&hw->done);
			break;
		}
	}

	while ((hw->ntx < hw->len) && (readl(hw->regs + S3C_SPI_STATUS) & SPI_STUS_TX_FIFORDY)) {
		if (hw->tx)
			writeb(hw->tx[hw->ntx], hw->regs + S3C_SPI_TX_DATA);
		else
			writeb(0, hw->regs + S3C_SPI_TX_DATA);
		hw->ntx++;
		if (hw->ntx >= hw->len) {
			unsigned long	inten;
			hw->tx = NULL;
			inten = readl(hw->regs + S3C_SPI_INT_EN);
			inten &= ~(SPI_INT_TX_FIFORDY_EN|SPI_INT_TX_UNDERRUN_EN|SPI_INT_TX_OVERRUN_EN);
			writel(inten, hw->regs + S3C_SPI_INT_EN);
//				complete(&hw->done);
			break;
		}
	}

	writel(0x1f, hw->regs + S3C_PENDING_CLR);

        return IRQ_HANDLED;
}

static int s3c_spi_txrx(struct spi_device *spi, struct spi_transfer *t)
{
        struct s3c_spi *hw = to_hw(spi);
	unsigned int inten = 0;

        dev_dbg(&spi->dev, "s3c_spi_txrx -> tx %p, rx %p, len %d\n",
                t->tx_buf, t->rx_buf, t->len);

        hw->tx = t->tx_buf;
	hw->ntx = 0;
        hw->rx = t->rx_buf;
        hw->nrx = 0;
        hw->len = t->len;

        init_completion(&hw->done);

	/* 4. Set SPI INT_EN register */
#if 0
	inten = SPI_INT_RX_FIFORDY_EN|SPI_INT_RX_UNDERRUN_EN|SPI_INT_RX_OVERRUN_EN|SPI_INT_TRAILING_EN;
#else
//	if (t->rx_buf)
		inten = SPI_INT_RX_FIFORDY_EN|SPI_INT_RX_UNDERRUN_EN|SPI_INT_RX_OVERRUN_EN|SPI_INT_TRAILING_EN;
#endif
//	if (t->tx_buf)
		inten |= SPI_INT_TX_FIFORDY_EN|SPI_INT_TX_UNDERRUN_EN|SPI_INT_TX_OVERRUN_EN;
	writel(0x1f, hw->regs + S3C_PENDING_CLR);
	writel(inten, hw->regs + S3C_SPI_INT_EN);

        wait_for_completion(&hw->done);

#if 1
        return hw->nrx;
#else
	if (t->rx_buf) {
//		printk("end of wait_for_completion : %d\n", hw->nrx);
		return hw->nrx;
	}
	if (t->tx_buf) {
		while (!(readl(hw->regs + S3C_SPI_STATUS) & SPI_STUS_TX_DONE))
			dev_dbg(&spi->dev, "s3c_spi_txrx -> waiting tx done\n");

		chcfg = readl(hw->regs + S3C_CH_CFG);
		chcfg &= ~SPI_CH_TXCH_ON;
		writel(chcfg, hw->regs + S3C_CH_CFG);
//		printk("end of wait_for_completion : %d\n", hw->ntx);
		return hw->ntx;
	}
//	printk("end of wait_for_completion : 0\n");
	return 0;
#endif
}

static int s3c_spi_setupxfer(struct spi_device *spi, struct spi_transfer *t)
{
        unsigned int bpw;

        dev_dbg(&spi->dev, "s3c_spi_setupxfer -> bits_per_word %d, len %d\n",
                (t ? t->bits_per_word : spi->bits_per_word), (t ? t->len : 0));

#if 0
	if (t && t->len > 0xFFFF)
		return -EINVAL;
#endif
        bpw = t ? t->bits_per_word : spi->bits_per_word;
	if (bpw == 0)
		bpw = 8;
	if (bpw != 8) {
                dev_err(&spi->dev, "invalid bits-per-word (%d)\n", bpw);
                return -EINVAL;
	}
	return 0;
}

static void s3c_spi_chipsel(struct spi_device *spi, int value)
{
        struct s3c_spi *hw = to_hw(spi);
        unsigned int cspol = spi->mode & SPI_CS_HIGH ? 1 : 0;

        dev_dbg(&spi->dev, "s3c_spi_chipsel -> cs %d, value %d\n",
		spi->chip_select, value);

        switch (value) {
        case BITBANG_CS_INACTIVE:
                hw->set_cs(hw->pdata, spi->chip_select, cspol^1);
                break;

        case BITBANG_CS_ACTIVE:
{
        unsigned int bpw;
        unsigned int hz;
        unsigned int div;
        unsigned int chcfg = 0;
	unsigned int clkcfg;
	unsigned int modecfg;
	unsigned int slavecfg;

        bpw = 8/*t->bits_per_word ? t->bits_per_word : spi->bits_per_word*/;
        hz  = /* t->speed_hz ? t->speed_hz : */ spi->max_speed_hz;

	/* hsspi software reset */
	writeb(readb(hw->regs + S3C_CH_CFG) | SPI_CH_SW_RST,
			hw->regs + S3C_CH_CFG);
	writeb(readb(hw->regs + S3C_CH_CFG) & (~SPI_CH_SW_RST),
			hw->regs + S3C_CH_CFG);

	/* 1. Set transfer type (CPOL & CPHA set) */
	if (spi->mode & SPI_CPHA)
		chcfg |= SPI_CH_FORMAT_B;
	else
		chcfg |= SPI_CH_FORMAT_A;
	if (spi->mode & SPI_CPOL)
		chcfg |= SPI_CH_FALLING;
	else
		chcfg |= SPI_CH_RISING;
	writel(chcfg, hw->regs + S3C_CH_CFG);

	/* 2. Set clock configuration register */
        clkcfg = SPI_ENCLK_ENABLE;
#if defined CONFIG_SPICLK_PCLK
        clkcfg |= SPI_CLKSEL_PCLK;
#elif defined CONFIG_SPICLK_EPLL
        clkcfg |= SPI_CLKSEL_ECLK;
#else
#error you must define correct confige file.
#endif

        div = clk_get_rate(hw->clk) / hz;
	/* is clk = pclk / (2 * (pre+1)), or is it
	 *    clk = (pclk * 2) / ( pre + 1) */
	div /= 2;
#if 0
        if (div > 0)
                div -= 1;
#endif
        if (div > 255)
                div = 255;
        dev_dbg(&spi->dev, "setting pre-scaler to %d (hz %d)\n", div, hz);
	clkcfg |= div;
        writel(clkcfg, hw->regs + S3C_CLK_CFG);

	/* 3. Set SPI MODE configuration register */
	modecfg = SPI_MODE_CH_TSZ_BYTE;
	modecfg |= SPI_MODE_TXDMA_OFF| SPI_MODE_SINGLE| SPI_MODE_RXDMA_OFF;
//	if (t->tx_buf)
		modecfg |= ((S3C_SPI_FIFO_DEPTH/2) << 5);
#if 0
	if (t->len <= S3C_SPI_FIFO_DEPTH)
		modecfg |= ((t->len-1) << 11);
	else
		modecfg |= ((S3C_SPI_FIFO_DEPTH/2) << 11);
#else
	modecfg |= (1 << 11);
#endif
	modecfg |= (0x3ff << 19);
	writel(modecfg, hw->regs + S3C_MODE_CFG);

#if 0
	/* 5. Set Packet Count configuration register */
	writel(t->len | SPI_PACKET_CNT_EN, hw->regs + S3C_PACKET_CNT);
#endif

	/* 6. Set Tx or Rx Channel on */
	chcfg = readl(hw->regs + S3C_CH_CFG);
#if 0
	chcfg |= SPI_CH_RXCH_ON;
#else
//	if (t->rx_buf)
		chcfg |= SPI_CH_RXCH_ON;
#endif
//	if (t->tx_buf)
		chcfg |= SPI_CH_TXCH_ON;
	writel(chcfg, hw->regs + S3C_CH_CFG);

	/* 7. Set nSS low to start Tx or Rx operation */
	slavecfg = readl(hw->regs + S3C_SLAVE_SEL);
//	slavecfg &= SPI_SLAVE_SIG_ACT;
	slavecfg &= ~SPI_SLAVE_SIG_INACT;
	writel(slavecfg, hw->regs + S3C_SLAVE_SEL);
}
                hw->set_cs(hw->pdata, spi->chip_select, cspol);
                break;
        }
}

/* the spi->mode bits understood by this driver: */
#define MODEBITS (SPI_CPOL | SPI_CPHA | SPI_CS_HIGH)

static int s3c_spi_setup(struct spi_device *spi)
{
        int ret;

	dev_dbg(&spi->dev, "s3c_spi_setup\n");

        if (!spi->bits_per_word)
                spi->bits_per_word = 8;

        if (spi->mode & ~MODEBITS) {
                dev_dbg(&spi->dev, "setup: unsupported mode bits %x\n",
                        spi->mode & ~MODEBITS);
                return -EINVAL;
        }

        ret = s3c_spi_setupxfer(spi, NULL);
        if (ret < 0) {
                dev_err(&spi->dev, "setupxfer returned %d\n", ret);
                return ret;
        }

        dev_dbg(&spi->dev, "%s: mode %d, %u bpw, %d hz\n",
                __func__, spi->mode, spi->bits_per_word,
                spi->max_speed_hz);

        return 0;
}

static int s3c_spi_hw_init(struct s3c_spi *hw)
{
	/* program defaults into the registers */
	writel(readl(S3C2443_SCLKCON)|(1<<14), S3C2443_SCLKCON);
        writel(readl(S3C2443_PCLKCON)|(1<<6), S3C2443_PCLKCON);

	writel(readl(S3C24XX_MISCCR)|S3C24XX_MISCCR_SPISEL, S3C24XX_MISCCR);

#if 0
	/* initialize the gpio */
	s3c2410_gpio_cfgpin(S3C2410_GPE11, S3C2410_GPE11_SPIMISO0);
	s3c2410_gpio_cfgpin(S3C2410_GPE12, S3C2410_GPE12_SPIMOSI0);
	s3c2410_gpio_cfgpin(S3C2410_GPE13, S3C2410_GPE13_SPICLK0);

	s3c2410_gpio_cfgpin(S3C2443_GPL13, S3C2443_GPL13_nSS0);
	s3c2410_gpio_cfgpin(S3C2443_GPL14, S3C2443_GPL14_nSS1);

	s3c2410_gpio_cfgpin(S3C2443_GPL11, S3C2443_GPL11_SPIMOSI1);
	s3c2410_gpio_cfgpin(S3C2443_GPL12, S3C2443_GPL12_SPIMISO1);
	s3c2410_gpio_cfgpin(S3C2443_GPL10, S3C2443_GPL10_SPICLK1);
#endif

#ifdef CONFIG_SPICLK_PCLK
	/*Enable PCLK into the HS SPI*/
	writel(readl(S3C2443_PCLKCON)|(1<<6), S3C2443_PCLKCON);

	clk_enable(hw->clk);

#elif defined CONFIG_SPICLK_EPLL
	/* implemetation when use EPLL clock */
	writel(0x800, S3C2443_LOCKCON1);
	writel( (readl( S3C2443_CLKSRC ) | (1 << 6) ), S3C2443_CLKSRC);  // EPLL Output

#if defined(CONFIG_CPU_S3C2443)
	writel((40<<16)|(1<<8)|(1<<0) ,S3C2443_EPLLCON);//96MHz
#elif defined(CONFIG_CPU_S3C2450) || defined(CONFIG_CPU_S3C2416)
	writel((48<<16)|(3<<8)|(1<<0) ,S3C2443_EPLLCON);//96MHz
#else
# error Any CPU Type is not selected
#endif

	writel( readl(S3C2443_EPLLCON)& (~(1<<24)) , S3C2443_EPLLCON );  //EPLL On
	writel(( readl(S3C2443_CLKDIV1) & (~(0x3<<24))) | (0x0 << 24) , S3C2443_CLKDIV1 ); // Epll Ratio is 1
	writel(( readl(S3C24XX_MISCCR) & (~(0x7<<8))) | (0x1 << 8) , S3C24XX_MISCCR );

	/* Epll, prescaler = 1					*/
	/* clock =  ( clock source / (2 * ( prescaler + 1)))	*/
	writel(0x501, hw->regs + S3C_CLK_CFG); 	//Use EPLL Clock and Clock On Prescaler 1
#else
#error you must define correct confige file.
#endif
	/* hspi software restet */

	writeb(readb(hw->regs + S3C_CH_CFG) | (1<<5), hw->regs + S3C_CH_CFG);
	writeb(readb(hw->regs + S3C_CH_CFG) & (~(1<<5)), hw->regs + S3C_CH_CFG);

        if (hw->pdata) {
#if 0
                if (hw->set_cs == s3c24xx_spi_gpiocs)
                        gpio_direction_output(hw->pdata->pin_cs, 1);
#endif
                if (hw->pdata->gpio_setup)
                        hw->pdata->gpio_setup(hw->pdata, 1);
        }

	return 0;
}

/* s3c_spi_probe
 *
 * called by the bus driver when a suitable device is found
 */

static int s3c_spi_probe(struct platform_device *pdev)
{
	struct s3c2410_spi_info *pdata;
	struct s3c_spi	*hw;
	struct spi_master *master;
	struct resource *res;
	int err = 0;

	master = spi_alloc_master(&pdev->dev, sizeof(struct s3c_spi));
	if (master == NULL) {
		dev_err(&pdev->dev, "No memory for spi_master\n");
		err = -ENOMEM;
		goto err_nomem;
	}

	hw = spi_master_get_devdata(master);
	memset(hw, 0, sizeof(struct s3c_spi));

	hw->master = spi_master_get(master);
	hw->pdata = pdata = pdev->dev.platform_data;
	hw->dev = &pdev->dev;

        if (pdata == NULL) {
                dev_err(&pdev->dev, "No platform data supplied\n");
                err = -ENOENT;
                goto err_no_pdata;
        }

	platform_set_drvdata(pdev, hw);
	init_completion(&hw->done);

        master->num_chipselect = hw->pdata->num_cs;
        master->bus_num = pdata->bus_num;

        /* setup the state for the bitbang driver */

        hw->bitbang.master         = hw->master;
        hw->bitbang.setup_transfer = s3c_spi_setupxfer;
        hw->bitbang.chipselect     = s3c_spi_chipsel;
        hw->bitbang.txrx_bufs      = s3c_spi_txrx;
        hw->bitbang.master->setup  = s3c_spi_setup;

	dev_dbg(hw->dev, "bitbang at %p\n", &hw->bitbang);

	/* find and map our resources */

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
        if (res == NULL) {
                dev_err(&pdev->dev, "Cannot get IORESOURCE_MEM\n");
                err = -ENOENT;
                goto err_no_iores;
        }

	hw->ioarea = request_mem_region(res->start, (res->end - res->start)+1,
                                        pdev->name);
	if (hw->ioarea == NULL) {
                dev_err(&pdev->dev, "Cannot reserve region\n");
                err = -ENXIO;
                goto err_no_iores;
        }

	hw->regs = ioremap(res->start, (res->end - res->start)+1);
        if (hw->regs == NULL) {
                dev_err(&pdev->dev, "Cannot map IO\n");
                err = -ENXIO;
                goto err_no_iomap;
        }

	hw->irq = platform_get_irq(pdev, 0);
        if (hw->irq < 0) {
                dev_err(&pdev->dev, "No IRQ specified\n");
                err = -ENOENT;
                goto err_no_irq;
        }

	err = request_irq(hw->irq, s3c_spi_irq, 0, pdev->name, hw);
        if (err) {
                dev_err(&pdev->dev, "Cannot claim IRQ\n");
                goto err_no_irq;
        }

	hw->clk = clk_get(&pdev->dev, "hsspi");
        if (IS_ERR(hw->clk)) {
                dev_err(&pdev->dev, "No clock for device\n");
                err = PTR_ERR(hw->clk);
                goto err_no_clk;
        }

	/* setup any gpio we can */

#if 0
        if (!pdata->set_cs) {
                if (pdata->pin_cs < 0) {
                        dev_err(&pdev->dev, "No chipselect pin\n");
                        goto err_register;
                }

                err = gpio_request(pdata->pin_cs, dev_name(&pdev->dev));
                if (err) {
                        dev_err(&pdev->dev, "Failed to get gpio for cs\n");
                        goto err_register;
                }

                hw->set_cs = s3c24xx_spi_gpiocs;
                gpio_direction_output(pdata->pin_cs, 1);
        } else
#endif
                hw->set_cs = pdata->set_cs;

	s3c_spi_hw_init(hw);

        /* register our spi controller */

        err = spi_bitbang_start(&hw->bitbang);
        if (err) {
                dev_err(&pdev->dev, "Failed to register SPI master\n");
                goto err_register;
        }

	printk(KERN_INFO "S3C2443 HSSPI Driver\n");

        return 0;

 err_register:
#if 0
        if (hw->set_cs == s3c24xx_spi_gpiocs)
                gpio_free(pdata->pin_cs);
#endif

        clk_disable(hw->clk);
        clk_put(hw->clk);

 err_no_clk:
        free_irq(hw->irq, hw);

 err_no_irq:
        iounmap(hw->regs);

 err_no_iomap:
        release_resource(hw->ioarea);
        kfree(hw->ioarea);

 err_no_iores:
 err_no_pdata:
        spi_master_put(hw->master);;

 err_nomem:
        return err;
}

/* s3c_spi_remove
 *
 * called when device is removed from the bus
*/
static int __exit s3c_spi_remove(struct platform_device *pdev)
{
	struct s3c_spi *hw = platform_get_drvdata(pdev);

	platform_set_drvdata(pdev, NULL);

	spi_unregister_master(hw->master);

	clk_disable(hw->clk);
	clk_put(hw->clk);

	free_irq(hw->irq, hw);
	iounmap(hw->regs);

#if 0
        if (hw->set_cs == s3c24xx_spi_gpiocs)
                gpio_free(hw->pdata->pin_cs);
#endif

	release_resource(hw->ioarea);
	kfree(hw->ioarea);

	spi_master_put(hw->master);
	return 0;
}

#ifdef CONFIG_PM
static int s3c_spi_suspend(struct platform_device *pdev, pm_message_t msg)
{
	struct s3c_spi *hw = platform_get_drvdata(pdev);
	if (hw->pdata && hw->pdata->gpio_setup)
                hw->pdata->gpio_setup(hw->pdata, 0);
	clk_disable(hw->clk);
	return 0;
}

static int s3c_spi_resume(struct platform_device *pdev)
{
	struct s3c_spi *hw = platform_get_drvdata(pdev);
	s3c_spi_hw_init(hw);
	return 0;
}
#else
#define s3c_spi_suspend NULL
#define s3c_spi_resume  NULL
#endif

/* device driver for platform bus bits */
MODULE_ALIAS("platform:s3c24xx-hsspi");
static struct platform_driver s3c_spi_driver = {
	.remove		= __exit_p(s3c_spi_remove),
	.suspend	= s3c_spi_suspend,
	.resume		= s3c_spi_resume,
	.driver		= {
		.name	= "s3c24xx-hsspi",
		.owner	= THIS_MODULE,
	},
};

static int __init s3c_spi_driver_init(void)
{
	return platform_driver_probe(&s3c_spi_driver, s3c_spi_probe);
}

static void __exit s3c_spi_driver_exit(void)
{
	platform_driver_unregister(&s3c_spi_driver);
}

module_init(s3c_spi_driver_init);
module_exit(s3c_spi_driver_exit);

MODULE_DESCRIPTION("S3C2443 HSSPI Bus driver");
MODULE_AUTHOR("Ryu Euiyoul<steven.ryu@samsung.com>");
MODULE_LICENSE("GPL");
