/*
 * alc5621.c  --  ALC5621 ALSA SoC Audio driver
 *
 * Copyright 2005 Openedhand Ltd.
 *
 * Author: Richard Purdie <richard@openedhand.com>
 *
 * Based on wm8731.c by Liam Girdwood
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>

#include "alc5621.h"

static int use_class_d_amp = 0;

#define ALC5621_VERSION "0.1"

#define MAX_RETRY	50

struct snd_soc_codec_device soc_codec_dev_alc5621;

/*
 * read from alc5621 register space
 */
static inline unsigned int alc5621_read(struct snd_soc_codec *codec,
	unsigned int reg)
{
	u8 data[3];
	struct i2c_client *i2c = (struct i2c_client *) codec->control_data;
	struct i2c_msg	msg[2];
	int i;

	data[0] = reg;

	msg[0].addr = i2c->addr;
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = data;

	msg[1].addr = i2c->addr;
	msg[1].flags = I2C_M_RD;
	msg[1].len = 2;
	msg[1].buf = data + 1;

	for (i = 0; i < MAX_RETRY; i++) {
		if (i2c_transfer(i2c->adapter, msg, 2) == 2)
			return ((unsigned int)data[1] << 8) | data[2];
		else {
			printk(KERN_ERR "ALC5621 : error on i2c_transfer\n");
		}
	}
	return 0;
}

/*
 * write to the ALC5621 register space
 */
static int alc5621_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int value)
{
	u8 data[3];
	int i;

	data[0] = reg;
	data[1] = (value >> 8) & 0xFF;
	data[2] = value & 0xFF;

	for (i = 0; i < MAX_RETRY; i++) {
		if (codec->hw_write(codec->control_data, data, 3) == 3) {
			return 0;
		} else {
			printk(KERN_ERR "ALC5621 : error on i2c_master_send\n");
		}
	}
	return -EIO;
}

/*
 * read from alc5621 index register space
 */
static inline unsigned int alc5621_index_read(struct snd_soc_codec *codec,
	unsigned int reg)
{
	if (alc5621_write(codec, ALC5621_IDXADR, reg) != 0)
		return 0;
	return alc5621_read(codec, ALC5621_IDXDAT);
}

/*
 * write to the ALC5621 index register space
 */
static int alc5621_index_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int value)
{
	if (alc5621_write(codec, ALC5621_IDXADR, reg) != 0)
		return -EIO;
	return alc5621_write(codec, ALC5621_IDXDAT, value);
}

#define alc5621_reset(c)	alc5621_write(c, ALC5621_RESET, 0)

static const char *alc5621_avol[] = {"1","2","4","8","16","32","64","128","256","512","1024"};

static const char *alc5621_spk_outn_source[] = {
	"RN/-(L+R)", "RP/L+R", "LN/-(L+R)", "Vmid"};
static const char *alc5621_spk_out_type_sel[] = {"Class AB", "Class D"};
static const char *alc5621_clsab_amp_sel[] = {"Strong Amp", "Weak Amp"};
static const char *alc5621_spkmono_in_sel[] = {
	"No input", "HP Mixer", "Speaker Mixer", "Mono Mixer"};
static const char *alc5621_hp_in_sel[] = {"No input", "HP Mixer"};

static const char *alc5621_mic_boost_ctrl[] = {"Bypass", "20dB", "30dB"};
static const char *alc5621_mic1_bias_voltage_ctrl[] = {"0.9AVDD", "0.75AVDD"};
static const char *alc5621_mic_bias_threshold[] = {"600uA", "1200uA", "1800uA"};

static const struct soc_enum alc5621_enum[] = {
	SOC_ENUM_SINGLE(ALC5621_AVOL, 0, 11, alc5621_avol),

	SOC_ENUM_SINGLE(ALC5621_OUTSW, 14, 4, alc5621_spk_outn_source),
	SOC_ENUM_SINGLE(ALC5621_OUTSW, 13, 2, alc5621_spk_out_type_sel),
	SOC_ENUM_SINGLE(ALC5621_OUTSW, 12, 2, alc5621_clsab_amp_sel),
	SOC_ENUM_SINGLE(ALC5621_OUTSW, 10, 4, alc5621_spkmono_in_sel),
	SOC_ENUM_DOUBLE(ALC5621_OUTSW, 9, 8, 2, alc5621_hp_in_sel),
	SOC_ENUM_SINGLE(ALC5621_OUTSW, 6, 4, alc5621_spkmono_in_sel),

	SOC_ENUM_SINGLE(ALC5621_MICCTL, 10, 3, alc5621_mic_boost_ctrl),
	SOC_ENUM_SINGLE(ALC5621_MICCTL, 8, 3, alc5621_mic_boost_ctrl),
	SOC_ENUM_SINGLE(ALC5621_MICCTL, 5, 2, alc5621_mic1_bias_voltage_ctrl),
	SOC_ENUM_SINGLE(ALC5621_MICCTL, 0, 3, alc5621_mic_bias_threshold),
};

static const struct snd_kcontrol_new alc5621_snd_controls[] = {
SOC_DOUBLE("Speaker Volume", ALC5621_SPKVOL, 8, 0, 31, 1),
SOC_DOUBLE("Speaker Mute", ALC5621_SPKVOL, 15, 7, 1, 0),
SOC_DOUBLE("Speaker ZC Switch", ALC5621_SPKVOL, 14, 6, 1, 0),
SOC_DOUBLE("Speaker Soft Mute", ALC5621_SPKVOL, 13, 5, 1, 0),

SOC_DOUBLE("Headphone Volume", ALC5621_HPVOL, 8, 0, 31, 1),
SOC_DOUBLE("Headphone Mute", ALC5621_HPVOL, 15, 7, 1, 0),
SOC_DOUBLE("Headphone ZC Switch", ALC5621_HPVOL, 14, 6, 1, 0),
SOC_DOUBLE("Headphone Soft Mute", ALC5621_HPVOL, 13, 5, 1, 0),

SOC_DOUBLE("AuxOut Volume", ALC5621_AOVOL, 8, 0, 31, 1),
SOC_DOUBLE("AuxOut Mute", ALC5621_AOVOL, 15, 7, 1, 0),
SOC_DOUBLE("AuxOut ZC Switch", ALC5621_AOVOL, 14, 6, 1, 0),
SOC_DOUBLE("AuxOut Soft Mute", ALC5621_AOVOL, 13, 5, 1, 0),

SOC_DOUBLE("AuxIn Volume", ALC5621_AIVOL, 8, 0, 31, 1),
SOC_SINGLE("AuxIn To Headphone Mixer Switch", ALC5621_AIVOL, 15, 1, 1),
SOC_SINGLE("AuxIn To Speaker Mixer Switch", ALC5621_AIVOL, 14, 1, 1),
SOC_SINGLE("AuxIn To Mono Mixer Switch", ALC5621_AIVOL, 13, 1, 1),

SOC_DOUBLE("LineIn Volume", ALC5621_LIVOL, 8, 0, 31, 1),
SOC_SINGLE("LineIn To Headphone Mixer Switch", ALC5621_LIVOL, 15, 1, 1),
SOC_SINGLE("LineIn To Speaker Mixer Switch", ALC5621_LIVOL, 14, 1, 1),
SOC_SINGLE("LineIn To Mono Mixer Switch", ALC5621_LIVOL, 13, 1, 1),

SOC_DOUBLE("Master Playback Volume", ALC5621_DACVOL, 8, 0, 31, 1),
#if 0
SOC_SINGLE("Master Playback To Headphone Mixer Switch", ALC5621_DACVOL, 15, 1, 1),
SOC_SINGLE("Master Playback To Speaker Mixer Switch", ALC5621_DACVOL, 14, 1, 1),
SOC_SINGLE("Master Playback To Mono Mixer Switch", ALC5621_DACVOL, 13, 1, 1),
#endif

SOC_SINGLE("Mic1 Volume", ALC5621_MICVOL, 8, 31, 1),
SOC_SINGLE("Mic2 Volume", ALC5621_MICVOL, 0, 31, 1),

SOC_SINGLE("Mic1 To Headphone Mixer Switch", ALC5621_MICSW, 15, 1, 1),
SOC_SINGLE("Mic1 To Speaker Mixer Switch", ALC5621_MICSW, 14, 1, 1),
SOC_SINGLE("Mic1 To Mono Mixer Switch", ALC5621_MICSW, 13, 1, 1),
SOC_SINGLE("Mic1 Differential Switch", ALC5621_MICSW, 12, 1, 0),

SOC_SINGLE("Mic2 To Headphone Mixer Switch", ALC5621_MICSW, 7, 1, 1),
SOC_SINGLE("Mic2 To Speaker Mixer Switch", ALC5621_MICSW, 6, 1, 1),
SOC_SINGLE("Mic2 To Mono Mixer Switch", ALC5621_MICSW, 5, 1, 1),
SOC_SINGLE("Mic2 Differential Switch", ALC5621_MICSW, 4, 1, 0),

SOC_DOUBLE("Capture Volume", ALC5621_ADCVOL, 7, 0, 31, 0),
SOC_DOUBLE("Capture To Headphone Mixer Switch", ALC5621_ADCVOL, 15, 14, 1, 1),
SOC_DOUBLE("Capture To Mono Mixer Switch", ALC5621_ADCVOL, 13, 12, 1, 1),
SOC_DOUBLE("Capture ZC Switch", ALC5621_ADCVOL, 6, 5, 1, 0),

SOC_DOUBLE("Capture From Mic1 Switch", ALC5621_ADCSW, 14, 6, 1, 1),
SOC_DOUBLE("Capture From Mic2 Switch", ALC5621_ADCSW, 13, 5, 1, 1),
SOC_DOUBLE("Capture From LineIn Switch", ALC5621_ADCSW, 12, 4, 1, 1),
SOC_DOUBLE("Capture From AuxIn Switch", ALC5621_ADCSW, 11, 3, 1, 1),
SOC_DOUBLE("Capture From Headphone Mixer Switch", ALC5621_ADCSW, 10, 2, 1, 1),
SOC_DOUBLE("Capture From Speaker Mixer Switch", ALC5621_ADCSW, 9, 1, 1, 1),
SOC_DOUBLE("Capture From Mono Mixer Switch", ALC5621_ADCSW, 8, 0, 1, 1),

SOC_ENUM("Avol Soft Volume Control Time", alc5621_enum[0]),

SOC_ENUM("SPK_Out_N Source Select", alc5621_enum[1]),
SOC_ENUM("SPK_Output Type Select", alc5621_enum[2]),
SOC_ENUM("Class AB Output Amp Select", alc5621_enum[3]),
SOC_ENUM("SPK Volume Output Input Select", alc5621_enum[4]),
SOC_ENUM("HP Volume Output Input Select", alc5621_enum[5]),
SOC_ENUM("MONO/AUX Volume Output Input Select", alc5621_enum[6]),

SOC_ENUM("Mic1 Boost Control", alc5621_enum[7]),
SOC_ENUM("Mic2 Boost Control", alc5621_enum[8]),
SOC_ENUM("MICBIAS1 Output Voltage Control", alc5621_enum[9]),
SOC_ENUM("MICBIAS1/2 Short Current Detector Threshold", alc5621_enum[10]),
};

/* add non dapm controls */
static int alc5621_add_controls(struct snd_soc_codec *codec)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(alc5621_snd_controls); i++) {
		err = snd_ctl_add(codec->card,
				  snd_soc_cnew(&alc5621_snd_controls[i],
						codec, NULL));
		if (err < 0)
			return err;
	}

	return 0;
}

static const struct snd_soc_dapm_widget alc5621_dapm_widgets[] = {
SND_SOC_DAPM_DAC("Left DAC", "Left Playback", ALC5621_PWR2, 9, 1),
SND_SOC_DAPM_DAC("Right DAC", "Right Playback", ALC5621_PWR2, 8, 1),
SND_SOC_DAPM_ADC("Left ADC", "Left Capture", ALC5621_PWR2, 7, 1),
SND_SOC_DAPM_ADC("Right ADC", "Right Capture", ALC5621_PWR2, 6, 1),
};

#if 0
/* Output Mixer */
static const struct snd_kcontrol_new alc5621_output_mixer_controls[] = {
SOC_DAPM_SINGLE("Line Bypass Switch", ALC5621_APANA, 3, 1, 0),
SOC_DAPM_SINGLE("Mic Sidetone Switch", ALC5621_APANA, 5, 1, 0),
SOC_DAPM_SINGLE("HiFi Playback Switch", ALC5621_APANA, 4, 1, 0),
};

/* Input mux */
static const struct snd_kcontrol_new alc5621_input_mux_controls =
SOC_DAPM_ENUM("Input Select", alc5621_enum[0]);

static const struct snd_soc_dapm_widget alc5621_dapm_widgets[] = {
SND_SOC_DAPM_MIXER("Output Mixer", ALC5621_PWR, 4, 1,
	&alc5621_output_mixer_controls[0],
	ARRAY_SIZE(alc5621_output_mixer_controls)),
SND_SOC_DAPM_DAC("DAC", "HiFi Playback", ALC5621_PWR, 3, 1),
SND_SOC_DAPM_OUTPUT("LOUT"),
SND_SOC_DAPM_OUTPUT("LHPOUT"),
SND_SOC_DAPM_OUTPUT("ROUT"),
SND_SOC_DAPM_OUTPUT("RHPOUT"),
SND_SOC_DAPM_ADC("ADC", "HiFi Capture", ALC5621_PWR, 2, 1),
SND_SOC_DAPM_MUX("Input Mux", SND_SOC_NOPM, 0, 0, &alc5621_input_mux_controls),
SND_SOC_DAPM_PGA("Line Input", ALC5621_PWR, 0, 1, NULL, 0),
SND_SOC_DAPM_MICBIAS("Mic Bias", ALC5621_PWR, 1, 1),
SND_SOC_DAPM_INPUT("MICIN"),
SND_SOC_DAPM_INPUT("RLINEIN"),
SND_SOC_DAPM_INPUT("LLINEIN"),
};

static const struct snd_soc_dapm_route intercon[] = {
	/* output mixer */
	{"Output Mixer", "Line Bypass Switch", "Line Input"},
	{"Output Mixer", "HiFi Playback Switch", "DAC"},
	{"Output Mixer", "Mic Sidetone Switch", "Mic Bias"},

	/* outputs */
	{"RHPOUT", NULL, "Output Mixer"},
	{"ROUT", NULL, "Output Mixer"},
	{"LHPOUT", NULL, "Output Mixer"},
	{"LOUT", NULL, "Output Mixer"},

	/* input mux */
	{"Input Mux", "Line In", "Line Input"},
	{"Input Mux", "Mic", "Mic Bias"},
	{"ADC", NULL, "Input Mux"},

	/* inputs */
	{"Line Input", NULL, "LLINEIN"},
	{"Line Input", NULL, "RLINEIN"},
	{"Mic Bias", NULL, "MICIN"},
};
#endif

static int alc5621_add_widgets(struct snd_soc_codec *codec)
{
	snd_soc_dapm_new_controls(codec, alc5621_dapm_widgets,
				  ARRAY_SIZE(alc5621_dapm_widgets));

#if 0
	snd_soc_dapm_add_routes(codec, intercon, ARRAY_SIZE(intercon));
#endif

	snd_soc_dapm_new_widgets(codec);
	return 0;
}

static int alc5621_hw_params(struct snd_pcm_substream *substream,
			    struct snd_pcm_hw_params *params,
			    struct snd_soc_dai *dai)
{
#if 0
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;
	struct alc5621_priv *alc5621 = codec->private_data;
	u16 iface = alc5621_read(codec, ALC5621_IFACE) & 0xfff3;
	int i = get_coeff(alc5621->sysclk, params_rate(params));
	u16 srate = (coeff_div[i].sr << 2) |
		(coeff_div[i].bosr << 1) | coeff_div[i].usb;

	alc5621_write(codec, ALC5621_SRATE, srate);

	/* bit size */
	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S16_LE:
		break;
	case SNDRV_PCM_FORMAT_S20_3LE:
		iface |= 0x0004;
		break;
	case SNDRV_PCM_FORMAT_S24_LE:
		iface |= 0x0008;
		break;
	}

	alc5621_write(codec, ALC5621_IFACE, iface);
#else
//	printk("alc5621_hw_params\n");
#endif
	return 0;
}

static int alc5621_pcm_prepare(struct snd_pcm_substream *substream,
			      struct snd_soc_dai *dai)
{
#if 0
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;

	/* set active */
	alc5621_write(codec, ALC5621_ACTIVE, 0x0001);
#else
//	printk("alc5621_pcm_prepare\n");
#endif

	return 0;
}

static void alc5621_shutdown(struct snd_pcm_substream *substream,
			    struct snd_soc_dai *dai)
{
#if 0
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;

	/* deactivate */
	if (!codec->active) {
		udelay(50);
		alc5621_write(codec, ALC5621_ACTIVE, 0x0);
	}
#else
//	printk("alc5621_shutdown\n");
#endif
}

static int alc5621_mute(struct snd_soc_dai *dai, int mute)
{
	struct snd_soc_codec *codec = dai->codec;

	u16 misc_reg = alc5621_read(codec, ALC5621_MISC);
	u16 dvol_reg = alc5621_read(codec, ALC5621_DACVOL);

	if (mute)  {
		alc5621_write(codec, ALC5621_DACVOL, dvol_reg & ~(7<<13));
		alc5621_write(codec, ALC5621_MISC, misc_reg | (3<<2));
	} else {
		alc5621_write(codec, ALC5621_MISC, misc_reg & ~(3<<2));
		alc5621_write(codec, ALC5621_DACVOL, dvol_reg & ~(7<<13));
	}

	return 0;
}

static int alc5621_set_dai_clkdiv(struct snd_soc_dai *codec_dai,
		int div_id, unsigned int div)
{
	struct snd_soc_codec *codec = codec_dai->codec;

	pr_debug("alc5621_set_dai_clkdiv -> %x\n", div);

	alc5621_write(codec, ALC5621_CLKCTL, div);
	return 0;
}

static int alc5621_set_dai_fmt(struct snd_soc_dai *codec_dai,
		unsigned int fmt)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	unsigned int	daireg;

	pr_debug("alc5621_set_dai_fmt -> %x\n", fmt);

	daireg = alc5621_read(codec, ALC5621_DAICTL);

	daireg &= ~0x03;
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		break;
	case SND_SOC_DAIFMT_RIGHT_J:
		daireg |= 1;
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		daireg |= 2;
		break;
	case SND_SOC_DAIFMT_DSP_A:
		daireg |= 3;
		daireg &= ~(1<<14);
		break;
	case SND_SOC_DAIFMT_DSP_B:
		daireg |= 3;
		daireg |= (1<<14);
		break;
	default:
		pr_debug("Unknown data format\n");
		return -EINVAL;
	}

	switch (fmt & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_NB_NF:
		daireg &= ~0x80;
		daireg &= ~0x30;
		break;
	case SND_SOC_DAIFMT_NB_IF:
		daireg &= ~0x80;
		daireg |= 0x30;
		break;
	case SND_SOC_DAIFMT_IB_NF:
		daireg |= 0x80;
		daireg &= ~0x30;
		break;
	case SND_SOC_DAIFMT_IB_IF:
		daireg |= 0x80;
		daireg |= 0x30;
		break;
	}

	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		daireg &= ~0x8000;
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		daireg |= 0x8000;
		break;
	default:
		pr_debug("unknwon master/slave format\n");
		return -EINVAL;
	}

	alc5621_write(codec, ALC5621_DAICTL, daireg);

	return 0;
}

static int alc5621_set_bias_level(struct snd_soc_codec *codec,
				 enum snd_soc_bias_level level)
{
#if 0
	u16 reg = alc5621_read(codec, ALC5621_PWR) & 0xff7f;

	switch (level) {
	case SND_SOC_BIAS_ON:
		/* vref/mid, osc on, dac unmute */
		alc5621_write(codec, ALC5621_PWR, reg);
		break;
	case SND_SOC_BIAS_PREPARE:
		break;
	case SND_SOC_BIAS_STANDBY:
		/* everything off except vref/vmid, */
		alc5621_write(codec, ALC5621_PWR, reg | 0x0040);
		break;
	case SND_SOC_BIAS_OFF:
		/* everything off, dac mute, inactive */
		alc5621_write(codec, ALC5621_ACTIVE, 0x0);
		alc5621_write(codec, ALC5621_PWR, 0xffff);
		break;
	}
#else
//	printk("alc5621_set_bias_level -> %d\n", (int) level);
#endif
	codec->bias_level = level;
	return 0;
}

#define ALC5621_RATES (SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 |\
		SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 |\
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 |\
		SNDRV_PCM_RATE_48000)

#define ALC5621_FORMATS (SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S20_3LE |\
	SNDRV_PCM_FMTBIT_S24_LE)

struct snd_soc_dai alc5621_dai = {
	.name = "ALC5621",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 1,
		.channels_max = 2,
		.rates = ALC5621_RATES,
		.formats = ALC5621_FORMATS,},
	.capture = {
		.stream_name = "Capture",
		.channels_min = 1,
		.channels_max = 2,
		.rates = ALC5621_RATES,
		.formats = ALC5621_FORMATS,},
	.ops = {
		.prepare = alc5621_pcm_prepare,
		.hw_params = alc5621_hw_params,
		.shutdown = alc5621_shutdown,
		.digital_mute = alc5621_mute,
		.set_clkdiv = alc5621_set_dai_clkdiv,
		.set_fmt = alc5621_set_dai_fmt,
	}
};
EXPORT_SYMBOL_GPL(alc5621_dai);

/*
 * initialise the ALC5621 driver
 * register the mixer and dsp interfaces with the kernel
 */
static void alc5621_codec_init(struct snd_soc_codec *codec)
{
	alc5621_reset(codec);

	msleep(1);

	/* power on device */
	alc5621_set_bias_level(codec, SND_SOC_BIAS_STANDBY);

	/* Pad drive[Strong], Power Gating[Enable], Vref Pwr Ctl[Enable] */
	alc5621_index_write(codec, ALC5621_IDX_DIGITAL, 0x9000);
	/* Main Bias Analog Circuit[Enable], others[Disable] */
	alc5621_write(codec, ALC5621_PWR3, 0x8000);
	/* VREF of All Analog Circuit[Enable], others[Disable] */
	alc5621_write(codec, ALC5621_PWR2, 0x2000);

	/* Speaker Output Volume : Mute */
	alc5621_write(codec, ALC5621_SPKVOL, 0x8080);
	/* Headphone Output Volume : Mute */
	alc5621_write(codec, ALC5621_HPVOL, 0x8080);
	if (use_class_d_amp) {
		/* SPK_OUT_N[RN], AMP[Class D] */
		alc5621_write(codec, ALC5621_OUTSW, 0x2700);
	} else {
		/* SPK_OUT_N[RN], AMP[Strong Class AB] */
		alc5621_write(codec, ALC5621_OUTSW, 0x0700);
	}

	/* AUXOUT[Diff], AB[1Vdd], D[1Vdd], AB[Differential Mode] */
	alc5621_write(codec, ALC5621_ACTL, 0x5C00);

	/* PowerOn */
	if (use_class_d_amp) {
		alc5621_index_write(codec, ALC5621_IDX_CLASSD, 0xCF00);
	}
	/* AB:X D:X AVREF:O PLL:X DAC:O ADC:X HPMixer:O SPKMixer:O */
	alc5621_write(codec, ALC5621_PWR2, 0x2738);
	/* Bias:O SPK_OUT:O HPAmp:O */
	alc5621_write(codec, ALC5621_PWR3, 0x9600);
	/* DAI:O HPAmp:Normal */
	alc5621_write(codec, ALC5621_PWR1, 0x8020);

	/* DACInput : Mute */
	alc5621_write(codec, ALC5621_MISC, 0x000C);
	if (use_class_d_amp) {
		alc5621_write(codec, ALC5621_PWR2, 0x6738);
	} else {
		alc5621_write(codec, ALC5621_PWR2, 0xA738);
	}
	alc5621_write(codec, ALC5621_SPKVOL, 0x0000);
	alc5621_write(codec, ALC5621_HPVOL, 0x0F0F);
 
	alc5621_write(codec, ALC5621_SPKVOL, 0x0000);
	alc5621_write(codec, ALC5621_HPVOL, 0x0F0F);
}

#if 0
static int alc5621_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

#if 1
	alc5621_write(codec, ALC5621_PWR3, 0);
	alc5621_write(codec, ALC5621_PWR1, 0);
	alc5621_write(codec, ALC5621_PWR2, 0);

	alc5621_index_write(codec, ALC5621_IDX_DIGITAL, 0x0000);
#else
	alc5621_set_bias_level(codec, SND_SOC_BIAS_OFF);
#endif
	return 0;
}

static int alc5621_resume(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

#if 1
	alc5621_codec_init(codec);
#else
	alc5621_set_bias_level(codec, SND_SOC_BIAS_STANDBY);
	alc5621_set_bias_level(codec, codec->suspend_bias_level);
#endif
	return 0;
}
#endif

static int alc5621_init(struct snd_soc_device *socdev)
{
	struct snd_soc_codec *codec = socdev->codec;
	int reg, ret = 0;

	codec->name = "ALC5621";
	codec->owner = THIS_MODULE;
	codec->read = alc5621_read;
	codec->write = alc5621_write;
	codec->set_bias_level = alc5621_set_bias_level;
	codec->dai = &alc5621_dai;
	codec->num_dai = 1;

	/* register pcms */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0) {
		printk(KERN_ERR "alc5621: failed to create pcms\n");
		goto pcm_err;
	}

	alc5621_codec_init(codec);
	msleep(1);
	alc5621_codec_init(codec);

	alc5621_add_controls(codec);
	alc5621_add_widgets(codec);
	ret = snd_soc_init_card(socdev);
	if (ret < 0) {
		printk(KERN_ERR "alc5621: failed to register card\n");
		goto card_err;
	}

	return ret;

card_err:
	snd_soc_free_pcms(socdev);
	snd_soc_dapm_free(socdev);
pcm_err:
	return ret;
}

static struct snd_soc_device *alc5621_socdev;

#if defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE)

/*
 * ALC5621 2 wire address is determined by GPIO5
 * state during powerup.
 *    low  = 0x1a
 *    high = 0x1b
 */

static int alc5621_i2c_probe(struct i2c_client *i2c,
			    const struct i2c_device_id *id)
{
	struct snd_soc_device *socdev = alc5621_socdev;
	struct snd_soc_codec *codec = socdev->codec;
	int ret;

	i2c_set_clientdata(i2c, codec);
	codec->control_data = i2c;
	ret = alc5621_init(socdev);
	if (ret < 0)
		pr_err("failed to initialise ALC5621\n");

	return ret;
}

static int alc5621_i2c_remove(struct i2c_client *client)
{
	struct snd_soc_codec *codec = i2c_get_clientdata(client);
	return 0;
}

static int alc5621_i2c_suspend(struct i2c_client *client, pm_message_t state)
{
	struct snd_soc_codec *codec = i2c_get_clientdata(client);

#if 1
	alc5621_write(codec, ALC5621_PWR3, 0);
	alc5621_write(codec, ALC5621_PWR1, 0);
	alc5621_write(codec, ALC5621_PWR2, 0);

	alc5621_index_write(codec, ALC5621_IDX_DIGITAL, 0x0000);
#else
	alc5621_set_bias_level(codec, SND_SOC_BIAS_OFF);
#endif
	return 0;
}

static int alc5621_i2c_resume(struct i2c_client *client)
{
	struct snd_soc_codec *codec = i2c_get_clientdata(client);

//	msleep(1000);
#if 1
	alc5621_codec_init(codec);
#else
	alc5621_set_bias_level(codec, SND_SOC_BIAS_STANDBY);
	alc5621_set_bias_level(codec, codec->suspend_bias_level);
#endif
	return 0;
}

static const struct i2c_device_id alc5621_i2c_id[] = {
	{ "alc5621", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, alc5621_i2c_id);

static struct i2c_driver alc5621_i2c_driver = {
	.driver = {
		.name = "ALC5621 I2C Codec",
		.owner = THIS_MODULE,
	},
	.probe =    alc5621_i2c_probe,
	.remove =   alc5621_i2c_remove,
	.suspend =  alc5621_i2c_suspend,
	.resume =   alc5621_i2c_resume,
	.id_table = alc5621_i2c_id,
};

static int alc5621_add_i2c_device(struct platform_device *pdev,
				 const struct alc5621_setup_data *setup)
{
	struct i2c_board_info info;
	struct i2c_adapter *adapter;
	struct i2c_client *client;
	int ret;

	ret = i2c_add_driver(&alc5621_i2c_driver);
	if (ret != 0) {
		dev_err(&pdev->dev, "can't add i2c driver\n");
		return ret;
	}

	memset(&info, 0, sizeof(struct i2c_board_info));
	info.addr = setup->i2c_address;
	strlcpy(info.type, "alc5621", I2C_NAME_SIZE);

	adapter = i2c_get_adapter(setup->i2c_bus);
	if (!adapter) {
		dev_err(&pdev->dev, "can't get i2c adapter %d\n",
			setup->i2c_bus);
		goto err_driver;
	}

	client = i2c_new_device(adapter, &info);
	i2c_put_adapter(adapter);
	if (!client) {
		dev_err(&pdev->dev, "can't add i2c device at 0x%x\n",
			(unsigned int)info.addr);
		goto err_driver;
	}

	return 0;

err_driver:
	i2c_del_driver(&alc5621_i2c_driver);
	return -ENODEV;
}
#endif

static int alc5621_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct alc5621_setup_data *setup;
	struct snd_soc_codec *codec;
//	struct alc5621_priv *alc5621;
	int ret = 0;

	pr_info("ALC5621 Audio Codec %s", ALC5621_VERSION);

	setup = socdev->codec_data;
	codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL);
	if (codec == NULL)
		return -ENOMEM;

#if 0
	alc5621 = kzalloc(sizeof(struct alc5621_priv), GFP_KERNEL);
	if (alc5621 == NULL) {
		kfree(codec);
		return -ENOMEM;
	}

	codec->private_data = alc5621;
#endif
	socdev->codec = codec;
	mutex_init(&codec->mutex);
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	alc5621_socdev = socdev;
	ret = -ENODEV;

#if defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE)
	if (setup->i2c_address) {
		codec->hw_write = (hw_write_t)i2c_master_send;
		ret = alc5621_add_i2c_device(pdev, setup);
	}
#endif

	if (ret != 0) {
//		kfree(codec->private_data);
		kfree(codec);
	}
	return ret;
}

/* power down chip */
static int alc5621_remove(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

	if (codec->control_data)
		alc5621_set_bias_level(codec, SND_SOC_BIAS_OFF);

	snd_soc_free_pcms(socdev);
	snd_soc_dapm_free(socdev);
#if defined(CONFIG_I2C) || defined(CONFIG_I2C_MODULE)
	i2c_unregister_device(codec->control_data);
	i2c_del_driver(&alc5621_i2c_driver);
#endif
//	kfree(codec->private_data);
	kfree(codec);

	return 0;
}

struct snd_soc_codec_device soc_codec_dev_alc5621 = {
	.probe = 	alc5621_probe,
	.remove = 	alc5621_remove,
#if 0
	.suspend = 	alc5621_suspend,
	.resume =	alc5621_resume,
#endif
};
EXPORT_SYMBOL_GPL(soc_codec_dev_alc5621);

static int __init alc5621_modinit(void)
{
	return snd_soc_register_dai(&alc5621_dai);
}
module_init(alc5621_modinit);

static void __exit alc5621_exit(void)
{
	snd_soc_unregister_dai(&alc5621_dai);
}
module_exit(alc5621_exit);

MODULE_DESCRIPTION("ASoC ALC5621 driver");
MODULE_LICENSE("GPL");
