/*
 * Modifications by Christian Pellegrin <chripell@evolware.org>
 *
 * s3c24xx_alc5621.c  --  S3C24XX_ALC5621 ALSA SoC Audio board driver
 *
 * Copyright 2007 Dension Audio Systems Ltd.
 * Author: Zoltan Devai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/clk.h>
#include <linux/mutex.h>
#include <linux/gpio.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/plat-s3c24xx/regs-s3c2412-iis.h>

#include "s3c24xx-pcm.h"
#include "s3c2412-i2s.h"

#include "../codecs/alc5621.h"

static struct platform_device *s3c24xx_alc5621_snd_device;

static int s3c24xx_alc5621_hw_params(struct snd_pcm_substream *substream,
					struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->dai->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->dai->cpu_dai;
	static unsigned int fstab[] = { 256, 384, 0 };
	int ret = 0;
	unsigned long rate = params_rate(params);
	struct s3c2412_rate_calc	rate_inf;

	pr_debug("%s\n", __func__);

	ret = s3c2412_iis_calc_rate(&rate_inf, fstab, rate,
					s3c2412_get_iisclk());
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_fmt(codec_dai, SND_SOC_DAIFMT_I2S |
				SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_fmt(cpu_dai, SND_SOC_DAIFMT_I2S |
				SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

#if 0
	ret = snd_soc_dai_set_sysclk(cpu_dai, clk_source , clk,
				SND_SOC_CLOCK_OUT);
	if (ret < 0)
		return ret;
#endif

	ret = snd_soc_dai_set_clkdiv(cpu_dai, S3C2412_DIV_RCLK,
				rate_inf.fs_div);
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_clkdiv(cpu_dai, S3C2412_DIV_BCLK,
				S3C2412_IISMOD_BCLK_32FS);
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_clkdiv(cpu_dai, S3C2412_DIV_PRESCALER,
				rate_inf.clk_div);
	if (ret < 0)
		return ret;

#if 0
	/* set the codec system clock for DAC and ADC */
	ret = snd_soc_dai_set_sysclk(codec_dai, 0, clk,
				SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;
#else
	if (rate_inf.fs_div == 256)
		ret = snd_soc_dai_set_clkdiv(codec_dai, 0, 0x0A69);
	else if (rate_inf.fs_div == 384)
		ret = snd_soc_dai_set_clkdiv(codec_dai, 0, 0x0C6B);
	if (ret < 0)
		return ret;
#endif

	return 0;
}

static struct snd_soc_ops s3c24xx_alc5621_ops = {
	.hw_params = s3c24xx_alc5621_hw_params,
};

static struct snd_soc_dai_link s3c24xx_alc5621_dai_link = {
	.name = "ALC5621",
	.stream_name = "ALC5621",
	.codec_dai = &alc5621_dai,
	.cpu_dai = &s3c2412_i2s_dai,
	.ops = &s3c24xx_alc5621_ops,
};

static struct snd_soc_card snd_soc_s3c24xx_alc5621 = {
	.name = "S3C24XX_ALC5621",
	.platform = &s3c24xx_soc_platform,
	.dai_link = &s3c24xx_alc5621_dai_link,
	.num_links = 1,
};

static struct alc5621_setup_data soc_codec_data_alc5621 = {
	.i2c_bus = 0,
	.i2c_address = 0x1a,
};

static struct snd_soc_device s3c24xx_alc5621_snd_devdata = {
	.card = &snd_soc_s3c24xx_alc5621,
	.codec_dev = &soc_codec_dev_alc5621,
	.codec_data = &soc_codec_data_alc5621,
};

static int __init s3c24xx_alc5621_init(void)
{
	int ret;

	printk(KERN_INFO "S3C24XX_ALC5621 SoC Audio driver\n");

	s3c24xx_alc5621_snd_device = platform_device_alloc("soc-audio", -1);
	if (!s3c24xx_alc5621_snd_device) {
		printk(KERN_ERR "S3C24XX_ALC5621 SoC Audio: "
		       "Unable to register\n");
		return -ENOMEM;
	}

	platform_set_drvdata(s3c24xx_alc5621_snd_device,
			     &s3c24xx_alc5621_snd_devdata);
	s3c24xx_alc5621_snd_devdata.dev = &s3c24xx_alc5621_snd_device->dev;
	ret = platform_device_add(s3c24xx_alc5621_snd_device);
	if (ret) {
		printk(KERN_ERR "S3C24XX_ALC5621 SoC Audio: Unable to add\n");
		platform_device_put(s3c24xx_alc5621_snd_device);
	}

	return ret;
}

static void __exit s3c24xx_alc5621_exit(void)
{
	platform_device_unregister(s3c24xx_alc5621_snd_device);
}

module_init(s3c24xx_alc5621_init);
module_exit(s3c24xx_alc5621_exit);

MODULE_DESCRIPTION("S3C24XX_ALC5621 ALSA SoC audio driver");
MODULE_LICENSE("GPL");
